/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	26-Jul-2005
 *
 * Added support for uncertainties, 09-Mar-2023, CT
 */

#include "usrbin.h"

#include "math/eps.h"
#include "tool/bstring.h"
#include "tool/fortran.h"

const char* Usrbin::_typeStr[] = {
		"XYZ_point",
		"RPhiZ_point",
		"region_point",
		"|X|_point",
		"|Y|_point",
		"|Z|_point",
		"|XYZ|_point",
		"RPhi|Z|_point",
		"special_point",
		"",
		"XYZ",
		"RPhiZ",
		"region",
		"|X|",
		"|Y|",
		"|Z|",
		"|XYZ|",
		"RPhi|Z|",
		"special",
		""
	};

constexpr char Usrbin::USRBINErrHeader[];

// ================================== Usrbin ===================================
/** cleanup */
void Usrbin::cleanup()
{
	MeshData::cleanup();
	runtitle[0] = 0;
	runtime[0]  = 0;

	lntzer = 0;
	bkusbn = 0.0;
	b2usbn = 0.0;
	tcusbn = 0.0;

	weipri = 0.0;
	ncase  = 0;
	mcase  = 0;
	nbatch = 0;

	//_total = 0.0;
} // cleanup

/** load data and possible errors if they exists
 * @return true  on success
 *         false on failure
 */
bool Usrbin::load(const char* fn, int det)
{
	// NEW: we now load usrbins with their associated errors
	switch (preLoad(fn,det)) {
		case  1: return true;
		case -1: return false;
		default: ;
	}

	FortranFile f(fn,"rb");
	if (!f) return false;

	char buffer[512];
	FortranParser parser(buffer, sizeof(buffer));
	int length = f.read(parser);
	parser.read(runtitle, RunTit_Length);
	parser.read(runtime,  RunTim_Length);
	weipri = parser.readFloat();
	ncase  = parser.readInt();
	if (parser.more()) mcase  = parser.readInt();
	if (parser.more()) nbatch = parser.readInt();

	for (int i=1; ;i++) {
		/* Read detector information */
		length = f.read(parser);
		if (length != USRBINDataHeaderLen) break;
		if (length<0) goto ERROR;
		if (i!=det) {
			/* skip detector */
			f.skip();
			continue;
		}

		/* load detector */
		parser.skipInt();
		char usrbin_name[TitUsb_Length+1];
		parser.read(usrbin_name, TitUsb_Length);
		name(usrbin_name);
		_type  = (Type)parser.readInt();
		_score = parser.readInt();

		xlow  = (double)parser.readFloat();
		xhigh = (double)parser.readFloat();
		nx    =         parser.readInt();
		dx    = (double)parser.readFloat();

		ylow  = (double)parser.readFloat();
		yhigh = (double)parser.readFloat();
		ny    =         parser.readInt();
		dy    = (double)parser.readFloat();

		zlow  = (double)parser.readFloat();
		zhigh = (double)parser.readFloat();
		nz    =         parser.readInt();
		dz    = (double)parser.readFloat();

		if (not In(_type, Type::region_point,  Type::region,
				  Type::special_point, Type::special)) {
			// Re-normalize dx,dy,dz to get more precision from float
			dx = (xhigh-xlow) / (double)nx;
			dy = (yhigh-ylow) / (double)ny;
			dz = (zhigh-zlow) / (double)nz;
		}

		lntzer = parser.readInt();
		bkusbn = parser.readFloat();
		b2usbn = parser.readFloat();
		tcusbn = parser.readFloat();

		_detector = det;
		allocateData();
		length = nx*ny*nz*sizeof(float);
		if (f.read(data, length) != length) goto ERROR;
		switch (_type) {
			case Type::XYZ_point:
			case Type::XYZ:
				break;
			case Type::RPhiZ_point:
			case Type::RPhiZ:
				rphizMesh = true;
				break;
			case Type::region_point:
			case Type::region:
				regionMesh = true;
				break;
			case Type::Xsym_point:
			case Type::Xsym:
				xsym = true;
				break;
			case Type::Ysym_point:
			case Type::Ysym:
				ysym = true;
				break;
			case Type::Zsym_point:
			case Type::Zsym:
				zsym = true;
				break;
			case Type::XYZsym_point:
			case Type::XYZsym:
				xsym = ysym = zsym = true;
				break;
			case Type::RPhiZsym_point:
			case Type::RPhiZsym:
				rphizMesh = true;
				zsym = true;
				break;
			case Type::special_point:
			case Type::special:
				break;
			default:
				std::cerr << "ERROR unknown USRBIN type " << (int)_type << std::endl;
				cleanup();
				return false;
		}
		postLoad();
		printf("\tType:%d [%s]\tScore:%d\n",(int)_type,_typeStr[(int)_type],_score);
	}

	// check if we have statistics and read the corresponding uncertainties
	if (length == USRBINErrHeaderLen) {
		for (int i=1; i<det; i++) f.skip();
		allocateError();
		length = nx*ny*nz*sizeof(float);
		if (f.read(_error, length) != length)
			goto ERROR;
		else
			_errorsPresent = true;

		// DEBUG
		/*
		if ( _errorsPresent ) {
			std::ofstream out( "usrbin-debug.dat" );
			if ( !out ) {
				std::cerr << "cannot write userbin debug" << std::endl;
				return true;
			}

			out << "Detector: " << name() << std::endl << std::endl;
			out << "------ DATA -----" << std::endl;

			for ( int i = 1; i <= nx * ny * nz; ++i ) {
				out << std::scientific << _error[i-1] << " ";
				if ( !(i % 10) )
					out << std::endl;
			}

			out.close();
		}
		*/
	}
	return true;

ERROR:
	std::cerr << "ERROR reading USRBIN file \"" << _filename << '\"' << std::endl;
	cleanup();
	return false;
} // load

/** save usrbin data to file
 * @param   filename
 * @return  true on success
 *          false on failure
 */
bool Usrbin::save(const char* fn)
{
	if (fn) _filename = fn;
	FortranFile f(_filename.c_str(), "wb");
	if (!f) return false;

	if (!writeHeader(f)) return false;
	if (!writeData(f))   return false;

	// write errors if available
	if (_errorsPresent) {
		if (!writeErrorHeader(f)) return false;
		if (!writeError(f)) return false;
	}

	return true;
} // save

/** save */
bool Usrbin::save(const char* fn, std::vector<Usrbin>& usrbins)
{
	if (usrbins.empty()) return false;
	FortranFile f(fn, "wb");
	if (!f) return false;
	usrbins[0].writeHeader(f);
	for (int i=0; i<(int)usrbins.size(); i++)
		usrbins[i].writeData(f,i+1);
	if (usrbins[0].errorsPresent()) {
		writeErrorHeader(f);
		for (auto usrbin : usrbins)
			usrbin.writeError(f);
	}
	return true;
} // save

/** writeHeader */
bool Usrbin::writeHeader(FortranFile& f)
{
	char buffer[512];
	FortranParser parser(buffer, sizeof(buffer));

	// Write title
	parser.write(runtitle, RunTit_Length);
	parser.write(runtime,  RunTim_Length);
	parser.write(weipri);
	parser.write(ncase);
	parser.write(mcase);
	parser.write(nbatch);
	f.write(parser);
	return true;
} // writeHeader

/** writeData */
bool Usrbin::writeData(FortranFile& f, int det)
{
	char buffer[512];
	FortranParser parser(buffer, sizeof(buffer));

	// Write detector header
	parser.write(det==0? _detector:det);		// detector number
	parser.write(name().c_str(), TitUsb_Length);	// usrbin detector title
	parser.write((int)_type);			// type
	parser.write(_score);				// score

	parser.write((float)xlow);
	parser.write((float)xhigh);
	parser.write(nx);
	parser.write((float)dx);

	parser.write((float)ylow);
	parser.write((float)yhigh);
	parser.write(ny);
	parser.write((float)dy);

	parser.write((float)zlow);
	parser.write((float)zhigh);
	parser.write(nz);
	parser.write((float)dz);

	parser.write(lntzer);
	parser.write(bkusbn);
	parser.write(b2usbn);
	parser.write(tcusbn);
	assert(parser.length() == USRBINDataHeaderLen);
	f.write(parser);

	// Write data
	int length = totalSize()*sizeof(float);
	if (f.write(data, length) != length) {
		std::cerr << "ERROR writing usrbin file" << std::endl;
		return false;
	}
	return true;
} // writeData

/** writeErrorHeader */
bool Usrbin::writeErrorHeader(FortranFile& f)
{
	char buffer[20];
	FortranParser parser(buffer, sizeof(buffer));
	parser.write(USRBINErrHeader, sizeof(USRBINErrHeader) / sizeof(char) - 1);
	parser.write(1);
	assert(parser.length() == USRBINErrHeaderLen);
	f.write(parser);
	return true;
} // writeErrorHeader

/** writeError */
bool Usrbin::writeError(FortranFile& f)
{
	int length = totalSize()*sizeof(*_error);
	if (f.write(_error, length) != length) {
		std::cerr << "ERROR writing uncertainties to usrbin file " << std::endl;
		return false;
	}
	return true;
} // writeError

/** create */
void Usrbin::create(int det, Type atype, int ascore,
		 const double axlow, const double axhigh, const int anx,
		 const double aylow, const double ayhigh, const int any,
		 const double azlow, const double azhigh, const int anz)
{
	cleanup();

	_detector = det;
	_type = atype;
	_score = ascore;

	nx = Max(1, anx);
	ny = Max(1, any);
	nz = Max(1, anz);
	allocateData();
	allocateError();

	xlow = axlow;
	ylow = aylow;
	zlow = azlow;

	xhigh = axhigh;
	yhigh = ayhigh;
	zhigh = azhigh;

	dx = (xhigh - xlow) / (double)nx;
	dy = (yhigh - ylow) / (double)ny;
	dz = (zhigh - zlow) / (double)nz;

	assert(xlow < xhigh);
	assert(ylow < yhigh);
	assert(zlow < zhigh);
} // create

/** getDetectorNames */
std::vector<std::string> Usrbin::getDetectorNames(const char* fn)
{
	std::vector<std::string> detNames;

	FortranFile f(fn, "rb");
	if (!f) return detNames;

	char buffer[512];
	FortranParser parser(buffer, sizeof(buffer));
	f.skip();
	while (true) {
		/* Read detector information */
		int length = f.read(parser);
		if (length != USRBINDataHeaderLen ) break;
		if (length < 0) break;

		parser.skipInt();
		char usrbin_name[TitUsb_Length + 1];
		parser.read(usrbin_name, TitUsb_Length);
		detNames.push_back(usrbin_name);
		f.skip();		 /* skip data */
	}
	return detNames;
} // getDetectorNames
