////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//

#ifndef ANISO_MICRO_DIST_H
#define ANISO_MICRO_DIST_H

#include "bxdfbase.h"
#include "bxdfmicrofacet.h"
#include "microfacetdistribution.h"

class CAnisotropicMicroDist : public CMicrofacetDistribution
{
public:
  CAnisotropicMicroDist( double X, double Y );
  virtual ~CAnisotropicMicroDist(void);

	double D( const Vector& HalfVec, const Vector& Normal ) const override
	{
    // float CosThetah = fabs( CosTheta(HalfVec) );
    double CosThetah = fabs( ( Normal.dot( HalfVec ) ) );
    double e = (m_ExpX * HalfVec.x * HalfVec.x + m_ExpY * HalfVec.y * HalfVec.y) / (1.0 - CosThetah * CosThetah);
    return static_cast<double>( sqrt(( m_ExpX + 2) * (m_ExpY + 2 ) ) / PI * pow( CosThetah, e) );
  }

  virtual void    Sample_f( const Vector& Out, Vector& In, const Vector& Normal, double u1, double u2, float& Pdf ) const override;
  virtual float   GetPDF( const Vector& Out, const Vector& In, const Vector& Normal ) const override;
  void            SampleFirstQuadrant(double u1, double u2, double& Phi, double& CosTheta) const;

private:
  double m_ExpX;
  double m_ExpY;
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Trowbridge Reitz - microfacets

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CTrowbridgeReitzDist : public CMicrofacetDistribution {
public:
	static inline double RoughnessToAlpha( double Roughness ) {
		Roughness = std::max( Roughness, 1e-3 );
		double x = std::log( Roughness );
		return 1.62142 + 0.819955 * x + 0.1734 * x * x + 0.0171201 * x * x * x + 0.000640711 * x * x * x * x;
	};

	CTrowbridgeReitzDist( double AlphaX, double AlphaY, bool bSampleVisArea = true ) : CMicrofacetDistribution(), m_AlphaX( std::max( double( 0.001 ), AlphaX ) ), m_AlphaY( std::max( double( 0.001 ), AlphaY ) ), m_bSampleVisArea(bSampleVisArea) {};

	void SetAlpha( double AlphaX, double AlphaY ) { m_AlphaX = AlphaX; m_AlphaY = AlphaY; }
	void SetRoughness( double x, double y ) { m_AlphaX = CTrowbridgeReitzDist::RoughnessToAlpha( x ); m_AlphaY = CTrowbridgeReitzDist::RoughnessToAlpha( y ); }

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	double D( const Vector& wh, const Vector& Normal ) const

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		double tan2Theta = Tan2Theta( wh, Normal );
		if(std::isinf( tan2Theta ))
			return 0.;

		const double cos4Theta = Cos2Theta( wh, Normal ) * Cos2Theta( wh, Normal );
		double e = (Cos2Phi( wh, Normal ) / ( m_AlphaX * m_AlphaX) + Sin2Phi( wh, Normal ) / (m_AlphaY * m_AlphaY)) * tan2Theta;
		return 1 / (PI * m_AlphaX * m_AlphaY * cos4Theta * (1 + e) * (1 + e));
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	double G( const Vector& Out, const Vector& In, const Vector& wh, const Vector& Normal ) const

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		UNUSED( wh );
		return 1 / (1 + Lambda( Out, Normal ) + Lambda( In, Normal ));
	}

	float  GetPDF( const Vector& Out, const Vector& In, const Vector& Normal ) const;

	void   Sample_f( const Vector& Out, Vector& In, const Vector& Normal, double u1, double u2, float& Pdf ) const;

private:

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	double Lambda( const Vector& Out, const Vector& Normal ) const

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		double AbsTanTheta = std::abs( TanTheta( Out, Normal ) );
		if(std::isinf( AbsTanTheta )) return 0.;

		// Compute _alpha_ for Out direction
		double Alpha = std::sqrt( Cos2Phi( Out, Normal ) * m_AlphaX * m_AlphaX + Sin2Phi( Out, Normal ) * m_AlphaY * m_AlphaY );
		double Alpha2Tan2Theta = (Alpha * AbsTanTheta) * (Alpha * AbsTanTheta);

		return (-1 + std::sqrt( 1.0 + Alpha2Tan2Theta )) / 2;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	double G1( const Vector& Out, const Vector& In, const Vector& wh, const Vector& Normal ) const

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		UNUSED( In );

		if (Out.dot(wh) * Out.dot( Normal ) < 0.)
			return 0.;

		return 1 / (1 + Lambda( Out, Normal ));
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static Vector TrowbridgeReitzSample( const Vector& wi, const Vector& Normal, double AlphaX, double AlphaY, double u1, double u2 )

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		// 1. stretch wi
		Vector wiStretched = Vector( AlphaX, AlphaY, 1 );
		wiStretched.normalize();
		wiStretched = CBaseSampler::canonicalToBasis( wiStretched, Normal );

		wiStretched.x *= wi.x, wiStretched.y *= wi.y, wiStretched.z *= wi.z;

		// 2. simulate P22_{wi}(x_slope, y_slope, 1, 1)
		double SlopeX, SlopeY;
		TrowbridgeReitzSample11( CosTheta( wiStretched, Normal ), u1, u2, &SlopeX, &SlopeY );

		// 3. rotate
		double tmp = CosPhi( wiStretched, Normal ) * SlopeX - SinPhi( wiStretched, Normal ) * SlopeY;
		SlopeY = SinPhi( wiStretched, Normal ) * SlopeX + CosPhi( wiStretched, Normal ) * SlopeY;
		SlopeX = tmp;

		// 4. unstretch
		SlopeX = AlphaX * SlopeX;
		SlopeY = AlphaY * SlopeY;

		// 5. compute normal
		Vector RetVec( -SlopeX, -SlopeY, 1. );
		RetVec.normalize();

		return RetVec;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static void TrowbridgeReitzSample11( double cosTheta, double u1, double u2, double* SlopeX, double* SlopeY )

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{		// special case (normal incidence)
		if(cosTheta > .9999) {
			double r = sqrt( u1 / (1 - u1) );
			double Phi = 2 * PI * u2;
			*SlopeX = r * cos( Phi );
			*SlopeY = r * sin( Phi );
			return;
		}

		double sinTheta = std::sqrt( std::max( 0.0, 1.0 - cosTheta * cosTheta ) );
		double tanTheta = sinTheta / cosTheta;
		double a = 1 / tanTheta;
		double G1 = 2 / (1 + std::sqrt( 1. + 1. / (a * a) ));

		// sample slope_x
		double A = 2 * u1 / G1 - 1;
		double Tmp = 1.0 / (A * A - 1.0);

		if(Tmp > 1e10)
			Tmp = 1e10;

		double B = tanTheta;
		double D = std::sqrt( std::max( double( B * B * Tmp * Tmp - (A * A - B * B) * Tmp ), double( 0 ) ) );
		double SlopeX1 = B * Tmp - D;
		double SlopeX2 = B * Tmp + D;
		*SlopeX = (A < 0 || SlopeX2 > 1.0 / tanTheta) ? SlopeX1 : SlopeX2;

		// sample slope_y
		double S;
		if(u2 > 0.5) {
			S = 1.0;
			u2 = 2.0 * (u2 - .5);
		}
		else {
			S = -1.0;
			u2 = 2.0 * (.5 - u2);
		}
		double z = (u2 * (u2 * (u2 * 0.27385 - 0.73369) + 0.46341)) / (u2 * (u2 * (u2 * 0.093073 + 0.309420) - 1.000000) + 0.597999);

		*SlopeY = S * z * std::sqrt( 1.0 + *SlopeX * *SlopeX );

	}


private:
	double			 m_AlphaX, m_AlphaY;
	bool				 m_bSampleVisArea;
};





#endif
// vim:ts=2
