////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#include "architectureintegrator.h"
#include "geometry/engine.h"
#include "viewer/d3layer.h"
#include "viewer/viewer.h"
#include "math/randsampler.h"
#include "math/niederreiter.h"
#include "rendermaterial.h"
#include "os.h"

#include <string>
#include <set>

#ifndef MSWIN
	#include <unistd.h>
#endif

using namespace std;

CArchiIntegrator::CArchiIntegrator() : CBaseIntegrator(), m_AOCMinIntensity( 0.0 ), m_AOCMaxIntensity(1.0), m_AOCMaxDist(INFINITY ), m_NrOfAOCSamples(48), m_MaxAOCDepth(0), m_AOCSampleMultiplier(1),
m_bTexAvailable(false)
{
}

CArchiIntegrator::~CArchiIntegrator( void )
{
	for ( size_t i = 0; i < m_PhysMatLUT.size(); ++i )
		delete m_PhysMatLUT[i];

	m_PhysMatLUT.clear();
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CArchiIntegrator::Li( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0, int CallerId /*= cStandard | cFirstCall for first call*/)

// main integrator routine returning the luminance for the passed ray & rendering pass
// [4/27/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( CallerId );

	Point	 Hit;
	const	 Material* pMaterial = nullptr;
	Color	 ShadingColor;
	Ray		 lightRay;

	// check next intersection on secondary ray generations
	if( pRay->depth > 0 && !m_pVisLayer->nextIntersection( pEngine, pRay ) ) {
		return ShadingColor;	// on end of ray, error or blackbody
	}

	// check if we have to deal with a projection of a data mesh instead of real geometry
	if ( pRay->project_insight ) {
		while ( true ) {

			// prevent reflection of data as it looks a bit strange and can lead to confusion even though it would physically be correct
			// TODO: prevent reflections by checking the ray generation??
			if ( pRay->project_insight && pEngine->applyProject( pRay, pRay->T() ) ) {

				pRay->project_alpha = m_pVisLayer->viewer.usrbin.shade( pRay->hit(), ShadingColor );

				if ( pRay->project_alpha != 0xFF ) {
					return ShadingColor;
				}
				else { // we hit a full transparent part of the USRBIN data. so let's continue tracing
					pRay->moveby( SMALL3D3 );
					pRay->project_hit = false;

					if ( !m_pVisLayer->nextIntersection( pEngine, pRay ) ) {
						return ShadingColor;	// on end of ray, error or blackbody
					}

					continue;
				}
			}

			break;
		} // end of while
	}

	// Get region/voxel color

	if( pRay->voxelreg > 0 ) {
		ShadingColor.set( m_pVisLayer->viewer.voxel().roiColor( pRay->voxelreg ) );

		if ( m_RenderMode == CBaseIntegrator::cSpeed ) {
			Color totalIntensity = ShadeHitPoint( pEngine, pRay, RenderPass, Light0 );
			float ambient = (float)m_pVisLayer->ambient / 128.0f;
			ShadingColor *= totalIntensity + Color( ambient, ambient, ambient );			// we intentionally brighten the ambient light a bit w.r.t. normal calculations
		}

		return ShadingColor;		// return to FARM which, based on the selected rendering quality, it will trigger the voxel renderer to obtain full volume rendering effects or not
	}
	else {
		pMaterial = pRay->hitRegion()->region()->material();
		ShadingColor = GetHitColor( pRay );
	}

		// If we are in the start of the plane do not apply any shadow
	if( pRay->start() ) return ShadingColor;

	//
	// calculate light intensity
	//
	Color totalIntensity = ShadeHitPoint( pEngine, pRay, RenderPass, Light0 );
	float ambient = (float)m_pVisLayer->ambient / 128.0f;
	ShadingColor *= totalIntensity + Color( ambient, ambient, ambient);				// we intentionally brighten the ambient light a bit w.r.t. normal calculations

	if(pRay->depth <= pRay->max_depth  &&		// We now allow going to 1 sub-generation even in speed mode in order to
		pMaterial != nullptr ) {								// be able to treat data projections in transparent objects.

		// refraction & transmission
		if(!pRay->hitRegion()->notTransparent()) {
			int oldBodyId = pEngine->bodyCheckId();		// Transmission requires spawning secondary rays and thus, we need to advance the ray cache
			Color TmpColor = Li_Transmission(pEngine, pRay, RenderPass, Light0);
			if(pRay->project_hit) {
				// for data we use blending instead of addition as we would do in normal ray tracing. this way we remain closer to the original color shades of the data
				ShadingColor.blend( TmpColor, 1.0 - pRay->hitRegion()->alpha() / 255.0 );
				pRay->project_hit = false;		// this has been treated by the renderer already and does not require further action like the std. transparency by
																			// blending via the d3layer. So we reset the hit flag for data projections.
			}
			else {
				ShadingColor += TmpColor;
			}
			pEngine->bodyCheckId(oldBodyId);	// restore the ray cache to ensure that there is no conflict with subsequent calls to obtain screen door transparency via d3layer::ShadeXRay()
		}
	}

	ShadingColor.clamp();

	return ShadingColor;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CArchiIntegrator::CalcAmbientOcclusion( GeometryEngine* eng, Ray* pixelRay,
																							 int nrOfSamples, const double cMaxDist, const double cMinIntensity,
																							 const double cMaxIntensity )


//	ambient occlusion integrator
// [6/3/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Point    hit;
	Vector   sampleDir, bodyNormal;
	Ray      ambiOccRay;
	double   aocFactor = 0.0, transparency = 0.0;
	double   randNr1, randNr2;

	// Find body normal
	hit = pixelRay->hit( -SMALL3D4 );
	bodyNormal = pixelRay->normal;

	// check that the normal is opposite to the direction
	if ( bodyNormal * pixelRay->dir() > 0.0 ) {
		bodyNormal.negate();
	}

	{
		// inherit and propagate general ray a
		ambiOccRay.use_clip = pixelRay->use_clip;
		ambiOccRay.use_project = pixelRay->use_project;

		int ScrambleBxDF = eng->random.integer( cPMJLUTSize );

		for ( int i = 0; i < nrOfSamples; i++ ) {
			CBaseSampler::getPMJ02Sample( i, randNr1, randNr2, ScrambleBxDF );
			sampleDir = CBaseSampler::uniformSampleHemisphere( bodyNormal, randNr1, randNr2 );

			ambiOccRay.init();
			RaySegment new_segment( hit, sampleDir, pixelRay->prevZone() );
			new_segment.tmin = SMALL3D3;
			new_segment.tmax = cMaxDist * 2.0;
			ambiOccRay.push( new_segment );

			int oldBodyCheckId = eng->incBodyCheckId();

			if ( !eng->intersectRay( &ambiOccRay, true ) ) {		// NOTE: if we pass false we would ignore projection planes which can create artifacts, CT 18/8/22
																													// return false = intersection occurred!
				VZone* pZone = ambiOccRay.hitZone();
				if ( pZone != nullptr &&
						 (pZone->region()->type() == GRegion::Type::NORMAL ||
						 pZone->region()->type() == GRegion::Type::VOXEL) ) {
					eng->bodyCheckId( oldBodyCheckId );
					if ( ambiOccRay.T() <= cMaxDist ) {
						// ATTENTION: the higRegion->transparent() function only returns true for FULLY transparent regions --> we need to check alpha!!
						transparency = ambiOccRay.hitRegion()->alpha() / 255.0;
						aocFactor += (1.0 - transparency) * sampleDir * bodyNormal;		// take transparency into account when calculating AOC
						continue;
					}
				}
			}

			eng->bodyCheckId( oldBodyCheckId );
			ambiOccRay.pop();
		}
	} // end of sequential

	aocFactor = (1.0 - aocFactor / ((double)nrOfSamples)) * 1.0 / M_PI;

	return cMinIntensity + (cMaxIntensity - cMinIntensity) * aocFactor;


} // calcAmbientOcclusion


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CArchiIntegrator::Li_Transmission( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// [5/4/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color						RefractionColor, TransparencyColor(1.0, 1.0, 1.0);	// maybe transparency color should be gray?
	Ray							RefractionRay;
	double					Dot = 0.0;

	// disable clipping bodies for refraction rays, to avoid visualization artifacts of background bleeding
	RefractionRay.use_clip = RefractionRay.clip_insight = false;

	// sample refraction BxDF to obtain direction of incoming light + the corresponding rho
	GetRefractionRay( *pRay, RefractionRay, pRay->normal );
	RefractionRay.use_project = pRay->use_project;

	// calculate incoming reflected luminance
	Dot = -pRay->normal.dot( RefractionRay.dir() );
	if( Dot > 0 ) {
		pEngine->incBodyCheckId();		// signal to cache that we have a new ray direction
		RefractionColor += Li( pEngine, &RefractionRay, RenderPass, Light0, cTransmission ) * Dot;

		// handle potential intersection with data mesh
		pRay->project_hit |= RefractionRay.project_hit;
		if(RefractionRay.project_hit) {
			pRay->project_alpha = max( pRay->project_alpha, RefractionRay.project_alpha );	// propagate back alpha channel to camera ray
		}
	}

	return RefractionColor * TransparencyColor * ( pRay->hitRegion()->alpha() / 255.0 );

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CArchiIntegrator::InitFromConfig()

// [8/19/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	assert( m_pRenderer );	// ensure the base-class Init() has been called before
	string FileName;

	// show config flags
	if( !m_pRenderer->GetConfig().Empty() ) {
#if _DEBUG>1
		m_pRenderer->GetConfig().Show();
#endif
  }

	// set up ambient occlusion
	/*
	m_AOCMinIntensity = 0.0; // m_pRenderer->GetConfig().GetValue<double>( "aoc_min", 0.0 );
	m_AOCMaxIntensity = 1.0; // m_pRenderer->GetConfig().GetValue<double>( "aoc_max", 1.0 );
	m_AOCMaxDist = 100; // m_pRenderer->GetConfig().GetValue<double>( "aoc_dist", 25.0 );
	m_MaxAOCDepth = 0; // m_pRenderer->GetConfig().GetValue<int>( "aoc_maxdepth", 0 );
	m_AOCSampleMultiplier = m_pRenderer->GetConfig().GetValue<int>( "aoc_sample_multiplier", 1 );
	*/

	SetupBxDFLUT();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CArchiIntegrator::ShadeHitPoint( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// [8/19/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( RenderPass );
	UNUSED( Light0 );
	double totalIntensity = 0.0;


	// All light rays should start from outside the object
	// Move a bit backwards
	// Should be bigger than what is inside distance

	if ( pRay->depth <= m_MaxAOCDepth ) {
		totalIntensity = CalcAmbientOcclusion( pEngine, pRay, m_NrOfAOCSamples * m_AOCSampleMultiplier,
																					 m_AOCMaxDist, m_AOCMinIntensity, m_AOCMaxIntensity );
	}

	return Color( totalIntensity, totalIntensity, totalIntensity);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CArchiIntegrator::GetRefractionRay( const Ray& Out, Ray& In, const Vector& Normal )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	/*
	/!\/!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\
	/!\
	/!\ For a physically correct refraction ray the code in BxDFSpecTransmission applying Snell's law should be used. However,
	/!\ in this case we don't care but we want pure "look-through" transparency and no distortions on curved surfaces!
  /!\
	/!\/!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\

	*/

	UNUSED( Normal );

	Vector T = Out.dir();
	// set up ray
	In = Out;
	In.init();
	In.depth = Out.depth + 1;

	RaySegment NewSegment( Out.hit( SMALL3D4 ), T, Out.prevZone() );
	NewSegment.tmin = SMALL3D3;
	NewSegment.tmax = INFINITE;

	In.push( NewSegment );

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Colored architectural renderer
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CColoredArchiIntegrator::ShadeHitPoint( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

//  [12/1/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( RenderPass );

	Point  Hit;
	const  Material* pMaterial = pRay->hitRegion()->region()->material();
	double randNr[2];
	Ray    lightRay;
	Color	 Rho( Color::White );
	double pdf = 1.0, GeoTMin = 0.0;
	Vector ShadingNormal, GeoNormal, GeoHit;

	UNUSED( GeoTMin );

	// All light rays should start from outside the object
	// Move a bit backwards
	// Should be bigger than what is inside distance

	GeoHit = pRay->hit();
	if ( pMaterial ) {
		ShadingNormal = m_PhysMatLUT[0]->GetShadingNormal( pRay );
	}
	Hit = pRay->hit( -SMALL3D3 );

	// Calculate illumination
	Color totalIntensity = Color::Black, curIntensity = Color::Black;
	Vector dir, curLPos;

	// ignore artificial clipping & projection bodies to avoid vis artifacts
	//lightRay.use_clip = false;
	lightRay.use_clip = pRay->use_clip;
	lightRay.use_project = false;

	int Scramble1 = pEngine->random.integer();
	int Scramble2 = pEngine->random.integer();

	for ( int i = 0; i < Min( pRay->nlights, m_pGeometry->nlights ); i++ ) {
		const Light& L = m_pGeometry->light[Light0 + i];	// Geometry Light properties
		const VLight& VL = m_pVisLayer->light[i];				// Viewer light coordinate & direction
		double  dist, power, penumbraScaling = 0.0;

		const int nsamples = L.nsamples;
		//assert(InRange(0,nsamples,100));
		curIntensity = Color::Black;

		// volumetric lights
		for ( int lightSample = 0; lightSample < nsamples; ++lightSample ) {
			switch ( L.type ) {
			case Light::Type::Sun:
				dir = -VL.dir;
				dist = INFINITE;
				power = L.power;
				if ( L.shadow && pRay->shadow ) {
					double dot = m_pVisLayer->view().matrix( 0, 2 ) * dir.x +
						m_pVisLayer->view().matrix( 1, 2 ) * dir.y +
						m_pVisLayer->view().matrix( 2, 2 ) * dir.z;
					if ( dot > SMALL3D3 && m_pVisLayer->view().projection != ViewPort::Projection::Perspective ) {
						dist = (m_pVisLayer->view().matrix( 0, 2 ) * (m_pVisLayer->view().matrix( 0, 3 ) - Hit.x) +
										 m_pVisLayer->view().matrix( 1, 2 ) * (m_pVisLayer->view().matrix( 1, 3 ) - Hit.y) +
										 m_pVisLayer->view().matrix( 2, 2 ) * (m_pVisLayer->view().matrix( 2, 3 ) - Hit.z)) /
							dot;
						if ( dist < SMALL3D3 ) continue;
					}
				}
				break;

			case Light::Type::Omni:
			case Light::Type::Spot:
				// move the light position for volumetric light sources
				curLPos = VL.pos;
				if ( lightSample > 0 ) {
					// we found that the (0,2) series by VDC & Sobol2 aggregation shows highest volumetric convergence
					// at relatively low sample rates. Thus, we stick to this one here instead of Niederreiter or PMJ02.
					randNr[0] = CBaseSampler::VanDerCorput( lightSample, Scramble1 );
					randNr[1] = CBaseSampler::Sobol2( lightSample, Scramble2 );

					dir = (VL.pos - Hit);
					dir.normalize();
					curLPos += CBaseSampler::uniformSampleDisk( dir, L.radius, randNr[0], randNr[1] );
				}
				dir = curLPos - Hit;
				dist = dir.normalize();

				if ( dist < SMALL3D3 ) continue;
				if ( dist > L.distance ) continue;
				if ( L.falloff != Light::FallOff::Constant ) {
					// FIXME could be converted to switch ()
					if ( L.falloff == Light::FallOff::Inverse )
						power = L.power * 100.0 / dist;
					else	// must be Light::FallOff::Inverse2
						power = L.power * 10000.0 / Sq( dist );
					if ( power < 0.001 ) continue;
				}
				else
					power = L.power;

				if ( L.shadow && pRay->shadow ) {
					double dot = m_pVisLayer->view().matrix( 0, 2 ) * dir.x +
						m_pVisLayer->view().matrix( 1, 2 ) * dir.y +
						m_pVisLayer->view().matrix( 2, 2 ) * dir.z;
					if ( dot > SMALL3D3 ) {
						double pdist = (m_pVisLayer->view().matrix( 0, 2 ) * (m_pVisLayer->view().matrix( 0, 3 ) - Hit.x) +
														 m_pVisLayer->view().matrix( 1, 2 ) * (m_pVisLayer->view().matrix( 1, 3 ) - Hit.y) +
														 m_pVisLayer->view().matrix( 2, 2 ) * (m_pVisLayer->view().matrix( 2, 3 ) - Hit.z)) /
							dot;
						dist = Min( dist, pdist );
						if ( dist < SMALL3D3 ) continue;
					}
				}
				power /= (double)nsamples;   // normalize the light power w.r.t. light samples
				break;

			default:
				continue;
			}

			// set up light ray & find intersection with light
			lightRay.init();

			RaySegment new_segment( Hit, dir, pRay->prevZone() );

			lightRay.push( new_segment );

			//double dot = dir * GeoNormal;
			double dot = dir * ShadingNormal;
			if ( dot < 0.00001 ) continue;

			if ( L.shadow && pRay->shadow ) {
				int oldBodyCheckId = pEngine->incBodyCheckId();

				if ( !pEngine->intersectRay( &lightRay, false ) ) {
					// Check if we hit a black hole or an error
					VZone* z = lightRay.hitZone();
					if ( z != nullptr ) {
						if ( (z->region()->type() == GRegion::Type::NORMAL ||
								 z->region()->type() == GRegion::Type::VOXEL) ) {
							pEngine->bodyCheckId( oldBodyCheckId );

							// account for "light shadows" due to transparent objects
							double Transparency = z->region()->alpha() / 255.0;
							penumbraScaling += Transparency / (nsamples);		// take transparency into account when calculating AOC
							continue;
						}
						else
							if ( z->region()->type() == GRegion::Type::BLACKHOLE )
								penumbraScaling = 1.0;
					}

				}
				else {
					if ( L.type == Light::Type::Sun )
						penumbraScaling = 1.0;	// no penumbra by the sun. we only use 1
																		// sample for the sunlight for reasons of efficiency
					else
						penumbraScaling += 1.0 / (double)nsamples;
				}
				pEngine->bodyCheckId( oldBodyCheckId );
			}
			else // no shadowing requested
				penumbraScaling = 1.0;

			////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
			// Evaluate BxDF for current material
			//
			if ( pMaterial ) {
				Rho = m_PhysMatLUT[0]->f( *pRay, lightRay, ShadingNormal, *pMaterial, Color( 1.0, 1.0, 1.0 ), power );
			}

			lightRay.pop();

			// take BxDF for current sample into account
			curIntensity += Rho * (power * dot / pdf);

			// bail out early after 1 sample in case of sunlight
			if ( L.type == Light::Type::Sun && lightSample == 0 )
				break;
		} // end of sampling for one light source

		totalIntensity += curIntensity * penumbraScaling;

	} // end of light loop

	// ambient occlusion
	double ambiOcc = 0;
	if ( pRay->depth <= m_MaxAOCDepth ) {
		ambiOcc = CalcAmbientOcclusion( pEngine, pRay, m_NrOfAOCSamples * m_AOCSampleMultiplier,
																					 m_AOCMaxDist, m_AOCMinIntensity, m_AOCMaxIntensity );
	}

	return totalIntensity * min(1.0, ambiOcc *  2.0 * M_PI);	// the multiplication with PI is physically wrong but gives better artistic effects
}

// vim:ts=2