////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Author: Vasilis.Vlachoudis@cern.ch
//			   Christian.Theis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// BxDF helper functions for Fresnel dielectrica
//

#pragma once
#ifndef BXDF_FRESNEL_H
#define BXDF_FRESNEL_H

#include <algorithm>
#include <cmath>

#include "bxdfbase.h"

class CFresnelHelpers
{
public:
	static double FresnelApproxEta( const double Fr ) {
		double Reflectance = Clamp( Fr, 0., .999 );
		double SqrRefl = std::sqrt( Reflectance );
		return (1. + SqrRefl) / (1. - SqrRefl);
	};

	static double FresnelApproxK( const double Fr ) {
		double Reflectance = Clamp( Fr, 0., .999 );
		return 2. * (Reflectance / std::sqrt( 1. - Reflectance ));
	};

	// Fresnel factor for dielectrica
	static double FrDiel( double CosI, double CosT, double EtaI, double EtaT ) {
		// R_parallel
		double Rparl = ((EtaT * CosI) - (EtaI * CosT)) / ((EtaT * CosI) + (EtaI * CosT));
		// R_perpendicular
		double Rperp = ((EtaI * CosI) - (EtaT * CosT)) / ((EtaI * CosI) + (EtaT * CosT));
		return (Rparl * Rparl + Rperp * Rperp) / 2.;
	};

	// Fresnel factor for conductors
	static double FrCond( double CosI, double Eta, double k ) {
		double Tmp = (Eta * Eta + k * k) * CosI * CosI;
		double Rparl2 = (Tmp - (2. * Eta * CosI) + 1) / (Tmp + (2. * Eta * CosI) + 1);
		double Tmp_f = Eta * Eta + k * k;
		double Rperp2 = (Tmp_f - (2. * Eta * CosI) + CosI * CosI) / (Tmp_f + (2. * Eta * CosI) + CosI * CosI);

		return (Rparl2 + Rperp2) / 2.;
	};

	static double SchlickFresnel( const double Rs, double CosTheta ) {
		double Temp = (1.0 - CosTheta) * (1.0 - CosTheta) * (1.0 - CosTheta) * (1.0 - CosTheta) * (1.0 - CosTheta);
		return Rs + (1.0 - Rs) * Temp;
	};
};


class CFresnel {
public:
	virtual ~CFresnel() {};
	virtual double Evaluate( double CosI, double Eta, double k ) = 0;
};

class CFresnelConductor : public CFresnel
{
public:
	CFresnelConductor() {};
	virtual ~CFresnelConductor() {};

	virtual double Evaluate( double CosI, double Eta, double k ) { return CFresnelHelpers::FrCond( fabs( CosI ), Eta, k ); };

};

class CFresnelPreDefConductor : public CFresnelConductor
{
public:
	CFresnelPreDefConductor( double k, double Eta ) : m_Absorption( k ), m_Eta( Eta ) {};
	virtual double	Evaluate( double CosI, double Dummy1 = 0, double Dummy2 = 0 ) { UNUSED( Dummy1 ); UNUSED( Dummy2 ); return CFresnelConductor::Evaluate( CosI, m_Eta, m_Absorption ); };
	void		SetEta( double Eta ) { m_Eta = Eta; };
	void		SetAbsorption( double k ) { m_Absorption = k; };

protected:
	double m_Absorption;
	double m_Eta;
};

class CFresnelDiel : public CFresnel
{
public:
	CFresnelDiel() {};
	virtual ~CFresnelDiel() {};

	// Par1 = EtaI, Par2 = EtaT
	virtual double Evaluate( double CosI, double EtaI, double EtaT ) {
		// Compute Fresnel reflectance for dielectric
		CosI = Clamp( CosI, -1., 1. );

		// Compute indices of refraction for dielectric
		// Compute _sint_ using Snell's law
		double SinT = EtaI / EtaT * std::sqrt( std::max( 0., 1. - CosI * CosI ) );
		if( SinT > 1. ) {
			// Handle total internal reflection
			return 1.;
		}
		else {
			double CosT = std::sqrt( std::max( 0., 1. - SinT * SinT ) );
			return CFresnelHelpers::FrDiel( fabs( CosI ), CosT, EtaI, EtaT );
		}
	};
};

class CFresnelPreDefDiel : public CFresnelDiel
{
public:
	CFresnelPreDefDiel( double EtaI, double EtaT ) : m_EtaI( EtaI ), m_EtaT( EtaT) {};
	virtual double	Evaluate( double CosI, double Dummy1 = 0, double Dummy2 = 0 ) { UNUSED( Dummy1 ); UNUSED( Dummy2 ); return CFresnelDiel::Evaluate( CosI, m_EtaI, m_EtaT ); };
	void						SetEta( double EtaI, double EtaT ) { m_EtaI = EtaI; m_EtaT = EtaT; };

protected:
	double m_EtaI;
	double m_EtaT;
};

#endif
// vim:ts=2
