////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M   -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//          Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "bxdflafortune.h"

#include <cmath>
#include <climits>

CBxDFLaFortune::CBxDFLaFortune( const Color& Diffuse, unsigned int NrOfLobes, const Color* pX, const Color* pY, const Color* pZ, const Color* pExp, CBxDFBase::BxDFType Type) :
              CBxDFBase( Type ), m_NrOfLobes( NrOfLobes ), m_Diffuse( Diffuse ), m_pX( pX ), m_pY( pY ), m_pZ( pZ ), m_pExp( pExp )
{
}

CBxDFLaFortune::~CBxDFLaFortune(void)
{
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBxDFLaFortune::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( LightColor ); UNUSED( Power ); UNUSED( Shade ); UNUSED( RenderMat );

  double          Rho = 0.0;
  Vector          InDir(  In.dir() );
  Vector          OutDir(-Out.dir() );
  Color           TmpCol;
  Color           RhoCol = m_Diffuse / PI;

	double Dot = Normal.dot( InDir );
	if(Dot < SMALL3D )
		return Color::Black;

	for( unsigned int i = 0; i < m_NrOfLobes; ++i) {
		// Add i_th Phong lobe
		Color CurLobe = m_pX[i] * ( OutDir.x * InDir.x) + m_pY[i] * ( OutDir.y * InDir.y) + m_pZ[i] * ( OutDir.z * InDir.z );

		TmpCol(0) = pow( CurLobe.red(),   m_pExp[i].red() );
		TmpCol(1) = pow( CurLobe.green(), m_pExp[i].green() );
		TmpCol(2) = pow( CurLobe.blue(),  m_pExp[i].blue() );

    RhoCol += TmpCol.spectralToRGB();
  }
  /*RhoCol = TmpCol.spectralToRGB();*/

  Rho = std::max(0.0, std::min( 1.0, RhoCol.luminance()));

	return LightColor * Rho;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBxDFLaFortune::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0 */ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Power ); UNUSED( Type );


	Vector InDir, OutDir, TmpVec;
  Matrix3 uvwBase;
	Pdf = 1.0;

	OutDir = -Out.dir();

	// Cosine-sample the hemisphere, flipping the direction if necessary
	// set up orthogonal sampling basis

	unsigned int CurLobe = (unsigned int)(pRandNr[0] * INT_MAX) % (m_NrOfLobes);
	if ( CurLobe == m_NrOfLobes - 1 ) {
		InDir = CBaseSampler::cosineSampleHemisphere( Normal, pRandNr[0], pRandNr[1] );
	}
	else {
	 // Sample lobe CurLobe for Lafortune BRDF
		double CosTheta = pow( pRandNr[0], 1.0 / (0.8 * m_pExp[CurLobe].luminance() + 1) );
		double SinTheta = sqrt( std::max( 0.0, 1.0 - CosTheta * CosTheta ) );
		double Phi = pRandNr[1] * 2.0 * PI;

		// create an orthogonal basis of the lobe centered around vZ
     Vector vZ( m_pX[CurLobe].red()    * OutDir.x + m_pY[CurLobe].red()   * OutDir.y + m_pZ[CurLobe].red()   * OutDir.z,
                m_pX[CurLobe].green()  * OutDir.x + m_pY[CurLobe].green() * OutDir.y + m_pZ[CurLobe].green() * OutDir.z,
                m_pX[CurLobe].blue()   * OutDir.x + m_pY[CurLobe].blue()  * OutDir.y + m_pZ[CurLobe].blue()  * OutDir.z );

		//Vector vZ( Normal );

		vZ.normalize();

		uvwBase.make( vZ );
		uvwBase.transpose();

		InDir = SphericalDirection( SinTheta, CosTheta, Phi );
    InDir = uvwBase * InDir;
	}

	// check on which side of the hemisphere we are and flip if necessary
	if ( OutDir.dot( InDir ) < 0 ) {
		InDir *= -1;
	}


  //////////////////////////////////////////////////////////////////////////////
  // sample incoming ray
	In = Out;
	In.init();
	In.incDepth();  // to avoid infinite loops
	In.push( RaySegment( Out.hit( -SMALL3D4 ), InDir, Out.prevZone() ) );

	Pdf = GetPDF( OutDir, InDir, Normal);
	if( Pdf == 0 ) {
		return Color::Black;
	}

  return f( Out, In, Normal, RenderMat, LightColor, Shade);
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFLaFortune::GetPDF( const Vector& Out, const Vector& In, const Vector& Normal ) const

// ATTENTION: Out should be directed away from in!!!
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Normal );

	if( !CBaseSampler::sameHemisphere( Out, In))
		return 0.f;

	double PdfSum = fabs( In.z ) * PI;
	for( unsigned int i = 0; i < m_NrOfLobes; ++i ) {
		double XLum = m_pX[i].luminance();
		double YLum = m_pY[i].luminance();
		double ZLum = m_pZ[i].luminance();
		Vector LobeCenter( XLum * (Out.x), YLum * (Out.y), ZLum * (Out.z) );
		double Exp = 0.8 * m_pExp[i].luminance();

		PdfSum += ( Exp + 1.0 ) * pow( std::max( 0.0, In.dot( LobeCenter ) ), Exp );
	}

	return fabs(PdfSum / (1.0 + m_NrOfLobes));
}
// vim:ts=2
