////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M  - Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//          Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "bxdfphong.h"
#include "bxdffresnel.h"
#include "bxdfmicrofacet.h"

CBxDFMicrofacet::CBxDFMicrofacet( CFresnel* pFresnel, CMicrofacetDistribution* pMicroDist ) : CBxDFBase( static_cast<BxDFType>( BSDF_REFLECTION | BSDF_GLOSSY) ), m_pFresnel( pFresnel), m_pMicroDist( pMicroDist )
{
	assert( pFresnel && pMicroDist );
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////


Color CBxDFMicrofacet::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Shade ); UNUSED( Power ); UNUSED( LightColor );

	Vector OutDir  = Out.dir();
	Vector InDir   = In.dir();
	Vector HalfVec = -OutDir + InDir;    // ATTENTION: In & out dir have opposite directions. Thus, for a correct calculation of the half-vector the outgoing direction must be reversed.
                                       // Otherwise the result would be zero!
	HalfVec.normalize();

	double EtaI = 1.0;  // could get this from ray!
	double EtaT = RenderMat.ior();
	double cosThetaO = fabs( Normal.dot( -OutDir) );
	double cosThetaI = fabs( Normal.dot( InDir ) );
	double cosThetaH = fabs( Normal.dot( HalfVec) );

	double F = m_pFresnel->Evaluate(cosThetaH, EtaI, EtaT);
	double D = m_pMicroDist->D( HalfVec, Normal );
	double RetG = G( OutDir, InDir, HalfVec, Normal);

	return LightColor * (D * RetG * F / (4.0 * cosThetaI * cosThetaO));
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBxDFMicrofacet::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0 */ )

// Sample_f samples an incoming ray matching to the outgoing ray + its Pdf. Furthermore it evaluates the
// BxDF according to In & Out and returns the color. This is used for reflection, refraction etc. which will
// continue recursively.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Shade ); UNUSED( Power ); UNUSED( Type ); UNUSED( LightColor );

	Vector OutDir = -Out.dir();
	Vector InDir;

	// TODO: Replace RndFloat with stratified samples
	m_pMicroDist->Sample_f( OutDir, InDir, Normal, pRandNr[0], pRandNr[1], Pdf);

	if ( !CBaseSampler::sameHemisphere( OutDir, InDir ) )					// check if in & out are on the same hemisphere
		return Color::Black;

	// sample incoming ray
	In = Out;
	In.init();
	In.incDepth();	// to avoid infinite loops
	In.push( RaySegment( Out.hit( -SMALL3D4 ), InDir, Out.prevZone() ) );

	return f( Out, In, Normal, RenderMat, LightColor );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFMicrofacet::GetPDF( const Vector& Out, const Vector& In, const Vector& Normal ) const

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if( !CBaseSampler::sameHemisphere( Out, In) )
		return 0.0;

	return m_pMicroDist->GetPDF( Out, In, Normal);
}
// vim:ts=2
