////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "bxdfphong.h"
#include "bxdffresnel.h"
#include "math/bmath.h"

CBxDFPhong::CBxDFPhong( void ) : CBxDFBase( CBxDFBase::BSDF_ALL ), m_pFresnelObj( new CFresnelDiel )
{
}

CBxDFPhong::~CBxDFPhong( void )
{
	delete m_pFresnelObj;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBxDFPhong::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Out );
	UNUSED( LightColor );

	double ResColor = 1.0 * Shade;

	// add specular component
	if( RenderMat.shine() > 0.0 ) {
		double Dot = Out.dir() * Normal;
		Vector Reflect = (-2.0 * Dot) * Normal + Out.dir();
		double Spec = In.dir() * Reflect;

		if( Spec > 0.0 )
			ResColor += Power * pow( Spec, RenderMat.shine() );
	}

	return LightColor * ResColor;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBxDFPhong::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0 */)

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Power );

	if( Type == CBxDFBase::BSDF_REFLECTION )
		return Sample_f_Reflection( Out, In, Pdf, Normal, RenderMat, pRandNr, LightColor );
	else if ( Type == CBxDFBase::BSDF_TRANSMISSION ) {
		return Sample_f_Transmission( Out, In, Pdf, Normal, RenderMat, pRandNr, LightColor );
	}
	else
		return CBxDFBase::Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade );

	return Color::Black;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBxDFPhong::Sample_f_Reflection( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, const Color& LightColor )
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( pRandNr );

	Pdf = 1.0f;

	// CT: we only render reflections if we are at the beauty pass (step = 1)
	double dot = Normal * Out.dir();
	if( dot < 0.0 ) {
		Vector reflect = (-2.0 * dot) * Normal + Out.dir();
		reflect.normalize();

		// Find intersection with light
		In = Out;
		In.init();
		In.depth = Out.depth + 1;
		In.push( RaySegment( Out.hit( -SMALL3D4 * 10.0 ), reflect, Out.prevZone() ) );

		return f( Out, In, Normal, RenderMat, LightColor );
	}

	return Color::Black;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBxDFPhong::Sample_f_Transmission( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, const Color& LightColor )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
		/*
	/!\/!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\
	/!\
	/!\ For a physically correct refraction ray the code in BxDFSpecTransmission applying Snell's law should be used. However,
	/!\ in this case we don't care but we want pure "look-through" transparency and no distortions on curved surfaces!
	/!\
	/!\/!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\ /!\

	*/


	UNUSED( pRandNr );

	bool   bEntering = Out.type == Ray::RayType::Transmission ? false : true;
	double EtaI = 1.0;			// to be physically precise this IOR should be carried by the outgoing ray --> TODO:
	double EtaT = RenderMat.ior();

	Pdf = 1.0;

	double Cosi = Clamp( Normal.dot( Out.dir() ), -1.0, 1.0  );
	if ( bEntering ) {	
		Cosi = -Cosi;	} 
	else {	
		std::swap( EtaI, EtaT );	
	}

	double k = 1 - (EtaI / EtaT) * (EtaI / EtaT) * (1.0 - Cosi * Cosi);
	if ( k < 0 ) {
		//return Sample_f_Reflection( Out, In, Pdf, Normal, RenderMat, pRandNr );
		return Color::Black;	// NOTE: we intentionally disable physical total reflection to have clear transparency
	}

	Vector Temp = (EtaI / EtaT) * Out.dir() + ((EtaI / EtaT) * Cosi - sqrtf(k)) * Normal;
	Temp.normalize();


	// set up ray
	In = Out;
	In.init();
	In.depth = Out.depth + 1;

	RaySegment NewSegment( Out.hit( SMALL3D4 * 100 ), Temp, Out.hitZone() );
	NewSegment.tmin = SMALL3D3;
	NewSegment.tmax = INFINITE;

	In.push( NewSegment );
	In.type = bEntering ? Ray::RayType::Transmission : Ray::RayType::Standard;	// if the out ray was entering then this is a transmission ray (inside a solid), otherwise it is an exiting ray

	// Fresnel transmission,
	double Rho = 1.0 / fabs( Cosi ) * (EtaI * EtaI) / (EtaT * EtaT) * (1.0 - m_pFresnelObj->Evaluate( Cosi, EtaI, EtaT ));
	return LightColor * Rho;


}
// vim:ts=2
//

