////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
// Based on SimpleGeo's CDIB class from 2003, ported from my 3D particle simulation system (PSG) 1998
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef __DIB_H
#define __DIB_H

#include <cassert>
#include <string>

#include "math/bit.h"

typedef unsigned long		COLORREF;
typedef unsigned char		BYTE;
typedef unsigned short	WORD;
typedef unsigned long		DWORD;
typedef unsigned long		DWORD_PTR, * PDWORD_PTR;

#define MAKE_RGB(r,g,b) ((COLORREF)(((BYTE)(r)|((WORD)((BYTE)(g))<<8))|(((DWORD)(BYTE)(b))<<16)))

#define GetRValue(rgb)      (LOBYTE(rgb))
#define GetGValue(rgb)      (LOBYTE(((WORD)(rgb)) >> 8))
#define GetBValue(rgb)      (LOBYTE((rgb)>>16))

#pragma pack(2)
typedef struct
{
	unsigned short bfType;
	unsigned int   bfSize;
	unsigned short bfReserved1;
	unsigned short bfReserved2;
	unsigned int   bfOffBits;
} BITMAPFILEHEADER;

#pragma pack()

#define BF_TYPE 0x4D42             /* "MB" */

typedef struct                       /**** BMP file info structure ****/
{
	unsigned int   biSize;           /* Size of info header */
	int            biWidth;          /* Width of image */
	int            biHeight;         /* Height of image */
	unsigned short biPlanes;         /* Number of color planes */
	unsigned short biBitCount;       /* Number of bits per pixel */
	unsigned int   biCompression;    /* Type of compression to use */
	unsigned int   biSizeImage;      /* Size of image data */
	int            biXPelsPerMeter;  /* X pixels per meter */
	int            biYPelsPerMeter;  /* Y pixels per meter */
	unsigned int   biClrUsed;        /* Number of colors used */
	unsigned int   biClrImportant;   /* Number of important colors */
} BITMAPINFOHEADER, * PBITMAPINFOHEADER;

/*
 * Constants for the biCompression field...
 */

#define BI_RGB       0             /* No compression - straight BGR data */
#define BI_RLE8      1             /* 8-bit run-length compression */
#define BI_RLE4      2             /* 4-bit run-length compression */
#define BI_BITFIELDS 3             /* RGB bitmap with RGB masks */

typedef struct                       /**** Colormap entry structure ****/
{
	unsigned char  rgbBlue;          /* Blue value */
	unsigned char  rgbGreen;         /* Green value */
	unsigned char  rgbRed;           /* Red value */
	unsigned char  rgbReserved;      /* Reserved */
} RGBQUAD;

typedef struct tagBITMAPINFO {
	BITMAPINFOHEADER    bmiHeader;
	RGBQUAD             bmiColors[1];
} BITMAPINFO, * PBITMAPINFO;


class CDIB
{
public:
	enum BitmapType
	{
		BMP,
		GIF,
		TIFF
	};
	CDIB();
	virtual ~CDIB();
  CDIB( const CDIB& rhs ) { *this = rhs; };
	CDIB& operator=( const CDIB& dib);


	// DIB creation/writing
	bool Create(int Width,int Beight, int Bits = 24);
	bool Create(BITMAPINFOHEADER& bmInfo);
	void InitDIB( COLORREF color );
	bool CopyDIB(const CDIB& dib);
	bool OpenDIB( const std::string& fileName);
	bool SaveDIB( const std::string& fileName,BitmapType type);
	void DestroyDIB();
	bool IsValid() { return (m_pVoid && Width() && Height()); };

  void*		GetBits() const { return m_pBits; };
	DWORD		GetDIBSize();

	std::string	GetName() const { return m_Name; };

	// palette manipulations
	int				GetPaletteSize( BITMAPINFOHEADER& bmInfo );
	int				GetPaletteSize();
	int				CountColors();
	int				EnumColors( BYTE* colors );
	void			ReplaceColor( unsigned char oldColor, unsigned char newColor );
	void			SetPalette(unsigned char *palette);
	void			SetPalette(RGBQUAD *pRGB);
	COLORREF	PaletteColor(int index);
	int				ClosestColor( RGBQUAD* pRgb );
	static unsigned int Distance( RGBQUAD& rgb1, RGBQUAD& rgb2 );

	// pixel data manipulation
	void			CopyLine( int source, int dest );
	void			SetPixel(int x,int y,COLORREF color);
  void			SetPixel(int x,int y, unsigned char Red, unsigned char Green, unsigned char Blue);
	void			SetPixel8(int x,int y,unsigned char color);
	COLORREF	GetPixel(int x,int y);
	void			GetPixel(unsigned int x,unsigned int y,int& pixel);
	void			StretchBlt(int nXDest,int nYDest,int nDWidth,int nDHeight,CDIB& dibSrc,int xSrc,int ySrc,int  nSWidth,int nSHeight);
	void			ExpandBlt(int nXDest,int nYDest,int xRatio,int yRatio,CDIB& dibSrc,int xSrc,int ySrc,int  nSWidth,int nSHeight);
	void			BitBlt( int nXDest, int nYDest, int nWidth, int nHeight, CDIB& dibSrc, int nSrcX, int nSrcY, unsigned char* colors );
	unsigned char* GetLinePtr( int line );
	void			FlipHorizontally();
	void			FlipVertically();

	// image data conversion
	bool Make8Bit( CDIB& dib );
	bool SwitchFromOne( CDIB& dib );
	bool SwitchFromFour( CDIB& dib );
	bool SwitchFrom24( CDIB& dib );
	bool SwitchPalette( CDIB& dib );

	// flags & attributes
	void				SetFlags(int flag) { m_nFlags = flag; };
	int					Height() const { return height ; };
	int					Width() const { return width ; };
	inline int	GetBitCount() const {return m_pInfo->bmiHeader.biBitCount;};
	BITMAPINFO* GetBitmapInfo() const { return m_pInfo;};
	void				UseGamma( float fg, bool bUse = true );

protected:
	bool OpenBMP( const std::string& csFileName);
	bool OpenCompressedBMP( const std::string& csFileName );
	bool SaveBMP( const std::string& csFileName);

	bool ReadBMP( std::istream& input );


	void CreateGammaCurve();
	void Expand(int nXDest,int nYDest,int xRatio,int yRatio,CDIB& dibSrc,int xSrc,int ySrc,int  nSWidth,int nSHeight);

protected:
	unsigned char*	m_pBits;
	BITMAPINFO*			m_pInfo;
	RGBQUAD*				m_pRGB;
	void*						m_pVoid;
	BYTE**					m_pLinePtr;

	int							height,bytes,width,m_nBits;
	int							m_nFlags;
	bool						m_bUseGamma;
	float						m_fGamma;
	float						m_fOldGamma;
	unsigned char		Gamma[256];
	RGBQUAD					CacheQuad[256];
	char						CachePtr[256];
	std::string			m_Name;
};

#endif
// vim:ts=2
