////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#pragma once
#ifndef SURFACE_INTEGRATOR_H
#define SURFACE_INTEGRATOR_H

#include "baseintegrator.h"
#include "graph/color.h"
#include "matphong.h"
#include "lightmap.h"
#include "biasedlightmap.h"

#include <vector>

class D3Layer;

class CSurfaceIntegrator : public CBaseIntegrator
{
public:
	CSurfaceIntegrator();
	virtual ~CSurfaceIntegrator( void );

	// initialize options configured in the optional config object in FARM
	void InitFromConfig();

	// initialization before each render task (e.g. delayed loading/unloading of textures etc.)
	bool PreDrawInit() override;

	// main integrator routine returning the luminance for the passed ray & rendering pass
	virtual Color Li( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0, int CallerId = CBaseIntegrator::cStandard ) override;

	// configure ambient occlusion
	void SetAOCParameters( double MinIntensity, double MaxIntensity, double MaxDist, int NrOfSamples, int Multiplier )
	{
		m_AOCMinIntensity = MinIntensity;
		m_AOCMaxIntensity = MaxIntensity;
		m_AOCMaxDist      = MaxDist;
		m_NrOfAOCSamples  = NrOfSamples;
		m_AOCSampleMultiplier = Multiplier;
	}

	void SetAOCSamples(int NrOfSamples) { m_NrOfAOCSamples = NrOfSamples; }
	int  GetAOCSamples() const          { return m_NrOfAOCSamples; }

	// configure global illumination
	int	 GetNrOfGISamples() const { return m_NrOfGISamples; }
	void SetGISamples( int NrOfSamples ) { m_NrOfGISamples = NrOfSamples;  }

	// retrieve the light intensity at the hit point -> used to apply shadows for data that is displayed
	double	GetLightIntensityAtHit( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 ) {
		const Material* pMaterial = nullptr;
		if ( pRay->type != Ray::RayType::Transmission )
			pMaterial = pRay->hitRegion()->region()->material();
		else
			pMaterial = pRay->prevZone()->region()->region()->material();

		Color RetVal = ShadeHitPoint( pEngine, pRay, pMaterial, RenderPass, Light0 ); 
		return RetVal.luminance();
	};

protected:
	// setup bi-directional scattering & reflection LUT
	virtual void    SetupBxDFLUT();

	// ambient occlusion integrator
	double CalcAmbientOcclusion( GeometryEngine* eng, Ray* pixelRay, int nrOfSamples, const double cMaxDist, const double cMinIntensity, const double cMaxIntensity );

	// incoming reflection
	virtual Color Li_Reflection( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 );

	// refraction & transmission
	virtual Color Li_Transmission( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 );

	// environmental light maps
	virtual Color Li_LightMap( GeometryEngine* pEngine, Ray* pRay );

	// global illumination
	virtual Color Li_GlobalIllumination( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 );

	// calculate shading intensity of the current hit point
	Color					ShadeHitPoint( GeometryEngine* pEngine, Ray* pRay, const Material* pMaterial, const int RenderPass, int Light0 );

protected:
	std::vector<CPhysMatInterface*> m_PhysMatLUT;
	CPhysMatInterface* m_pPhysMat;

	double		m_AOCMinIntensity;
	double		m_AOCMaxIntensity;
	double		m_AOCMaxDist;
	int				m_NrOfAOCSamples;
	int				m_MaxAOCDepth;
	int				m_AOCSampleMultiplier;

	bool			m_bUseLightMap;

	CBiasedLightMap*	m_pLightMap;
	int								m_NrOfLightMapSamples;

	int				m_NrOfGISamples;
	bool			m_bUseGlobalIllumination;
	double    m_GIScaling;

	bool			m_bBumpmapData;
	double		m_BumpmapScaling;
	int				m_BumpDirection;

	double		m_MetalBumpScaling;
	double		m_PlasticBumpScaling;
	double		m_BrushedBumpScaling;
	double		m_ConcreteBumpScaling;

	bool			m_bTexAvailable;

	double		m_MaxGIDist;
};

#endif
// vim:ts=2
