////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "matbrushedmetal.h"


CMatBrushedMetal::CMatBrushedMetal( CBaseIntegrator* pInt, bool bNoiseTex /*= false*/, double BumpScaling /*= 1*/) : CPhysMatInterface(pInt, static_cast<CBxDFBase::BxDFType>(CBxDFBase::BSDF_DIFFUSE | CBxDFBase::BSDF_REFLECTION | CBxDFBase::BSDF_GLOSSY), true /* false = no MC sampling required */),
                                                                                                                       m_bUseNoise( bNoiseTex), m_BumpScaling( BumpScaling )
{
  static Color Diffuse( 0.60,    0.60,   0.65 );
  static float xy0[3] =     {  -1.11854f, -1.11845f, -1.11999f  };
  static float z0[3] =      {   1.01272f,  1.01469f,  1.01942f  };
  static float e0[3] =      {  15.8708f,  15.6489f,  15.4571f   };
  static float xy1[3] =     {  -1.05334f, -1.06409f, -1.08378f  };
  static float z1[3] =      {   0.69541f,  0.662178f, 0.626672f };
  static float e1[3] =      { 111.267f,   88.9222f,  65.2179f   };
  static float xy2[3] =     {  -1.01684f, -1.01635f, -1.01529f  };
  static float z2[3] =      {   1.00132f,  1.00112f,  1.00108f  };
  static float e2[3] =      { 180.181f,  184.152f,  195.773f    };

  static Color XY[3] = { Color( xy0[0], xy0[1], xy0[2]), Color( xy1[0], xy1[1], xy1[2]), Color( xy2[0], xy2[1], xy2[2]) };
  static Color  Z[3] = { Color( z0[0], z0[1], z0[2]), Color( z1[0], z1[1], z1[2]), Color( z2[0], z2[1], z2[2]) };
  static Color  E[3] = { Color( e0[0], e0[1], e0[2]), Color( e1[0], e1[1], e1[2]), Color( e2[0], e2[1], e2[2]) };


  m_pBxDFDiffuse = new CBxDFLaFortune( Diffuse, 3, XY, XY, Z, E, static_cast<CBxDFBase::BxDFType>( CBxDFBase::BSDF_DIFFUSE | CBxDFBase::BSDF_REFLECTION | CBxDFBase::BSDF_GLOSSY) );

	m_NoiseTexture.SetMinMax( 0.0, 1.0 );       // restrict noise
	m_NoiseTexture.m_Style = CNoiseTexture::cNormal;
}

CMatBrushedMetal::~CMatBrushedMetal(void)
{
  delete m_pBxDFDiffuse;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CMatBrushedMetal::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );

  auto Diffuse = m_pBxDFDiffuse->f( Out, In, Normal, RenderMat, LightColor, Shade );

  double NoiseScaling = Out.hitRegion()->region()->material()->roughness();
  double NoiseFact = m_bUseNoise ? m_NoiseTexture.Value( 0, 0, Out.hit(), NoiseScaling )[0] : 1.0;    // for the moment we only use the 1st channel of our noise, though we could use 3 of them if we want to influence

  //std::cout << "NoiseFact " << NoiseFact << std::endl;

  if( Normal.dot( In.dir() ) > SMALL3D )
    return Diffuse * NoiseFact;

  return Color::Black;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CMatBrushedMetal::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, CBxDFBase::BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0*/ )

// Sample_f samples an incoming ray matching to the outgoing ray + its Pdf. Furthermore it evaluates the
// BxDF according to In & Out and returns the color. This is used for reflection, refraction etc. which will
// continue recursively.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );

  if( Type == CBxDFBase::BSDF_REFLECTION || Type == CBxDFBase::BSDF_DIFFUSE )
    return m_pBxDFDiffuse->Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade);

  return Color::Black;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const Vector& CMatBrushedMetal::GetShadingNormal( const Ray* pCurRay )

//  [9/12/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  // check if we need to evaluate and update our cached shading normal based on Perlin noise bump-mapping

  if ( m_bUseNoise && ( !m_pCachedShadingRay || pCurRay != m_pCachedShadingRay ) ) {
    m_CachedShadingNormal = m_NoiseTexture.GetBumpNormal( pCurRay, m_BumpScaling );
    m_pCachedShadingRay = const_cast<Ray*>(pCurRay);
  }
  else if ( !m_bUseNoise ) {
    return pCurRay->normal;
  }

  return m_CachedShadingNormal;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CMatBrushedMetal::ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine )

//  [9/14/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  m_NoiseTexture.DisplaceHitRay( pCurRay, pEngine, m_BumpScaling );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// brushed metal with textures & displacement mapping
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const Vector& CMatBrushedMetalTex::GetShadingNormal( const Ray* pCurRay )

// Brushed metal with normal displacement texture maps
//  [10/26/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  return GetBumpNormalFromTex( pCurRay );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CMatBrushedMetalTex::ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine )

// brushed metal with normal displacement texture maps
//  [10/26/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  ModifyShadingGeometryByTex( pCurRay, pEngine );
}

// vim:ts=2
