////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef MAT_PLASTIC_H
#define MAT_PLASTIC_H

#include "physmatinterface.h"
#include "bxdfbase.h"
#include "baseintegrator.h"
#include "bxdffresnel.h"
#include "bxdfmicrofacet.h"
#include "bxdfphong.h"
#include "blinnmicrodist.h"
#include "bxdfspecrefl.h"
#include "bxdfspectransmission.h"
#include "bxdflambert.h"
#include "noisetexture.h"
#include "anisotropicmicrodist.h"

class CBxDFPhong;
class CBxDFMicrofacet;
class CFresnelPreDefDiel;
class CBlinnMicroDist;
class CBxDFSpecReflection;
class CBxDFSpecTransmission;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Standard polished plastic with Blinn's microfacet distribution
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatPlastic : public CPhysMatInterface
{
public:
  explicit CMatPlastic( CBaseIntegrator* pInt, const float cRoughness = 1.0, bool bNoiseTex = false);
  virtual ~CMatPlastic(void);

	// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
	// Out = ray coming FROM the camera, In = ray going towards the light
	virtual Color f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade = 1.0f ) override;

	// By default only diffuse sampling is performed. This function needs to be overwritten by the actual physical materials that
	// have reflective and probably translucent components!
	virtual Color Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, CBxDFBase::BxDFType Type, const Color& LightColor = Color( 1.0, 1.0, 1.0 ), double Power = 1.0, float Shade = 1.0f ) override;

protected:
	CBxDFBase*							m_pBxDFDiffuse;
  CBxDFMicrofacet*        m_pBxDFSpecular;
  CFresnelPreDefDiel*     m_pFresnelDiel;
  CBxDFSpecReflection*    m_pBxDFSpecReflection;
  CBxDFSpecTransmission*  m_pBxDFSpecTransmission;
	CBlinnMicroDist*				m_pBlinnMicroDist;

	bool											m_bUseNoise;
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Rough plastic with Trowbridge Reitz microfacet distributions
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatPlasticEx : public CMatPlastic
{
public:
	explicit CMatPlasticEx( CBaseIntegrator* pInt, CMicrofacetDistribution* pMicroDist, bool bNoiseTex = false, bool bDisplacement = false, double BumpScaling = 1.0 );
	virtual ~CMatPlasticEx() { delete m_pMicroFacets; }

	const Vector& GetShadingNormal( const Ray* pCurRay ) override;

	void ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine ) override;

protected:
	bool											m_bDisplacement;
	double										m_BumpScaling;

private:
	CMicrofacetDistribution* m_pMicroFacets;
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Rough plastic with Trowbridge Reitz microfacet distributions as they are more advanced and textures & displacement mapping
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatRoughPlasticTex : public CMatPlasticEx
{
public:
	explicit CMatRoughPlasticTex( CBaseIntegrator* pInt, double Roughness, const std::string& TexFileName, double TexDimU, double TexDimV, double BumpScaling = 1 );
	virtual ~CMatRoughPlasticTex() = default;

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	virtual const Vector& GetShadingNormal( const Ray* pCurRay ) override
	{
		return GetBumpNormalFromTex( pCurRay );
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	virtual void ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine ) override
	{
		ModifyShadingGeometryByTex( pCurRay, pEngine );
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// override color by blending the base color with the image texture
	Color GetHitColor( const Ray* pRay ) override
	{
		/*
		if ( !m_ImageTex.IsTextureLoaded() )
			return pRay->hitRegion()->color();

		auto TexColor = m_ImageTex.GetTexColor( pRay );
		Color BaseColor = pRay->hitRegion()->color();

		BaseColor += TexColor;
		BaseColor.clamp();
		return BaseColor; */

		return MixBaseWithTexColor( pRay, 0.9 );
	}
};

#endif
// vim:ts=2
