////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef MAT_SHINY_METAL_H
#define MAT_SHINY_METAL_H

#include "physmatinterface.h"
#include "bxdfbase.h"
#include "baseintegrator.h"
#include "bxdffresnel.h"
#include "bxdfmicrofacet.h"
#include "bxdfphong.h"
#include "blinnmicrodist.h"
#include "bxdfspecrefl.h"
#include "bxdfspectransmission.h"
#include "anisotropicmicrodist.h"
#include "noisetexture.h"


class CBxDFPhong;
class CFresnelPreDefConductor;
class CBxDFSpecReflection;
class CBxDFSpecTransmission;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Standard shiny metal with Blinn's microfacet distribution
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatShinyMetal : public CPhysMatInterface
{
public:
  explicit CMatShinyMetal( CBaseIntegrator* pInt, const float cRoughness = 1.0, const float cAbsorption = 0.0, bool bNoiseTex = false, bool bDisplacement = false, double BumpScaling = 1 );
  virtual ~CMatShinyMetal(void);

	// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
	// Out = ray coming FROM the camera, In = ray going towards the light
	virtual Color f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade = 1.0f ) override;

	// By default only diffuse sampling is performed. This function needs to be overwritten by the actual physical materials that
	// have reflective and probably translucent components!
	virtual Color Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, CBxDFBase::BxDFType Type, const Color& LightColor = Color( 1.0, 1.0, 1.0 ), double Power = 1.0, float Shade = 1.0f ) override;

	// override shading normal & geometry modififiers
	const Vector& GetShadingNormal( const Ray* pCurRay ) override;
	void ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine ) override;

protected:
  CBxDFPhong*               m_pBxDFDiffuse;      // microfacet diffuse reflector
  CBxDFSpecReflection*      m_pBxDFSpecular;     // specular reflection
  CFresnelPreDefConductor*  m_pFresnelReflection;
  CFresnelPreDefConductor*  m_pFresnelSpecular;
  CBxDFMicrofacet*          m_pDiffuseMicroFacets;

  float                     m_Absorption;

	CNoiseTexture							m_NoiseTexture;
	bool											m_bUseNoise;
	bool											m_bDisplacement;
	double										m_BumpScaling;

private:
	CBlinnMicroDist* m_pBlinnMicroDist;

};


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Shiny metal with the possibility to use Trowbridge Reitz microfacet distributions as they are more advanced
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatShinyMetalEx : public CMatShinyMetal
{
public:
	explicit CMatShinyMetalEx( CBaseIntegrator* pInt, const float cRoughnessX, const float cRoughnessY, const float cAbsorption = 0.0, bool bNoiseTex = false, bool bDisplacement = false, double BumpScaling = 1 );
  virtual ~CMatShinyMetalEx();

private:
  CTrowbridgeReitzDist*     m_pTrowbridgeReitz;

};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Shiny metal with the possibility to use Trowbridge Reitz microfacet distributions as they are more advanced
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class CMatPittedMetalTex : public CMatShinyMetalEx
{
public:
	explicit CMatPittedMetalTex( CBaseIntegrator* pInt, const std::string& TexFileName, double TexDimU, double TexDimV, double RoughnessX = 25, double RoughnessY = 50, double Absorption = 0.2, double BumpScaling = 1.0 ) : CMatShinyMetalEx( pInt, RoughnessX, RoughnessY, Absorption)
	{
		m_BumpScaling = BumpScaling;
		m_TexFileName = TexFileName;
		m_TexDimU = TexDimU;
		m_TexDimV = TexDimV;

		//InitTextures();
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	virtual const Vector& GetShadingNormal( const Ray* pCurRay ) override
	{
		return GetBumpNormalFromTex( pCurRay );
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	virtual void ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine ) override
	{
		ModifyShadingGeometryByTex( pCurRay, pEngine );
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// override color by blending the base color with the image texture
	Color GetHitColor( const Ray* pRay ) override
	{
		return MixBaseWithTexColor( pRay, 0.25 );
		//return MixBaseWithTexColor( pRay, 0 );
	}

};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Galvanized metal with textures and displacement mapping
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatGalvanizedMetalTex : public CMatPittedMetalTex
{
public:
	explicit CMatGalvanizedMetalTex( CBaseIntegrator* pInt, const std::string& TexFileName, double TexDimU, double TexDimV, double BumpScaling = 1.0 ) : CMatPittedMetalTex( pInt, TexFileName, TexDimU, TexDimV, 5, 5, 0.02, BumpScaling )
	{
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// override color by blending the base color with the image texture
	Color GetHitColor( const Ray* pRay ) override
	{
		return MixBaseWithTexColor( pRay, 0.2 );
	}

};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Painted metal with textures and displacement mapping
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class CMatPaintedMetal : public CMatPittedMetalTex
{
public:
	explicit CMatPaintedMetal( CBaseIntegrator* pInt, const std::string& TexFileName, double TexDimU, double TexDimV, double BumpScaling = 1.0 ) : CMatPittedMetalTex( pInt, TexFileName, TexDimU, TexDimV, 5, 5, 0.02, BumpScaling )
	{
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// override color by blending the base color with the image texture
	Color GetHitColor( const Ray* pRay ) override
	{
		auto Col1 = MixBaseWithTexColor( pRay, 0.10 );
		auto Col2 = AddBaseAndTexColor( pRay );

		return Col1 * Col2;
	}
};


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Scratched metal with textures and displacement mapping
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatScratchedMetalTex : public CMatPittedMetalTex
{
public:
	explicit CMatScratchedMetalTex( CBaseIntegrator* pInt, const std::string& TexFileName, double TexDimU, double TexDimV, double BumpScaling = 1.0 ) : CMatPittedMetalTex( pInt, TexFileName, TexDimU, TexDimV, 15, 36, 0.4, BumpScaling )
	{
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// override color by blending the base color with the image texture
	Color GetHitColor( const Ray* pRay ) override
	{
		//return MixBaseWithTexColor( pRay, 0.75 );
		return MultiplyBaseWithTexColor( pRay );
	}
};


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Rusty metal with textures and displacement mapping
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatRustyMetalTex : public CMatPittedMetalTex
{
public:
	explicit CMatRustyMetalTex( CBaseIntegrator* pInt, const std::string& TexFileName, double TexDimU, double TexDimV, double BumpScaling = 1.0 ) : CMatPittedMetalTex( pInt, TexFileName, TexDimU, TexDimV, 15, 36, 0.4, BumpScaling )
	{
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// override color by blending the base color with the image texture
	Color GetHitColor( const Ray* pRay ) override
	{
		return MixBaseWithTexColor( pRay, 0.15 );
	}
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Rusty metal with textures and displacement mapping
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatMatteMetalTex : public CMatPittedMetalTex
{
public:
	explicit CMatMatteMetalTex( CBaseIntegrator* pInt, const std::string& TexFileName, double TexDimU, double TexDimV, double BumpScaling = 1.0 ) : CMatPittedMetalTex( pInt, TexFileName, TexDimU, TexDimV, 15, 36, 0.4, BumpScaling )
	{
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// override color by blending the base color with the image texture
	Color GetHitColor( const Ray* pRay ) override
	{
		return MixBaseWithTexColor( pRay, 0.15 );
	}
};


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Rusty metal with textures and displacement mapping
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatBrushedSteelTex : public CMatPittedMetalTex
{
public:
	explicit CMatBrushedSteelTex( CBaseIntegrator* pInt, const std::string& TexFileName, double TexDimU, double TexDimV, double BumpScaling = 1.0 ) : CMatPittedMetalTex( pInt, TexFileName, TexDimU, TexDimV, 2, 2, 0.0, BumpScaling )
	{
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// override color by blending the base color with the image texture
	Color GetHitColor( const Ray* pRay ) override
	{
		return MixBaseWithTexColor( pRay, 0.05 );
	}
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Marble with textures and displacement mapping
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatMarble : public CMatShinyMetal
{
public:
	CMatMarble( CBaseIntegrator* pInt, const std::string& TexFileName, double TexDimU, double TexDimV ) : CMatShinyMetal( pInt, 1.0, 0.5 )
	{
		m_TexFileName = TexFileName;
		m_TexDimU = TexDimU;
		m_TexDimV = TexDimV;

		//InitTextures();
	};

	virtual ~CMatMarble() = default;


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	virtual const Vector& GetShadingNormal( const Ray* pCurRay ) override
	{
		return GetBumpNormalFromTex( pCurRay );
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	virtual void ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine ) override
	{
		ModifyShadingGeometryByTex( pCurRay, pEngine );
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// override color by blending the base color with the image texture
	Color GetHitColor( const Ray* pRay ) override
	{
		return MixBaseWithTexColor( pRay, 0.6 );
	}
};

#endif
// vim:ts=2
