////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#pragma once
#ifndef PHYS_MAT_INTERFACE_H
#define PHYS_MAT_INTERFACE_H

#include "bxdfbase.h"
#include "baseintegrator.h"
#include "graph/color.h"
#include "math/randsampler.h"
#include "noisetexture.h"
#include "texture.h"

//
// Interface class of all the physics models used for different materials. This class can act as a
// container of different BxDF's that make up the attributes of different materials.
//
class CPhysMatInterface
{
public:
	CPhysMatInterface( CBaseIntegrator* pIntegrator, CBxDFBase::BxDFType Type, bool bSamplingRequired ) : m_pIntegrator( pIntegrator ), m_Type( Type ), m_bMCSamplingRequired( bSamplingRequired ), m_pCachedShadingRay( nullptr ), m_BumpScaling( 1.0 ), m_bTexInitialized( false ),
		m_TexDimU( 0 ), m_TexDimV( 0 )
	{
		m_NormalTex.SetGeometry( pIntegrator->GetGeometryPtr() );
		m_ImageTex.SetGeometry( pIntegrator->GetGeometryPtr() );
	};

	virtual ~CPhysMatInterface( void ) {};

	bool MatchesType( CBxDFBase::BxDFType Flags ) const { return (m_Type & Flags); };
	bool RequiresMCSampling() const { return m_bMCSamplingRequired; };

	// default hit color = region color, but this can be overriden if texture are implemented in the BxDF
	virtual Color GetHitColor( const Ray* pRay ) { 
		if ( pRay->type != Ray::RayType::Transmission )
			return pRay->hitRegion()->color();

		assert( pRay->prevZone() );
		return pRay->prevZone()->region()->region()->material()->color();

	};

	// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
	// Out = ray coming FROM the camera, In = ray going towards the light
	virtual Color f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade = 1.0f ) = 0;

	// By default only diffuse sampling is performed. This function needs to be overwritten by the actual physical materials that
	// have reflective and probably translucent components!
	virtual Color Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, CBxDFBase::BxDFType Type, const Color& LightColor = Color( 1.0, 1.0, 1.0 ), double Power = 1.0,  float Shade = 1.0f )
	{
		UNUSED( Type );

		Color  SampleColor;
		Vector Intersection = Out.hit( -SMALL3D4 );
		Vector Dir, TmpVec;

		// sample incoming ray
		Dir = CBaseSampler::cosineSampleHemisphere( Normal, pRandNr[0], pRandNr[1] );

		In = Out;
		In.init();

		In.use_clip = Out.use_clip;
		In.use_project = Out.use_project;
		RaySegment new_segment( Intersection, Dir, Out.prevZone() );
		new_segment.tmin = SMALL3D3;
		new_segment.tmax = INFINITE;
		In.push( new_segment );

		In.incDepth();

		// determine pdf which needs to be normalized to the hemi-sphere
		// Pdf = PI;
		Pdf = 1.0 / M_PI;
		assert( Pdf > 0.0 );

		// default light color was set to white
		return f( Out, In, Normal, RenderMat, LightColor, Power, Shade );
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// initialize textures
	bool InitTextures()
	{
		if ( m_bTexInitialized )
			return true;

		m_bTexInitialized = false;
		m_bTexInitialized |= LoadNormalTexture( m_TexFileName, m_TexDimU, m_TexDimV );
		if ( !m_bTexInitialized )
			std::cout << "Warning: normal texture not found - " << m_TexFileName << std::endl;

		m_bTexInitialized |= LoadImageTexture( m_TexFileName, m_TexDimU, m_TexDimV );
		if ( !m_bTexInitialized )
			std::cout << "Warning: image texture not found - " << m_TexFileName << "_tex.bmp" << std::endl;

		//std::cout << "### loading - " << m_TexFileName << std::endl;

		return m_bTexInitialized;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// clear textures
	void ClearTextures()
	{
		m_bTexInitialized = false;

		if ( m_TexFileName == "" )
			return;

		if ( m_NormalTex.IsTextureLoaded() )
			m_NormalTex.UnloadTexture();

		if ( m_ImageTex.IsTextureLoaded() )
			m_ImageTex.UnloadTexture();

	}

	// textures
	bool		RequiresTexture() const { return m_TexFileName != ""; }
	bool		IsTextureLoaded() const { return m_bTexInitialized; };

	// return shading normal for this material & ray. default = geometric normal stored by the intersecting ray but
	// can be overridden by the respective material that might support bump maps etc.
	virtual const Vector& GetShadingNormal( const Ray* pCurRay ) { return pCurRay->normal; };

	// possibility to override the hit point etc. (e.g. for displacement mappings)
	virtual void ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine ) { UNUSED( pCurRay ); UNUSED( pEngine ); };

protected:
	bool LoadNormalTexture( const std::string& TexFileName, double TexDimU, double TexDimV )
	{
		m_NormalTex.SetTexDim( TexDimU, TexDimV );
		m_NormalTex.LoadTexture( TexFileName + std::string( ".bmp" ) );
		return m_NormalTex.IsTextureLoaded();
	}

	bool LoadImageTexture( const std::string& TexFileName, double TexDimU, double TexDimV )
	{
		m_ImageTex.SetTexDim( TexDimU, TexDimV );
		m_ImageTex.LoadTexture( TexFileName + std::string( "_tex.bmp" ) );
		return m_ImageTex.IsTextureLoaded();
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//
	Color MixBaseWithTexColor( const Ray* pRay, double BlendFactor )
	{
		if ( !m_ImageTex.IsTextureLoaded() )
			return pRay->hitRegion()->color();

		auto TexColor = m_ImageTex.GetTexColor( pRay );
		Color BaseColor = pRay->hitRegion()->color();

		BaseColor.blend( TexColor, BlendFactor );
		return BaseColor;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	Color MultiplyBaseWithTexColor( const Ray* pRay )
	{
		if ( !m_ImageTex.IsTextureLoaded() )
			return pRay->hitRegion()->color();

		auto TexColor = m_ImageTex.GetTexColor( pRay );
		Color BaseColor = pRay->hitRegion()->color();

		return BaseColor * TexColor;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	Color AddBaseAndTexColor( const Ray* pRay )
	{
		if ( !m_ImageTex.IsTextureLoaded() )
			return pRay->hitRegion()->color();

		auto TexColor = m_ImageTex.GetTexColor( pRay );
		Color BaseColor = pRay->hitRegion()->color();

		auto TmpCol = BaseColor + TexColor;
		TmpCol.clamp();

		return TmpCol;
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	const Vector& GetBumpNormalFromTex( const Ray* pCurRay, double Intensity = 1.0 )
	{
		if ( !m_NormalTex.IsTextureLoaded() )
			return pCurRay->normal;

		m_CachedShadingNormal = m_NormalTex.GetBumpNormal( pCurRay, m_BumpScaling * Intensity);
		return m_CachedShadingNormal;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	void ModifyShadingGeometryByTex( Ray* pCurRay, GeometryEngine* pEngine, double Intensity = 1.0 )
	{
		if ( m_NormalTex.IsTextureLoaded() ) {
			m_NormalTex.DisplaceHitRay( pCurRay, pEngine, m_BumpScaling * Intensity );
		}
	}


protected:
	CBaseIntegrator*					m_pIntegrator;
	const CBxDFBase::BxDFType m_Type;
	bool                      m_bMCSamplingRequired;
	Vector										m_CachedShadingNormal;
	Ray*											m_pCachedShadingRay;				// a get shading normal function that overrides the default can store here the ray ptr for which the cached shading normal has been determined already
																												// and thus, just needs to be returned without re-evaluation

	// textures
	CNoiseTexture							m_NoiseTexture;
	CGenericNormalTex					m_NormalTex;
	CImageTexture							m_ImageTex;
	double										m_BumpScaling;
	std::string								m_TexFileName;
	bool											m_bTexInitialized;
	double										m_TexDimU;
	double										m_TexDimV;
};


#endif
// vim:ts=2
