////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <cmath>
#include "texture.h"

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CImageTexture::CImageTexture( const std::string& FileName, double DimU /*= 1*/, double DimV /*= 1 */)

//  [10/20/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	m_TexDimU = DimU;
	m_TexDimV = DimV;
	LoadTexture( FileName );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CImageTexture::LoadTexture( const std::string& FileName )

//  [10/20/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if ( (m_pTexture = m_pGeometry->bitmapLoaded( FileName )) ) {
		return true;
	}

	m_pTexture = m_pGeometry->loadBitmap( FileName );

	if ( !m_pTexture ) {
		return false;
	}

	m_pTexture->FlipHorizontally();	// to respect the convention that the origin is the lower left corner, whereas in the file it is the upper left one
	return true;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CImageTexture::GetTexColor( const Ray* pCurRay )

//  [10/20/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double		u, v;
	COLORREF	PixelRGB;
	Color			TmpCol;

	// we abuse the roughness parameter to scale the texture. Higher roughnes = smaller texture;
	double Scale = 1.0 / pCurRay->hitRegion()->region()->material()->roughness();

	// set up texture mapping
	if ( !GetTextureCoordinates( pCurRay, u, v ) )
		return pCurRay->hitRegion()->region()->material()->color();

	// create normalized texture coordinates by assuming that we have toroidal topology
	double PixelsPerU = m_pTexture->Width() / (m_TexDimU * Scale);
	double PixelsPerV = m_pTexture->Height() / (m_TexDimV * Scale);

	int i = (int)(u * PixelsPerU) % m_pTexture->Width();
	int j = (int)(v * PixelsPerV) % m_pTexture->Height();

	if ( i < 0 )
		i += m_pTexture->Width();
	if ( j < 0 )
		j += m_pTexture->Height();

	PixelRGB = m_pTexture->GetPixel( i, j );

	// TODO: implement texture filtering
	// let's go for bilinear interpolation
	/*
	COLORREF	RightPixel, UpperPixel, DiagonalPixel;
	double		Weights[4];
	RightPixel		= m_TexFile.GetPixel( min( i + 1, m_TexFile.Width() - 1 ), j );
	UpperPixel		= m_TexFile.GetPixel( i, min( j + 1, m_TexFile.Height() - 1 ) );
	DiagonalPixel = m_TexFile.GetPixel( min( i + 1, m_TexFile.Width() - 1 ), min( j + 1, m_TexFile.Height() - 1 ) );

	double s = fmod( u * PixelsPerU, 1 );
	double t = fmod( v * PixelsPerV, 1 );

	Weights[0] = (min( i + 1, m_TexFile.Width() - 1 ) - s) * (min( j + 1, m_TexFile.Height() - 1 ) - t); // hit pixel
	Weights[1] = (min( i + 1, m_TexFile.Width() - 1 ) - s) * (t - j);	// right pixel
	Weights[2] = (s - i) * (min( j + 1, m_TexFile.Height() - 1 ) - t);	// upper pixel
	Weights[3] = (s - i) * (t - j); // diagonal pixel
	PixelRGB = PixelRGB * Weights[0] + RightPixel * Weights[1] + UpperPixel * Weights[2] + DiagonalPixel * Weights[3];
	*/

	TmpCol.set( GetRValue(PixelRGB) / 255.0, GetGValue(PixelRGB) / 255.0, GetBValue(PixelRGB) / 255.0);

	return TmpCol;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//  N O R M A L   D I S P L A C E M E N T   T E X T U R E


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Vector CNormalTex::GetNormalFromTex( const Ray* pCurRay, double u, double v )

// Get normal vector from texture via texture coordinates. The scale parameter
// ensures that higher roughness results in smaller image texture sizes, which makes
// the texture "rougher" or "noisier"
// [10/20/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Vector		Normal;
	COLORREF  PixelRGB;

	// we abuse the roughness parameter to scale the texture. Higher roughnes = smaller texture;
	double Scale = 1.0 / pCurRay->hitRegion()->region()->material()->roughness();

	// create normalized texture coordinates by assuming that we have toroidal topology
	double PixelsPerU = m_pTexture->Width() / (m_TexDimU * Scale);
	double PixelsPerV = m_pTexture->Height() / (m_TexDimV * Scale);

	int i = (int)(u * PixelsPerU) % m_pTexture->Width();
	int j = (int)(v * PixelsPerV) % m_pTexture->Height();

	if ( i < 0 )
		i += m_pTexture->Width();
	if ( j < 0 )
		j += m_pTexture->Height();


	PixelRGB = m_pTexture->GetPixel( i, j );
	// we map R = X [-1,1], G = Y [-1,1], B = Z [-1,1]
	Normal.set( -1.0 + 2.0 * (GetRValue( PixelRGB ) / 255.0), -1.0 + 2.0 * (GetGValue( PixelRGB ) / 255.0), -1.0 + 2.0 * (GetBValue( PixelRGB ) / 255.0) );
	Normal.normalize();

	return Normal;
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Vector CNormalTex::GetBumpNormal( const Ray* pCurRay, double Scale /*= 1.0*/ )

// return displacement normal for a specific hit position using spherical texture mapping
//  [10/20/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if ( !m_pTexture )
		return pCurRay->normal;

	Vector	NewNormal, OffsetNormal;
	Matrix3	uvwBase;
	double	u, v;

	// set up texture mapping
	if ( GetTextureCoordinates( pCurRay, u, v ) )
		return pCurRay->normal;

	uvwBase.make( pCurRay->normal );
	uvwBase.transpose();

	OffsetNormal = GetNormalFromTex( pCurRay, u, v );	// NOTE: Set param 3 to TexScaling in order to scale by roughness

	OffsetNormal = uvwBase * OffsetNormal;		// transform into coordinate system of the geometry normal

	NewNormal = pCurRay->normal + OffsetNormal * Scale;
	NewNormal.normalize();

	return NewNormal;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CNormalTex::DisplaceHitRay( Ray* pCurRay, GeometryEngine* pEngine, double BumpScaling )

// displace the hit point of the hit ray by the texture to implement displacement mapping
//
//  [1/9/2024 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if ( !m_pTexture )
		return;

//	// flag required cache update as we change the last ray direction
	pEngine->incBodyCheckId();

	Vector dF = GetBumpNormal( pCurRay, BumpScaling );

	// NOTE: This is not true displacement as this would either require triangulated mesh displacement of the geometry or
	//			 iterative re-tracking, which is too expensive. However, visually it gives a satisfying result if we are not too close to notice
	Vector CurHit = pCurRay->segment().hit( -SMALL3D4 );
	CurHit += dF * BumpScaling;		// displaced position

	Vector dir = CurHit - pCurRay->segment().pos;
	dir.normalize();
	pCurRay->segment().dir = dir;

}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Vector CNormalTex::GetTexValue( const Ray* pCurRay )

//  [10/20/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double	u, v;

	// set up texture mapping
	if ( !GetTextureCoordinates( pCurRay, u, v ) )
		return pCurRay->normal;

	return GetNormalFromTex( pCurRay, u, v );
}
// vim:ts=2
