/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "gregion.h"

#include <set>

#include "geometry.h"
#include "math/obbox.h"

using std::endl;

/* ================================ GREGION ================================= */
/** FLUKA region definition */
GRegion::GRegion(const std::string& aname) :
	_name(aname),
	_id(-1),
	_type(Type::NORMAL),
	_generation(0),
	_hash(0),
	_material(NULL),
	_hasMatrix(false),
	_universe(0),
	_fill(0),
	_lattice(nullptr),
	_latMatrixSize(0),
	_latMatrixId(nullptr),
	_latMatrix(nullptr),
	show(0)
{
} // GRegion

/** clone the region
 * WARNING not the expression!
 *         The expression is cloned in the @see Geometry::cloneRegion
 *         which requires to generate cloned bodies as well
 */
GRegion* GRegion::clone(const std::string& aname) const
{
	auto region = new GRegion(aname);
	// copy variables
	region->_type     = _type;
	region->_material = _material;
	region->_universe = _universe;
	region->_fill     = _fill;
	region->show      = show;
	if (_hasMatrix)
		region->matrix(_matrix);

	return region;
} // clone

/** initialize lattice:
 *   - set the limits
 *   - find the translation matrix
 *   - allocate memory
 * @param ilow, ihigh   low and high limit on i direction
 * @param jlow, jhigh   low and high limit on j axis
 * @param klow, khigh   low and high limit on k axis
 *
 * @return true on success, false on error
 */
bool GRegion::latticeInit(int il, int ih, int jl, int jh, int kl, int kh)
{
	ilow =  il;
	ihigh = ih;
	jlow =  jl;
	jhigh = jh;
	klow =  kl;
	khigh = kh;
	_fill =  0;

	if (ilow>ihigh) Swap(ilow, ihigh);
	if (jlow>jhigh) Swap(jlow, jhigh);
	if (klow>khigh) Swap(klow, khigh);

	// check if region fits to the lattice requirement
	if (not In(type(), Type::RECTLATTICE, Type::HEXLATTICE))
		throw std::string("Cell do not have the proper lattice type");

	if (size() != 1)
		throw std::string("Cell has more than one zone");

	if (_lattice) delete [] _lattice;
	_lattice = new unsigned short[latticeSize()];
	return _lattice != nullptr;
} // latticeInit

/** initialize transformation matrices for lattices
 * @param size	number of transformation matrices to allocate
 */
bool GRegion::latticeMatrixInit(const unsigned size)
{
	assert(_latMatrix==nullptr);
	assert(_latMatrixId==nullptr);
	assert(_latMatrixSize==0);

	_latMatrixSize = size;

	_latMatrixId = new unsigned short[latticeSize()];
	if (!_latMatrixId) return false;
	std::fill_n(_latMatrixId, latticeSize(), USHRT_MAX);

	_latMatrix   = new Matrix4[size];
	return _latMatrix != nullptr;
} // latticeMatrixInit

/** set transformation matrix which is used for lattices/universe redirection */
void GRegion::matrix(const Matrix4& M)
{
	if (!_hasMatrix) {
		// First time set the matrix
		_hasMatrix = true;
		_matrix = M;
	} else {
		// Multiply with the existing matrix
		// _matrix = M-1 * _matrix * M
		Matrix4 Minv;
		Minv.inverse(M);
		_matrix = Minv * (_matrix * M);
	}
	_matrix.fix();
	_invMatrix.inverse(_matrix);		// Invert matrix
	_invMatrix.fix();
} // matrix

/** transform ALL bodies of region with the specified matrix
 * this operation is done on the bodies that define the matrix
 */
void GRegion::transform(const Matrix4& M)
{
	// remember bodies that are already transformed
	std::set<GBody*> transformed;

	for (auto zone : _zones) {
		for (int ib=0; ib<zone->size(); ib++) {
			GBody* body = (*zone)[ib];
			// WARNING: it doesn't transform the GRegionBody
			//          this is the default behaviour in Phits, Moira
			//          HOWEVER is not the default in Penelope
			if (not body->isBody()) continue;
			if (transformed.find(body) != transformed.end()) continue;
			body->matrix(M);
			body->create();
			transformed.insert(body);
		}
	}
} // transform

/** @return memory used by region */
size_t GRegion::memory() const
{
	size_t mem = sizeof(GRegion);
	for (auto zone : _zones)
		mem += zone->memory();
	return mem;
} // memory

/** addZone */
GZone* GRegion::addZone(ZoneFormat aformat)
{
	GZone* zone = new GZone(this);

	zone->id(_zones.size());
	zone->format(aformat);

	_zones.push_back(zone);
	return zone;
} // addZone

/** addZone */
GZone* GRegion::addZone(GZone* z)
{
	GZone* zone = addZone(z->format());
	zone->reserve(z->size());
	for (int i=0; i<z->size(); i++)
		zone->add((*z)[i]);
	return zone;
} // addZone

/** insert a new zone at position pos */
GZone* GRegion::insertZone(int pos, GZone* z)
{
	GZone* zone = new GZone(this);
	_zones.insert(_zones.begin()+pos, zone);
	zone->id(pos);			// warning subsequent ids are wrong!
	zone->format(z->format());
	zone->reserve(z->size());
	for (int i=0; i<z->size(); i++)
		zone->add((*z)[i]);
	return zone;
} // insertZone

/** delZone */
void GRegion::delZone(int i)
{
	delete _zones[i];
	_zones.erase(_zones.begin()+i);
} // delZone

/** add2exp add token/body to last zone
 * @return true on success
 */
bool GRegion::add2exp(const char* token, GBody* body)
{
	assert(!_zones.empty());
	return _zones.back()->add(token, body);
} // add2exp

/** convert a region's zones in normal format
 * to multiple zones in either standard or rpn format
 */
void GRegion::convertFromNormal()
{
	for (int iz=0; iz<size(); iz++) {
		auto zone = _zones[iz];
		if (not zone->normal()) {
			zone->id(iz);
			continue;
		}
		_zones.erase(_zones.begin()+iz);	// remove from vector

		int iadd = 0;
		int from = 0;
		// skip leading |
		if (not zone->empty() and (*zone)[from]==&GBody::tunion) from++;
		for (int to=0; ; from=to+1) {
			to = zone->findZone(from);
			if (to<0) break;
			GZone subzone;
			zone->subZone(&subzone, from, to);
			insertZone(iz+iadd, &subzone);
			iadd++;
		}
		delete zone;	// to remove all references
	}
} // convertFromNormal

/** call setup on all zones */
void GRegion::setup()
{
	for (auto zone : _zones)
		zone->setup();

	// for a possible repeated lattice
	// find translation vectors
	if (In(type(), Type::RECTLATTICE, Type::HEXLATTICE)) {
		// Check the expression
		const GZone& z = *zone(0);
		if (!z.std()) {
			// FIXME return some kind of error!!!!!
			type(Type::NORMAL);
			return;
		}

		const Quad* qil = nullptr;
		const Quad* qih = nullptr;
		const Quad* qjl = nullptr;
		const Quad* qjh = nullptr;
		const Quad* qkl = nullptr;
		const Quad* qkh = nullptr;

		// Check the expression
		switch (z.size()) {
			// 1. +MB
			//    macro body BOX, RPP, HEX, RHP
			case 1:
				assert(_fill>0 or latticeiSize()>=1);
				assert(_fill>0 or latticejSize()>=1);
				assert(_fill>0 or latticekSize()>=1);
				if (z[0]->isBox()) {
					qil = &z[0]->Q(0);
					qjl = &z[0]->Q(1);
					qkl = &z[0]->Q(2);

					qih = &z[0]->Q(3);
					qjh = &z[0]->Q(4);
					qkh = &z[0]->Q(5);
				} else
					assert(false);
				break;

			// 3. +Pih[0] nil[1] -Pil[2]
			//    plane i high and low
			case 3:
				assert(_fill>0 or latticeiSize()>=1);
				assert(_fill>0 or latticejSize()==1);
				assert(_fill>0 or latticekSize()==1);
				qih = &z[0]->Q(0);
				assert(z[1]->isMinus());
				qil = &z[2]->Q(0);
				break;

			// 5. +Pih[0] +Pjh[1] nil[2] -Pil[3] -Pjl[4]
			//    plane i,j high and low
			case 5:
				assert(_fill>0 or latticeiSize()>=1);
				assert(_fill>0 or latticejSize()>=1);
				assert(_fill>0 or latticekSize()==1);
				qih = &z[0]->Q(0);
				qjh = &z[1]->Q(0);
				assert(z[2]->isMinus());
				qil = &z[3]->Q(0);
				qjl = &z[4]->Q(0);
				break;

			// 7. +Pih[0] +Pjh[1] +Pkh[2] nil[3] -Pil[4] -Pjl[5] -Pkl[6]
			//     plane i,j,k high and low
			case 7:
				assert(_fill>0 or latticeiSize()>=1);
				assert(_fill>0 or latticejSize()>=1);
				assert(_fill>0 or latticekSize()>=1);
				qih = &z[0]->Q(0);
				qjh = &z[1]->Q(0);
				qkh = &z[2]->Q(0);
				assert(z[3]->isMinus());
				qil = &z[4]->Q(0);
				qjl = &z[5]->Q(0);
				qkl = &z[6]->Q(0);
				break;

			default:
				// error!
				assert(false);
		}

		Point pC;

		assert(qih->isplane());
		assert(qil->isplane());
		ilat = qih->axis();
		assert(Eq(Abs(qil->axis()*ilat),1,SMALL));
		pC.x   = qil->C;			// C = - normal * position
		if (z.size()==1) pC.x = -pC.x;		// for macro bodies the 'low' points outside
		dlat.x = pC.x - qih->C;

		if (qjh != nullptr) {
			assert(qjh->isplane());
			assert(qjl->isplane());
			jlat = qjh->axis();
			assert(Eq(Abs(qjl->axis()*jlat),1,SMALL));
			pC.y   = qjl->C;		// C = - normal * position
			if (z.size()==1) pC.y = -pC.y;	// for macro bodies the 'low' points outside
			dlat.y = pC.y - qjh->C;
		} else {
			jlat   = ilat.orthogonal();
			pC.y   = 0;
			dlat.y = 0;
			assert(false);
		}

		if (qkh != nullptr) {
			assert(qkh->isplane());
			assert(qkl->isplane());
			klat = qkh->axis();
			assert(Eq(Abs(qkl->axis()*klat),1,SMALL));
			pC.z   = qkl->C;		// C = - normal * position
			if (z.size()==1) pC.z = -pC.z;	// for macro bodies the 'low' points outside
			dlat.z = pC.z - qkh->C;
		} else {
			// last vector orthogonal to others
			klat   = ilat ^ jlat;
			klat.normalize();
			pC.z   = 0;
			dlat.z = 0;
		}

		// Find location of lattice by solving the intersecting
		// system of the three planes
		// Solve system and find vertices,
		// add them to the bbox
		//    Cax*x + Cay*y + Caz*z + Ca = 0
		//    Cbx*x + Cby*y + Cbz*z + Cb = 0
		//    Ccx*x + Ccy*y + Ccz*z + Cc = 0
		//    A x X = B
		Matrix3 M;
		M.make(ilat, jlat, klat);
		if (M.inverse())
			plat = M.mult(-pC);
	}
} // setup

/** delete zone if it is a duplicate of another
 * @return true if the zone is removed
 */
bool GRegion::removeDup(int iz)
{
	GZone* zone = _zones[iz];
	if (zone->empty()) return false;
	for (auto izone : _zones) {
		if (zone == izone) continue;		// skip self
		if (zone->isSubsetOf(*izone)) {
			//std::cout << "+++ Duplicate Zone: " << *zone << std::endl;
			//std::cout << "              with: " << *izone << std::endl;
			zone->clear();
			zone->updateBBox();
			return true;
		}
	}
	return false;
} // removeDup

/** merge current zone defines by iz with other zones in the region
 * @param iz	index of zone to merge with others
 */
bool GRegion::merge(int iz)
{
	GZone* zone = _zones[iz];
	if (not zone->std() or zone->empty()) return false;
	bool changed = false;
	for (auto izone : _zones) {
		if (zone == izone) continue;		// skip self
		if (not izone->std()) continue;		// skip non STD zones
		if (zone->merge(*izone))
			changed = true;
	}
	return changed;
} // merge

/** @return true if region contains zone as expression of subset */
bool GRegion::contains(const GZone& zone)
{
	for (auto izone : _zones)
		if (zone.isSubsetOf(*izone))
			return true;
	return false;
} // contains

/** expand possible #regions present in the zones */
bool GRegion::expandRegions(Geometry* geometry)
{
	try {
		convertFromNormal();
	} catch (...) {
		return false;
	}
	for (auto zone : _zones)
		if (zone->format() == GZone::Format::RPN)
			zone->expandRegions(geometry);
	return true;
} // expandRegions

/** optimize zones */
bool GRegion::optimize()
{
	bool changed = false;
	//std::cout << "----- Before --------" << std::endl;
	//for (int j=0; j<size(); j++)
	//	std::cout << *(_zones[j]) << std::endl;
	bool mod1;
	do {	// loop until no more changes are done
		mod1 = false;
		for (int i=0; i<size(); i++) {
			GZone* zone = _zones[i];
			bool mod2;		// modified within the loop
			do {
				mod2 = false;
				if (not zone->bbox().isValid()) {
					changed = true;
					zone->clear();
					zone->setup();
					break;
				}
				// 1st try to geometrically optimize
				if (zone->std())
					mod2 = zone->optimize();
				else
				if (zone->rpn())
					mod2 = zone->optimizeProducts();

				// 2nd try to remove possible duplicate
				if (removeDup(i)) mod2 |= true;

				// 3rd try to merge with other zones
				if (merge(i)) mod2 |= true;
				mod1 |= mod2;
				changed |= mod2;
			} while (mod2 and not zone->empty());
			// printout
			//std::cout << "++++ Step " << i << std::endl;
			//for (int j=0; j<size(); j++)
			//	std::cout << *(_zones[j]) << std::endl;
		}
	} while (mod1 and size());
	return changed;
} // optimize

/** @return hash value of expression */
dword GRegion::hash() const
{
	if (_hash>0) return _hash;
	_hash = 0;
	int i=1;
	for (auto zone : _zones) {
		_hash += i * zone->hash();
		i += 2;
	}
	return _hash;
} // hash

/** clear */
void GRegion::clear(bool fast)
{
	// Clean the zones by popping the elements in order
	for (auto zone : _zones) {
		if (fast) zone->clear(fast);
		delete zone;
	}
	_zones.erase(_zones.begin(), _zones.end());
	_hash = 0;
	_latMatrixSize = 0;
	if (_lattice)     delete [] _lattice;
	if (_latMatrixId) delete [] _latMatrixId;
	if (_latMatrix)   delete [] _latMatrix;
	_lattice     = nullptr;
	_latMatrixId = nullptr;
	_latMatrix   = nullptr;
	//clearBB();
} // clear

/** update bbox of region zones */
void GRegion::updateBBox()
{
	for (auto zone : _zones)
		zone->updateBBox();
} // updateBbox

/** guess of 3D bounding box
 * @return bounding box of zone
 */
BBox GRegion::bbox() const
{
	BBox bb;

	for (auto zone : _zones)
		if (bb.isValid())
			bb |= zone->bbox();
		else
			bb = zone->bbox();
	return bb;
} // bbox

#ifdef USE_OBBOX
/** guess of 3D oriented bounding box
 * @return bounding box of zone
 */
OBBox GRegion::obbox() const
{
	OBBox obb;

	for (auto zone : _zones)
		if (obb.isValid())
			obb.Union(zone->obbox());
		else
			obb = zone->obbox();
	return obb;
} // obbox
#endif

/** check if point (x,y,z) with (u,v,w) direction is inside the zone or not
 * @param x,y,z    location of point to search
 * @param u,v,w    direction in case on boundary
 * @return zone    return zone that location is inside otherwise NULL
 *
 * @warning	NON-Caching function
 */
GZone* GRegion::inside(const double x, const double y, const double z,
		       const double u, const double v, const double w) const
{
	for (auto zone : _zones)
		if (zone->inside(x,y,z, u,v,w))
			return zone;
	return NULL;
} // inside

/** inside */
GZone* GRegion::inside(const Point& p, const Vector& d,
			const double t, Cache<CBody3D>& cache) const
{
	for (auto zone : _zones)
		if (zone->inside(p,d,t,cache))
			return zone;
	return nullptr;
} // inside

/** calculate track length in a region and the number of boundary crossings
 * given a position (pos) an direction (dir)
 * @param pos		initial position
 * @param dir		direction
 * @param tlength	track length inside region
 * @param cache		caching information of bodies
 * @return		number of boundary crossings
 */
unsigned GRegion::boundaryCrossings(const Point& pos, const Point& dir,
			double* tlength, Cache<CBody3D>& cache) const
{
	double tmin = 0;
	//double tmax = INFINITE; //*tlength;
	double tmax = *tlength;

	unsigned nbdx = 0;	// boundary crossings counter
	*tlength = 0;		// reset track length

	GZone* zone = inside(pos, dir, tmin, cache);
	assert(zone==nullptr); // we must always be outside!!!

	// assert check intersection with bounding box
	if (!bbox().distance(pos, dir, tmin, tmax))
		return nbdx;

	while (tmin<tmax) {
		if (zone==nullptr) {
			// find zone that we are entering
			zone = nullptr;
			double tmax2 = tmax;
			for (auto zzz : _zones) {
				double tmin2 = tmin;
				if (zzz->distance2In(pos,dir, &tmin2,tmax2, cache)) {
					zone  = zzz;
					tmax2 = tmin2;
				}
			}
			if (zone == nullptr) return nbdx;
			nbdx++;	// tally boundary crossing
			tmin = tmax2;
			assert(zone->inside(pos,dir,tmin,cache));
			DUMP(std::cout << "In: " << pos+dir*tmin << std::endl);
		}

		// we are inside zone
		*tlength -= tmin;		// tally negative enter distance

		// find distance to exit
		DEBUG(GBody* body =)
		zone->distance2Out(pos,dir, &tmin,tmax, cache);
		assert(body != nullptr);	// we should always exit

		*tlength += tmin;		// tally positive exit distance

		// find next zone at current location
		assert(!zone->inside(pos,dir,tmin,cache));
		zone = inside(pos,dir, tmin, cache);
		if (zone==nullptr) {
			DUMP(std::cout << "Out: " << pos+dir*tmin << std::endl);
			nbdx++;	// tally boundary crossing
		}
	}
	return 0;
} // boundaryCrossings

/** operator << */
std::ostream& operator << (std::ostream& s, const GRegion& region)
{
	s << region.name() << std::endl;

	for (int i=0; i<region.size(); i++)
		s << "    " << *region[i] << std::endl;
	return s;
} /* operator << */
