/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 *
 * - DXF export - 22.06.2010 from C. Theis
 */

#include "kernel.h"

// Body/Zone border type flags
const int SIDE_NONE   = 0x00;	// Body not found on any side (border)
const int SIDE_IN     = 0x01;	// Body found on side and is an overlap in the IN
const int SIDE_OUT    = 0x02;	// Body found on side and is an overlap in the OUT
const int SIDE_OK     = 0x04;	// Body found on side which is OK
const int SIDE_INOUT  = SIDE_IN|SIDE_OUT;
const int SIDE_BORDER = SIDE_IN|SIDE_OUT|SIDE_OK;
#define SIDE_STR(s)	(s&SIDE_IN? 'I':'-') << (s&SIDE_OUT?'O':'-') << (s&SIDE_OK? 'B':'-')

using std::cout;
using std::cerr;
using std::endl;

/* ============================== GeometryKernel ============================ */
/** GeometryKernel */
GeometryKernel::GeometryKernel(const Geometry& geo, int nt) :
	geometry(geo),
	_engine(&geo, this),
#ifdef THREAD
	threadengines(NULL),
#endif
	voxel(geo.voxel),
	view(640,480),
	_boundingVolTree(nullptr)
{
	_minSx2    = Sq(1e-5); // min of a pixel to be ignored
	_errors    = 0;
	_errorhash = 0;
	_bodyShow  = 0;
	universe   = 0;

	// thread
	if (nt>=0) initThreads(nt);
	pthread_mutex_init(&mutex, NULL);
} // GeometryKernel

/** ~GeometryKernel */
GeometryKernel::~GeometryKernel()
{
#ifdef THREAD
	for (int i=0; i<nthreads(); i++) {
		if (threadengines && threadengines[i].boundingVolTree())
			delete threadengines[i].boundingVolTree();
	}

	if (threadengines) delete[] threadengines;
#endif

	if (_boundingVolTree) {
		delete _boundingVolTree;
		_boundingVolTree = nullptr;
	}

	//stop();
	clearBodies();
	clearRegions();

	pthread_mutex_destroy(&mutex);

} /* ~GeometryKernel */

/** initialize threadpool
 * @param nt	number of threads, 0=all cpu/cores, 1=synchronous mode
 */
#ifdef THREAD
void GeometryKernel::initThreads(int nt)
#else
void GeometryKernel::initThreads(int)
#endif
{
#ifdef THREAD
	if (nt==0)	// 0 = use all available threads
		threadpool.init(ThreadPool::num_cores());
	else
	if (nt==1)	// 1 = run in sync mode, no threadpool
		threadpool.init(0);
	else	// n = up to max number of threads to use
		// limit to maximum number of cores
		threadpool.init(Min(nt, ThreadPool::num_cores()));
	if (threadengines) {
		delete [] threadengines;
		threadengines = NULL;
	}
#endif
} // initThreads

/** clearBodies */
void GeometryKernel::clearBodies()
{
	for (auto body: bodies)
		if (body) delete body;
	bodies.clear();
	clipBodyClear();
	projectBodyClear();
} // clearBodies

/** clearRegions */
void GeometryKernel::clearRegions()
{
	for (auto region : regions)
		if (region) delete region;
	regions.clear();
} // clearRegions

/** clear visible flag from all bodies */
void GeometryKernel::clearVisibleBodies()
{
	for (auto body : bodies)
		body->visible = false;
} // clearVisibleBodies

/** Calculate OR-ed all body->show flag */
// FIXME it should better be in Geometry
void GeometryKernel::updateBodyShow()
{
	//if (_bodyShow != BIT_INVALID) return;
	_bodyShow = 0;
	for (auto body : bodies)
		_bodyShow |= body->show();
} // updateBodyShow

/** projectBodyClear */
void GeometryKernel::projectBodyClear()
{
	_engine.projectBodyClear();
#ifdef THREAD
	if (threadengines!=NULL)
		for (int i=0; i<nthreads(); i++)
			engine(i)->projectBodyClear();
#endif
} // projectBodyClear

/** projectBodyAdd */
void GeometryKernel::projectBodyAdd(int idx)
{
	_engine.projectBodyAdd(idx);
#ifdef THREAD
	for (int i=0; i<nthreads(); i++)
		engine(i)->projectBodyAdd(idx);
#endif
} // projectBodyAdd

/** clipBodyClear */
void GeometryKernel::clipBodyClear()
{
	_engine.clipBodyClear();
#ifdef THREAD
	if (threadengines!=NULL)
		for (int i=0; i<nthreads(); i++)
			engine(i)->clipBodyClear();
#endif
} // clipBodyClear

/** clipBodyAdd */
void GeometryKernel::clipBodyAdd(int idx)
{
	_engine.clipBodyAdd(idx);
#ifdef THREAD
	if (threadengines!=NULL)
		for (int i=0; i<nthreads(); i++)
			engine(i)->clipBodyAdd(idx);
#endif
} // clipBodyAdd

/** clipBodyNegative */
void GeometryKernel::clipBodyNegative(int idx)
{
	_engine.clipBodyNegative(idx);
#ifdef THREAD
	if (threadengines!=NULL)
		for (int i=0; i<nthreads(); i++)
			engine(i)->clipBodyNegative(idx);
#endif
} // clipBodyNegative

// Convert the Bodies,Zones,Regions to Array(VBody,VZone,VRegion)
// do not delete anything but reassign
//
/** derive - derive the VBodies, VZones and VRegions from Geometry
 * It initializes the list of bodies, zones and regions in the Viewer,
 * note that it deletes all previous objects and creates new ones
 */
void GeometryKernel::derive()
{
#ifdef STAT
	Timer timer;
	timer.start();
#endif
	lock();

#if 0
	// WARNING: bodies and regions vectors can contain more or equal
	// number of VBodies vs geometry.bodies/regions
	//
	// Synchronize the bodies
	clearBodies();
	bodies.resize(geometry.nBodies());
	for (size_t i=0; i<geometry.nBodies(); i++) {
		GBody* gbody = geometry.bodies[i];
		bodies[i] = new VBody(gbody);
	}
#else
//#define DUMP(x) x
	DUMP(cout << endl);
	DUMP(cout << " ===================== DERIVE ======================" << endl);
	size_t oldSize = bodies.size();

	if (_boundingVolTree) {
		delete _boundingVolTree;
		_boundingVolTree = nullptr;
	}

	// Synchronize the bodies
	if (geometry.nBodies() < oldSize) {
		// Free the extra bodies if size if reduced
		for (size_t i=geometry.nBodies(); i<bodies.size(); i++)
			if (bodies[i]) {
				delete bodies[i];
				bodies[i] = NULL;
			}
		oldSize = geometry.nBodies();
	}
	bodies.resize(geometry.nBodies());
	for (size_t i=0; i<oldSize; i++) {
		// Update existing elements
		assert(bodies[i]!=NULL);
		GBody* gbody = geometry.bodies[i];
		VBody* vbody = bodies[i];
		DUMP(cout << "VBody " << gbody->name() << " eq=" << (vbody->body() == gbody) <<
			" invalid=" << vbody->invalid() << endl);
		if (vbody->body() != gbody || vbody->invalid()) {
			DUMP(cout << "VBody " << gbody->name() << " [UPDATE]" << endl);
			vbody->init(gbody);
		} else {
			DUMP(cout << "VBody " << gbody->name() << " [OK]" << endl);
		}
	}
	for (size_t i=oldSize; i<bodies.size(); i++) {
		assert(bodies[i]==NULL);
		GBody* gbody = geometry.bodies[i];
		DUMP(cout << "VBody " << gbody->name() << " [NEW]" << endl);
		bodies[i] = new VBody(gbody);
	}
#endif

#if 0
	clearRegions();
	regions.resize(nRegions());
	for (int i=0; i<nRegions(); i++) {
		GRegion* gregion = geometry.regions[i];
		regions[i] = new VRegion(gregion, this);
	}
#else
	// Synchronize regions
	oldSize = regions.size();
	if (geometry.nRegions() < oldSize) {
		// Free the extra bodies if size if reduced
		for (size_t i=geometry.nRegions(); i<regions.size(); i++)
			if (regions[i]) {
				delete regions[i];
				regions[i] = NULL;
			}
		oldSize = geometry.nRegions();
	}

	regions.resize(geometry.nRegions());
	for (size_t i=0; i<oldSize; i++) {
		GRegion* gregion = geometry.regions[i];
		VRegion* vregion = regions[i];
		DUMP(cout << "VRegion " << gregion->name()
			<< " eq=" << (vregion->region() == gregion)
			<< " hash_eq=" << (vregion->hash() == gregion->hash())
			<< " ggeneration=" << gregion->generation()
			//<< " vgeneration=" << vregion->generation()
			<< " invalid=" << vregion->invalid() << endl);
		if (vregion->region() != gregion ||
		    vregion->hash() != gregion->hash() ||
		    vregion->invalid()) {
			DUMP(cout << "VRegion " << gregion->name() << " [UPDATE]" << endl);
			vregion->init(gregion, this);
		} else {
			DUMP(cout << "VRegion " << gregion->name() << " [OK]" << endl);
		}
	}
	for (size_t i=oldSize; i<regions.size(); i++) {
		GRegion* gregion = geometry.regions[i];
		DUMP(cout << "VRegion " << gregion->name() << " [NEW]" << endl);
		regions[i] = new VRegion(gregion, this);
	}
	DUMP(cout << " ===================================================" << endl);
	DUMP(cout << endl);
#endif

	// derive engines
	_engine.derive();

	// build BVH
	buildBVH();
	_engine.boundingVolTree(_boundingVolTree);

	// pass original BVH to main engine and copies to the threaded engines
#ifdef THREAD
	if (threadengines==NULL && nthreads()>0) {
		threadengines = new GeometryEngine[nthreads()];
		for (int i=0; i<nthreads(); i++) {
			threadengines[i].geometry(&geometry);
			threadengines[i].kernel(this);
		}
	}
	for (int i=0; i<nthreads(); i++) {
		if (threadengines[i].boundingVolTree())
			delete threadengines[i].boundingVolTree();

		threadengines[i].derive();

		// attach a copy of the BVH to the thread-engine
		if (_boundingVolTree) {
			auto pLBVHCopy = new CLinearBoundingVolHierarchy(*static_cast<CLinearBoundingVolHierarchy*>(_boundingVolTree));
			//auto pLBVHCopy = new CSAHBoundingVolHierarchy( *static_cast<CSAHBoundingVolHierarchy*>(_boundingVolTree) );
			threadengines[i].boundingVolTree(pLBVHCopy);
		}
	}
#endif

	unlock();

#ifdef STAT
	timer.stop();
	cerr << "Time to derive geometry: " << timer << endl;
#endif
} // derive

/** origin */
void GeometryKernel::origin(const double x, const double y, const double z)
{
	stopThread();
	view.origin(x,y,z);
} // origin

/** moveViewOriginTo0
 * Move view origin to center of "view"
 */
void GeometryKernel::moveViewOriginTo0()
{
	stopThread();
	view.moveOriginTo0();
} // moveViewOriginTo0

/** matrix */
void GeometryKernel::matrix(const Matrix4& m)
{
	stopThread();		// Stop thread before modifying the matrix
	view.matrix(m);
} // matrix

/** window */
void GeometryKernel::window(double ex, double ey)
{
	if (ISZERO(ey)) ey = ex;
	view.window(-ex,-ey, ex,ey);
} // window

/** resize */
void GeometryKernel::resize(int w, int h)
{
	view.init(w, h);
} // resize

/* makeBodyConics
 * Creates the viewport conics for all the bodies
 * takes into account the _projectAll flag, otherwise it only affects invalid
 * bodies
 */
void GeometryKernel::makeBodyConics(bool all)
{
	_engine.incBodyCheckId();
	for (auto body : bodies) {
		if (all or body->invalid()) {
			IFSTAT(stat("invalid_bodies"));
			body->makeConics(view);
			body->intersectSelf(view);
			body->intersectViewport(view);
			body->updateLocation(view);
			//DUMP(cout << *body << endl);
		} else {
			IFSTAT(stat("valid_bodies"));
			// check if it intersects and invalid object
			// and clean its vertices
			body->removeInvalidVertices();
		}
	}
} // makeBodyConics

/** intersectBody */
void GeometryKernel::intersectBody(VBody* a, bool all)
{
#if 1
	// Generate all body combinations and intersect them
	for (size_t j=a->id()+1; j<bodies.size(); j++) {
		VBody* b = getBody(j);
		if (b->ignore()) continue;
		if (all || a->invalid() || b->invalid())
			a->intersectBody(b, view);
	}
#else
	// it goes slightly slower (I don't know why?)
	// Use a pseudo-random starting point
	int from = a->id()+1;
	int end   = _viewer->nBodies();
	if (end==from) return;

	// start from a random position
	int start = (from * 12997 + 981251) % (end-from) + from;

	// Generate all body combinations and intersect them
	// Split into two loops to avoid workers processing the same bodies
	for (int j=start; j<end; j++) {
		VBody* b = _viewer->getBody(j);
		if (b->ignore()) continue;
		if (all || a->invalid() || b->invalid())
			a->intersectBody(b, view);
	}

	for (int j=from; j<start; j++) {
		VBody* b = _viewer->getBody(j);
		if (b->ignore()) continue;
		if (_projectAll || a->invalid() || b->invalid())
			a->intersectBody(b, view);
	}
#endif
} // intersectBody

/** compare zone list a against zone list b if regions match
 * @return true if are different
 */
static inline bool equalZones(VZone* a[], VZone* b[], int n)
{
	for (int i=0; i<n; i++) {
		// check first the same position for faster lookup
		if (a[i]->region() == b[i]->region()) continue;

		bool found = false;
		for (int j=0; j<n; j++) {
			if (i == j) continue;
			if (a[i]->region() == b[j]->region()) {
				found = true;
				break;
			}
		}

		if (!found)
			return false;
	}
	return true;
} // equalZones

/** scanDirection fill the pIn, pOut structure with the search position/direction
 * @param body	body whose conic to scan
 * @param cid	conic id
 * @param u	u location of the middle segment
 * @param v	v location of the middle segment
 * @param pIn	zone of point with inward direction
 * @param pOut	zone of point with outward direction
 */
bool GeometryKernel::scanDirection(VBody* body, const int cid,
		const double u, const double v,
		ZoneOfPoint* pIn, ZoneOfPoint* pOut)
{
	// --------------------
	//   get the gradient
	// --------------------
	double tx = view.uv2x(u,v);
	double ty = view.uv2y(u,v);
	double tz = view.uv2z(u,v);

	double dx, dy, dz;
	body->Q(body->c2q[cid]).grad(tx,ty,tz,&dx,&dy,&dz);
	// normalize
	double dd = 1.0/sqrt(dx*dx+dy*dy+dz*dz);
	dx *= dd;
	dy *= dd;
	dz *= dd;

	/*
	 *             S\  ^ w
	 *               \ |   _-' normal
	 *                \|_-'
	 * ---view plane---+-------------------
	 *                / \`-_
	 *               /   \  `-_OUT
	 *              IN    \
	 * calculate the half angle from the surface
	 * with the view plane as IN and OUT
	 *
	 * IN = -normal-z
	 * AFTER the calculation of the shift
	 * adjust for the case of dz=0.0
	 */
	pIn->dx = -dx - view.matrix(0,2);
	pIn->dy = -dy - view.matrix(1,2);
	pIn->dz = -dz - view.matrix(2,2);

	//OUT = normal-z
	pOut->dx = dx - view.matrix(0,2);
	pOut->dy = dy - view.matrix(1,2);
	pOut->dz = dz - view.matrix(2,2);

	// Normalize directions
	double in2  = Sq(pIn->dx)  + Sq(pIn->dy)  + Sq(pIn->dz);
	double out2 = Sq(pOut->dx) + Sq(pOut->dy) + Sq(pOut->dz);
	// In case gradient is // to w-vector (perpendicular to screen)
	// then we have a problem. We are dealing with a touching surface
	// and we should normally ignore this point
	if (in2<1e-10) {
		assert(out2>1e-10);
		dd = 1.0/sqrt(out2);
		pOut->dx *= dd;
		pOut->dy *= dd;
		pOut->dz *= dd;
		pIn->dx = -pOut->dx;
		pIn->dy = -pOut->dy;
		pIn->dz = -pOut->dz;
		return false;
	} else
	if (out2<1e-10) {
		assert(in2>1e-10);
		dd = 1.0/sqrt(in2);
		pIn->dx *= dd;
		pIn->dy *= dd;
		pIn->dz *= dd;
		pOut->dx = -pIn->dx;
		pOut->dy = -pIn->dy;
		pOut->dz = -pIn->dz;
		return false;
	} else {
		dd = 1.0/sqrt(in2);
		pIn->dx *= dd;
		pIn->dy *= dd;
		pIn->dz *= dd;

		dd = 1.0/sqrt(out2);
		pOut->dx *= dd;
		pOut->dy *= dd;
		pOut->dz *= dd;
	}
#if 1
	// Move a bit in the direction of the search
	pIn->x  = tx + pIn->dx *SMALL*Abs(tx);
	pIn->y  = ty + pIn->dy *SMALL*Abs(ty);
	pIn->z  = tz + pIn->dz *SMALL*Abs(tz);
	pOut->x = tx + pOut->dx*SMALL*Abs(tx);
	pOut->y = ty + pOut->dy*SMALL*Abs(ty);
	pOut->z = tz + pOut->dz*SMALL*Abs(tz);
#else
	pIn->x  = tx;
	pIn->y  = ty;
	pIn->z  = tz;
	pOut->x = tx;
	pOut->y = ty;
	pOut->z = tz;
#endif
	return true;
} // scanDirection

/** scanBodySegments, a simple non-threaded version that scans
 * segments only of referenced bodies
 */
void GeometryKernel::scanBodySegments(bool all)
{
	clearErrorHash();
	for (auto body : bodies) {
		if (body->notReferenced()) continue;
		scanBodySegments(body, &_engine, all);
	}
} // scanBodySegments

/** scanBodySegments */
void GeometryKernel::scanBodySegments(VBody* body, GeometryEngine* eng, bool all)
{
	ZoneOfPoint pIn;
	ZoneOfPoint pOut;

#ifdef _DUMP
	if (body->show()) {
		cout << "********************************" << endl;
		cout << "+++ VBody-segments: " << body->name() << endl;
		for (int jj=0; jj < body->nC; jj++) {
			cout << "   Acc: " << body->acc[jj] << endl;
			for (int ii=0; ii<(int)body->V[jj].size(); ii++) {
				Segment& vv = body->V[jj][ii];
				cout << "   " << ii << " "
				     << (vv.body?vv.body->name():"NULL")
				     << '\t' << vv.t
				     << " [" << vv.x << ", " << vv.y << "]"
				     << " [" << view.uv2x(vv.x, vv.y)
				     << ", " << view.uv2y(vv.x, vv.y)
				     << ", " << view.uv2z(vv.x, vv.y) << "]"
				     << " invalid=" << vv.invalid
				     << endl;
			}
			cout << endl;
		}
		cout << "********************************" << endl;
	}
#endif
	// Check if any of the zones that refer to this body is visible
	// in the current viewport
	bool referenced = !body->notReferenced();

	for (int i=0; i<body->nC; i++) {
		//DUMP(cout << "+++ " << body->name() << " conic=" << i << endl);
		auto& V = body->V[i];
		bool isHyperbola = body->C[i].type()==Conic::Type::HYPERBOLA;
		bool isEllipse   = body->C[i].type()==Conic::Type::ELLIPSE;

		if (V.size() <= 1) continue;
		pIn.body  = body;
		pOut.body = body;
		pIn.V = pOut.V = &V;
		Segment* v = &V[0]; // Start point

		double tp = v->t;
		double xp = v->x;
		double yp = v->y;
		v->err = 0;
		v->type = Segment::Type::IGNORE;
		v->invalid = false;

		pIn.zero();
		pOut.zero();

		IFSTAT(stat.inc("segments_total", V.size()));

		// Segment invalid flag
		bool invalid = all || body->invalid();

		eng->incBodyCheckId();

		// Go through all the segments and find their inside and
		// outside zones by checking at midpoint
		for (int j=1; j<(int)V.size(); j++) {
			v = &V[j]; // next endpoint
			double t = v->t;
			double x = v->x;
			double y = v->y;
#ifdef EXPERIMENTAL
			// Reset the checkId of all objects crossed
			if (v->body!=NULL && v->body!=body) {
				eng->getBody(v->body)->resetCheckId();
				// and for all touching ones
				for (auto b : v->body->body()->touching)
					eng->getBody(b)->resetCheckId();
			} else
				eng->incBodyCheckId();
#endif
			//DUMP(cout << "<*< "<<body->name()<<" conic=" << i<<" vertex="<<j
			//	<< "\tv->body: " << (v->body? v->body->name() :"<win>") << endl);
			//DUMP(cout << "-*- "<<body->name()<<" x=" <<x<<" y="<<y<<" t="<<t<<endl);

			// Check if there is a reference to invalid body
			if (v->invalid) invalid = true;

			// Skip singularities
			if (Sq((x-xp)*view.Sx())+Sq((y-yp)*view.Sy())<_minSx2 or t-tp<SMALL)
				// do not skip empty ellipse with t=PI and tp=-PI
				if (!isEllipse or t-tp<SMALL) {
					IFSTAT(stat("segments_skipped_1"));
					// set the previous type
					tp = t;
					xp = x;
					yp = y;
					v->err = 0;
					v->type = Segment::Type::IGNORE;
					v->invalid = false;

					// reset the current body for the next search
					// increase the id
					// if goes outside the window or
					// intersect it self
#if 0
					// Reset the checkId of all objects crossed
					if (v->body != NULL && v->body != &body)
						eng->getBody(v->body)->resetCheckId();
					else
						eng->incBodyCheckId();
#endif
					continue;
				}

			// Hyperbola needs special checking
			if (isHyperbola and
			   ((tp<-PIover2 and t>-PIover2) or
			    (tp< PIover2 and t> PIover2))) {
				IFSTAT(stat("segments_skipped_2"));
				// Hyperbola on joining branches from - to + or vice versa
				// ignore them
				tp = t;
				xp = x;
				yp = y;
				v->err = 0;
				v->type = Segment::Type::IGNORE;
				v->invalid = false;
#ifdef EXPERIMENTAL
				// Is like moving out of window
				eng->incBodyCheckId();
#endif
				pIn.init();
				pOut.init();
				continue;
			}

			// Check window location of the segment
			// e.g. an arc at the viewport boundary
			double tm = 0.5*(t+tp);
			double xm, ym;
			body->C[i].getXY(tm,&xm,&ym);
			if (!view.inside(xm,ym)) {
				IFSTAT(stat("segments_skipped_3"));
				tp = t;
				xp = x;
				yp = y;
				v->err = 0;
				v->type = Segment::Type::IGNORE;
				v->invalid = false;
#ifdef EXPERIMENTAL
				// outside window increase id
				eng->incBodyCheckId();
#endif
				pIn.init();
				pOut.init();
				continue;
			}

			// If segment is valid ignore it
			// FIXME needs checking sometimes creates problems
			if (!invalid) { // reset flag
				IFSTAT(stat("segments_invalid"));
				if (v->invalid) invalid = true;
				tp = t;
				xp = x;
				yp = y;
#ifdef EXPERIMENTAL
				eng->incBodyCheckId();
#endif
				pIn.init();
				pOut.init();
				continue;
			}

			// Find search direction
			if (!scanDirection(body, i, xm, ym, &pIn, &pOut)) {
				IFSTAT(stat("segments_touching"));
				if (v->invalid) invalid = true;
				tp = t;
				xp = x;
				yp = y;
#ifdef EXPERIMENTAL
				eng->incBodyCheckId();
#endif
				pIn.init();
				pOut.init();
				continue;
			}

			// --- Find inside point ---
			pIn.to = j;
			pIn.copy2Old();

			// XXX XXX we should cache the information on the bodies
			// and reset the checkId only on the bodies that are crossing the Segment
			//DUMP(cout <<"<+< "<<body->name()<<" vertex="<<j<<" xm="<< xm << " ym="<<ym<<endl);
			// check also against the body
#ifdef EXPERIMENTAL
			// reset check id for current body
			eng->getBody(body)->resetCheckId();
			// and for all touching ones
			for (auto b : body->body()->touching)
				eng->getBody(b)->resetCheckId();
#else
			eng->incBodyCheckId();
#endif
			bool bodyIn = eng->getBody(body)->inside2D(
					pIn.x,pIn.y,pIn.z, pIn.dx,pIn.dy,pIn.dz);

			// and inside the geometry
			if (referenced) eng->where2D(&pIn);
			pIn.from = pIn.to;
			v->zone  = pIn.n? pIn.zone[0] : NULL;

			//DUMP(cout << "<+< "<<body->name()<<" inside="<<bodyIn<<endl);

			// --- Find outside point ---
			pOut.to = j;
			pOut.copy2Old();

			//DUMP(cout << "<-< "<<body->name()<<" vertex="<<j<<" xm="<<xm<<" ym="<<ym<<endl);
			// check against the body
#ifdef EXPERIMENTAL
			// reset check id for current body
			eng->getBody(body)->resetCheckId();
			// and for all touching ones
			for (auto b : body->body()->touching)
				eng->getBody(b)->resetCheckId();
#else
			eng->incBodyCheckId();
#endif
			bool bodyOut = eng->getBody(body)->inside2D(
					pOut.x,pOut.y,pOut.z, pOut.dx,pOut.dy,pOut.dz);
			// and inside the geometry
			if (referenced) eng->where2D(&pOut);
			pOut.from = pOut.to;
#ifdef EXPERIMENTAL
			// Invalidate last crossed bodies
			if (v->body!=NULL && v->body!=body) {
				eng->getBody(v->body)->resetCheckId();
				// and for all touching ones
				for (auto b : v->body->body()->touching)
					eng->getBody(b)->resetCheckId();
			} else
				eng->incBodyCheckId();
			// Invalidate current body
			eng->getBody(body)->resetCheckId();
			// and for all touching ones
			for (auto b : body->body()->touching)
				eng->getBody(b)->resetCheckId();
#endif
#if 0 // #ifdef _DUMP
			// strong debugging....
			if (body->show() && bodyIn==bodyOut) {
				cout << endl << "BREAK: " << body->name() << " j=" << j << " i="
					<< i << " in=" << bodyIn << " out=" << bodyOut << endl;
//					body->QN[body->c2q[i]].grad(xm,ym,&dx,&dy,&dz);
//					dd = sqrt(dx*dx+dy*dy+dz*dz);
				cout << "\ttm="<< tm << " xm="<< xm << " ym=" << ym << endl;
//					cout << "\t   : dx="<< dx << " dy=" << dy << " dz=" << dz << " Abs()<=" << (Abs(dz)<SMALL) << endl;
//					cout << "QN[i] = " << body->QN[body->c2q[i]] << endl;
				cout << "C[i]  = " << body->C[i] << endl;
//					cout << "\td= " << dx << ", " << dy << ", " << dz << "\t|d|=" << dd << endl;
				cout << "\tacc[i]= " << body->acc[i] << "  *SMALL=" << body->acc[i]*SMALL << endl;
//					cout << "\tQ(xm,ym,0)=" << body->QN[body->c2q[i]](xm,ym,0.) << endl;
//					cout << "\taccQ(xm,ym,0)=" << body->QN[body->c2q[i]].acc(xm,ym,0.) << endl;
//					cout << "\tQin(xi,yi,zi) =" << body->QN[body->c2q[i]](pIn.x,pIn.y,pIn.z) << endl;
//					cout << "\tQout(xo,yo,zo)=" << body->QN[body->c2q[i]](pOut.x,pOut.y,pOut.z) << endl;
//					cout << "\taccQin(xi,yi,zi) =" << body->QN[body->c2q[i]].acc(pIn.x,pIn.y,pIn.z) << endl;
//					cout << "\taccQout(xo,yo,zo)=" << body->QN[body->c2q[i]].acc(pOut.x,pOut.y,pOut.z) << endl;
				cout << "\tIn : x="<< pIn.x << " y=" << pIn.y << " z=" << pIn.z << endl;
				cout << "\t     dx="<< pIn.dx << " dy=" << pIn.dy << " dz=" << pIn.dz << endl;
				cout << "\tOut: x="<< pOut.x << " y=" << pOut.y << " z=" << pOut.z << endl;
				cout << "\t     dx="<< pOut.dx << " dy=" << pOut.dy << " dz=" << pOut.dz << endl;
			}
#endif
			//DUMP(cout << "<-< "<<body->name()<<" outside="<<bodyOut<<endl);
#if 0
			if (v->body==NULL || &body==v->body) {
				pIn.zero();
				pOut.zero();
			}
#endif
#if 0
			if (body->show()) {
			cout << "*-* t=" << tm
				<< " name=" << (v->body?v->body->name():"-null-") << " In=" << endl;
			for (int ii=0; ii<pIn.n; ii++)
				cout << pIn.zone[ii]->region->name() <<":" << *(pIn.zone[ii]) << endl;
			cout << "    Out=" << endl;
			for (int ii=0; ii<pOut.n; ii++)
				cout << pOut.zone[ii]->region->name() <<":" << *(pOut.zone[ii]) << endl;
			}
#endif
			// Compare the regions...
			if (pIn.nRegions == pOut.nRegions) {
				if (pIn.nRegions == 1) {
					v->err = 0;
					if (pIn.zone[0]->region() != pOut.zone[0]->region())
						v->type = Segment::Type::REGION;
					else
					if (pIn.zone[0] != pOut.zone[0])
						v->type = Segment::Type::ZONE;
					else
						v->type = Segment::Type::EMPTY;
				} else
				if (pIn.n == pOut.n &&
				    equalZones(pIn.zone, pOut.zone, pIn.n)) {
					// it is an error but the same on
					// both sides, ignore to show only the
					// borders below
					v->err  = 0;
					v->type = Segment::Type::EMPTY;
				} else {
					if (pIn.n  != pIn.nOld  ||
					    pOut.n != pOut.nOld ||
					    !equalZones(pIn.zone,  pIn.zoneOld,  pIn.n) ||
					    !equalZones(pOut.zone, pOut.zoneOld, pOut.n))
// FIXME _errors and errorhash should be protected with semaphores!!!!
						_errors++;
					_errorhash += (body->id()+1)*(i+1)*(j+1);
					v->err  = _errors;
					v->type = Segment::Type::ERROR;
				}
#if 0
			} else
			// Compare the zones...
			if (pIn.n == pOut.n) {
				if (pIn.n == 1) {
					v->err = 0;
					if (pIn.zone[0]->region() != pOut.zone[0]->region())
						v->type = Segment::Type::REGION;
					else
					if (pIn.zone[0] != pOut.zone[0])
						v->type = Segment::Type::ZONE;
					else
						v->type = Segment::Type::EMPTY;
				} else
				if (equalZones(pIn.zone, pOut.zone, pIn.n)) {
					// it is an error but the same on
					// both sides, ignore to show only the
					// borders below
					v->err  = 0;
					v->type = Segment::Type::EMPTY;
				} else {
					if (pIn.n  != pIn.nOld  ||
					    pOut.n != pOut.nOld ||
					    !equalZones(pIn.zone,  pIn.zoneOld,  pIn.n) ||
					    !equalZones(pOut.zone, pOut.zoneOld, pOut.n)) {
						//cout << "error 2" << endl;
						_errors++;
					}
					_errorhash += (body->id()+1)*(i+1)*(j+1);
					v->err  = _errors;
					v->type = Segment::Type::ERROR;
				}
#endif
			} else
			if (((pIn.n==1 and  pIn.zone[0]->region()->type()==GRegion::Type::BLACKHOLE) && pOut.n==0) or
			    ((pOut.n==1 and pOut.zone[0]->region()->type()==GRegion::Type::BLACKHOLE) && pIn.n==0) or
			    (geometry.mode()==Geometry::Mode::Penelope and pIn.n+pOut.n==1)) {
				// if one of the two is zero and the other is BLACKHOLE then ignore
				v->err  = 0;
				v->type = Segment::Type::REGION;
			} else {
				if (pIn.n  != pIn.nOld  ||
				    pOut.n != pOut.nOld ||
				    !equalZones(pIn.zone,  pIn.zoneOld,  pIn.n) ||
				    !equalZones(pOut.zone, pOut.zoneOld, pOut.n))
					_errors++;
				_errorhash += (body->id()+1)*(i+1)*(j+1);
				v->err  = _errors;
				v->type = Segment::Type::ERROR;
			}

			// On body boundary
			if (bodyIn != bodyOut) v->type |= Segment::Type::BODY;
#if 0
			// FIXME doesn't work properly
			// Check&Skip next vertex if pointing on notReferenced body
			for (int j1=j+1; j1<V.size(); j1++) {
				Segment* vn = V[j1];
				if (vn->body && body!=vn->body && vn->body.notReferenced()) {
					// Copy information from previous segment
					t = vn->t;
					x = vn->x;
					y = vn->y;
					vn->zone  = v->zone;
					vn->err  = v->err;
					vn->type = v->type;
					vn->invalid = false;
					j = j1;
				} else
					break;
			}
#endif
			// save position as previous of the next step
			tp = t;
			xp = x;
			yp = y;
			// reset flag
			invalid = all || body->invalid();
			if (v->invalid) {
				invalid = true;
				v->invalid = false;
			}
		}
	}

	// Set valid flag
	body->setValid();
} // scanBodySegments

/** error
 * return error information on err #err, or closest to point (x,y) if err<0
 * @param err		error index (if <0 then search closest to point (x,y)
 * @param x,y		I/O point to check error, and return error position
 * @param pIn, pOut	region of point information
 * @param xmin, xmax	x min and max limits of the error
 * @param ymin, ymax	y min and max limits of the error
 * @return NULL if no error is found otherwise the body containing the error
 */
VBody* GeometryKernel::error(const int err,
			double* u, double* v,
			ZoneOfPoint* pIn, ZoneOfPoint* pOut,
			double* umin, double* umax,
			double* vmin, double* vmax)
{
	int i=0,j;
	VBody* body = NULL;
	std::vector<Segment>* V = NULL;
	Segment* vertex;
	int from=-1, to=-1;		// error range

	for (auto b : bodies) {
		for (i=0; i<b->nC; i++) {
			V = &(b->V[i]);
			if (V->size() < 2) continue;
			for (j=1; j<(int)V->size(); j++) {
				vertex = &(*V)[j];
				if (vertex->err == err) {
					if (from < 0) {
						from = j-1;
						*umin = *umax = (*V)[from].x;
						*vmin = *vmax = (*V)[from].y;
					}
					to = j;
					*umin = Min(*umin, vertex->x);
					*umax = Max(*umax, vertex->x);
					*vmin = Min(*vmin, vertex->y);
					*vmax = Max(*vmax, vertex->y);
				} else
				if (from>=0 && vertex->err!=err) {
					body = b;
					goto FOUND;
				}
			}
			if (from>=0) {
				body = b;
				goto FOUND;
			}
		}
	}
FOUND:
	if (from<0) return NULL;

	vertex = &(*V)[from];
	Segment* vn = &(*V)[to];
	double tm = 0.5 * (vertex->t + vn->t);
	body->C[i].getXY(tm, u, v);
	scanDirection(body,i,*u,*v, pIn, pOut);

	// Prepare for search
	pIn->from = pOut->from = from;
	pIn->to   = pOut->to   = to;
	pIn->nOld = pOut->nOld = 0;
	pIn->body = pOut->body = body;

	// Find inside point
	_engine.incBodyCheckId();
	_engine.where2D(pIn);

	// Find outside point
	_engine.incBodyCheckId();
	_engine.where2D(pOut);

	return body;
} // error

/* updateRegionLocation */
void GeometryKernel::updateRegionLocation(bool all)
{
	for (auto region : regions) {
		if (all || region->invalid()) {
			IFSTAT(stat("invalid_regions"));
			DUMP(if (!all) cout << "Found invalid region: " << region->name() << endl);
			region->updateLocation(view);
			region->setValid();
		}
		IFSTAT(else
			stat("valid_regions"));
	}
} // updateRegionLocation

/** correctOverlaps, remove overlaps with other regions from that zone
 * @param region	region to remove overlaps from
 * @param zoneid	zone from region to remove overlaps from (-1=ALL)
 */
bool GeometryKernel::correctOverlaps(VRegion* region, int zoneid)
{
	bool regionModified = false;
	int  side;

	DUMP(cout << endl << "<<< " << *region->region() << endl);

	// Remember initial number of zones
	int nzones = region->size();
	while (true) {
		GZone* zone2 = NULL;	// additional zone
		VBody* body = NULL;
		side = SIDE_NONE;

		for (int i=0; i<region->size(); i++) {
			if (zoneid>=0 && i<nzones && i!=zoneid) continue;
			VZone* zone = region->zones()[i];
			DUMP(cout << endl << "Zone: " << *zone->zone() << endl);
			// FIXME don't check for bodies already been checked
			// keep id and start from that
			body = zoneOverlaps(zone, &side);
			DUMP(cout << "    SIDE= " << SIDE_STR(side) << endl);

			// Do we need both sides?
			if (body==NULL && (side&SIDE_INOUT)) {
				DUMP(cout << "   *********** We have to add another body..." << endl);
				// I should guess the remaining bodies from the zones
				// of the region that we have the problem
				side = SIDE_NONE;
			} else
			if ((side & SIDE_INOUT) == SIDE_INOUT) {
				DUMP(cout << "    ADD BOTH +/-" << body->name() << endl);
				zone2 = region->region()->addZone(zone->zone());
				zone->zone()->addPlus(body->body());
				zone->zone()->optimize();
				zone2->addMinus(body->body());
				zone2->optimize();
				zone2 = NULL;	// do not scan to remove it later
			} else
			// Only the inside
			if (side & SIDE_IN) {
				DUMP(cout << "    ADD +" << body->name() << endl);
				zone2 = region->region()->addZone(zone->zone());
				zone->zone()->addPlus(body->body());
				zone->zone()->optimize();
				zone2->addMinus(body->body());	// add the other side to test
				zone2->optimize();
			} else
			// Only the outside
			if (side & SIDE_OUT) {
				DUMP(cout << "    ADD -" << body->name() << endl);
				zone2 = region->region()->addZone(zone->zone());
				zone->zone()->addMinus(body->body());
				zone->zone()->optimize();
				zone2->addPlus(body->body());	// add the other side to test
				zone2->optimize();
			}

			// break
			if (side & SIDE_INOUT) break;
		}

		if (!(side & SIDE_INOUT)) break;
		regionModified = true;
		region->init(region->region(), this);
		derive();

		body->setInvalid();
		scanBodySegments(false);

		// check if we have to keep the temporary zone2
		if (zone2) {
			int side2;
			// Find vzone2
			VZone* vzone2=NULL;
			for (int i=0; i<region->size(); i++)
				if (region->zone(i)->zone() == zone2) {
					vzone2 = region->zone(i);
					break;
				}
			assert(vzone2!=NULL);
			DUMP(cout << "    Test: " << *zone2 << endl);
			VBody* body2 = zoneOverlaps(vzone2, &side2);
			DUMP(cout << "    SIDE= " << SIDE_STR(side2) << endl);
			if (body2==NULL || side2==SIDE_NONE) {
				DUMP(cout << "*** WRONG SIDE DELETE ***" << endl);
				// Delete last zone
				region->region()->delZone(region->size()-1);
				region->init(region->region(), this);
				derive();
				body->setInvalid();
				scanBodySegments(false);
				DUMP(cout << "*-* " << *region->region() << endl);
			} else {
				DUMP(cout << "*** OK KEEP ***" << endl);
			}
		}
	}

	DUMP(cout << ">>> " << *region->region() << endl);
	return regionModified;
} // correctOverlaps

/** isZoneValid *
bool GeometryKernel::isZoneValid(VZone* zone, VBody* body)
{
	return false;
} * isZoneValid */

/** zoneOverlaps
 * @param zone	zone to scan
 * @param side	return which side of body to to keep
 * @return body	that overlaps or appears on the border of the zone
 */
VBody* GeometryKernel::zoneOverlaps(VZone* zone, int* side)
{
	*side = SIDE_NONE;
	int side2 = SIDE_NONE;
	for (auto body : bodies) {
		if (body->notReferenced()) continue;
		//if (body->body()->hasZone(zone->zone())) continue;
		int s = bodySide(zone, body);
		DUMP(cout << "  Body: " << body->name() << "\tside= " << SIDE_STR(s) << endl);
		if (body->body()->hasZone(zone->zone()))
			side2 |= s;
		else {
			*side |= s;
			if (*side & SIDE_INOUT) return body;
		}
	}
	*side |= side2;
	return NULL;
} // zoneOverlaps

/** bodySide, remove overlaps with other regions from that zone
 * @param zone	zone to check for body if belongs to border
 * @param body	body to check
 * @return sides of body found
 */
int GeometryKernel::bodySide(VZone* zone, VBody* body)
{
	/* Scan all errors to find problematic bodies */
//	const int mask = Segment::Type::EMPTY | Segment::Type::REGION | GRegion::Type::ERROR;
	int side = SIDE_NONE;

	for (int i=0; i<body->nC; i++) {
		auto* V = &(body->V[i]);
		if (V->size() < 2) continue;
		for (int j=1; j<(int)V->size(); j++) {
			Segment* vertex = &(*V)[j];
			if (vertex->type == Segment::Type::IGNORE) continue;
			side |= segmentSide(zone, body, i, j-1, j);
		}
		// early stop
		if ((side&SIDE_BORDER) == SIDE_BORDER) return side;
	}
	return side;
} // bodySide

/** segmentSide, remove overlaps with this body
 * @param zone	zone to remove overlaps from
 * @param body	body might causing the overlap
 * @param i	conic of body
 * @param from	from vertex
 * @param to	to vertex
 * @return	SIDE_FLAGS
 */
int GeometryKernel::segmentSide(VZone* zone, VBody* body, int cid, int from, int to)
{
	ZoneOfPoint pIn;
	ZoneOfPoint pOut;

	Segment* vf = &(body->V[cid][from]);
	Segment* vt = &(body->V[cid][to]);
	double tm = 0.5 * (vf->t + vt->t);
	double u,v;
	body->C[cid].getXY(tm, &u, &v);

	scanDirection(body,cid,u,v, &pIn, &pOut);

	// Prepare for search
	pIn.from = pOut.from = from;
	pIn.to   = pOut.to   = to;
	pIn.nOld = pOut.nOld = 0;
	pIn.body = pOut.body = body;

	// Find inside point
	engine()->incBodyCheckId();
	engine()->where2D(&pIn);

	// Find outside point
	engine()->incBodyCheckId();
	engine()->where2D(&pOut);

	// To be an overlap:
	//   - zone should exist on both sides
	//   - one side should have only one zone (the current)
	//   - the other side should have more than one zones/regions and the current
	bool inZone  = pIn.hasZone(zone);
	bool outZone = pOut.hasZone(zone);

	// If zone is on neither side
	if (!inZone && !outZone) return SIDE_NONE;	// nothing to do with this zone

	// If zone is only on one side then...
	if (inZone && !outZone) {
		if (pIn.nRegions ==1)
			return SIDE_OK;		// valid for this zone
		else
			return SIDE_IN;		// overlap in IN
	}
	if (!inZone &&  outZone) {
		if (pOut.nRegions==1)
			return SIDE_OK;		// valid for this zone
		else
			return SIDE_OUT;	// overlap in OUT
	}

	// Both sides contain zone
	if (pIn.nRegions==1 && pOut.nRegions==1) return SIDE_NONE;	// valid for the region of the zone
	if (pIn.nRegions>1  && pOut.nRegions>1) return SIDE_NONE;	// overlap from both sides

	// Overlap: one side has more than one regions
	if (pIn.nRegions==1)
		return SIDE_IN;
	else
		return SIDE_OUT;
} // segmentSide

#ifdef STAT
/** resetStatistics */
void GeometryKernel::resetStatistics()
{
	stat.reset();
	engine()->stat.reset();
#ifdef THREAD
	if (threadengines) {
		for (int i=0; i<nthreads(); i++)
			engine(i)->stat.reset();
	}
#endif
} // resetStatistics

/** mergeStatistics */
void GeometryKernel::mergeStatistics()
{
#ifdef THREAD
	if (threadengines) {
		for (int i=0; i<nthreads(); i++)
			engine()->stat += engine(i)->stat;
	}
#endif
} // mergeStatistics
#endif

/** bodiesMemory */
size_t GeometryKernel::bodiesMemory() const
{
	size_t bmem = bodies.capacity() * sizeof(VBody*);
	for (auto body : bodies)
		bmem += body->memory();
	return bmem;
} // bodiesMemory

/** regionsMemory */
size_t GeometryKernel::regionsMemory() const
{
	size_t rmem = regions.capacity() * sizeof(VRegion*);
	for (auto region : regions)
		rmem += region->memory();
	return rmem;
} // regionsMemory

/** memory */
size_t GeometryKernel::memory() const
{
	size_t mem = sizeof(GeometryKernel)
	       + bodiesMemory()
	       + regionsMemory();
#ifdef THREAD
	mem += (nthreads()+1)*_engine.memory();
#else
	mem += _engine.memory();
#endif
	return mem;
} // memory

/** printMemory */
void GeometryKernel::printMemory() const
{
	cout << endl << "GeometryKernel:" << endl;
	cout << "Memory:"     << endl;
	cout << "\tSelf:\t"   << sizeof(GeometryKernel) << endl;
	cout << "\tBodies:\t" << bodiesMemory() << endl;
	cout << "\tRegion:\t" << regionsMemory() << endl;

#ifdef THREAD
	cout << "\tEngine:\t"	<< (nthreads()+1)
				<< " x " << _engine.memory()
				<< " = " << (nthreads()+1)*_engine.memory() << endl;
#else
	cout << "\tEngine:\t"	<< _engine.memory() << endl;
#endif

	cout << "\tTotal:\t"   << memory() << endl;
} // printMemory

/** Build bounding volume hierarchy */
void GeometryKernel::buildBVH()
{
#if defined(USE_BBOX_HIERARCHIES)
	// WIP: hierarchical bounding volumes for accelerated tracking
	// CT 4/26/2023

	// tighten potentially infinite bounding boxes
	tightenBBoxes();

	if (_boundingVolTree) {
		delete _boundingVolTree;
		_boundingVolTree = nullptr;
	}

	if (!_boundingVolTree) {
		// build zone list based on the engine array and pass it to BVH for construction
		// TODO: evaluate if BVH is converted from vector to Array in ctor

		// NOTE: we use the zonesSorted member of the engine because in this
		// case we are sure all connected cache bodies etc. have already been
		// initialized at this stage
		const std::vector<VZone*>& zoneList = _engine.getAllSortedZones();
		// CLinearBoundingVolHierarchy takes a copy of zoneList vector.
		_boundingVolTree = new CLinearBoundingVolHierarchy(zoneList, true, 1);
		//_boundingVolTree = new CSAHBoundingVolHierarchy( zoneList, true, true, 1 );
		// _boundingVolTree = new CBoundingVolHierarchy( zoneList, true, true, 1 );
		_boundingVolTree->BuildTree();
		_boundingVolTree->PrintTree();
	}
#endif
} // buildBVH

/** Try to eliminate infinite dimensions from the bounding boxes and tighten them
*   by setting them to the largest finite value discovered in the geometry
*/
void GeometryKernel::tightenBBoxes( double maxDim /*= FLOATBIG*/ )
{
	// maxDim extrema.
	const auto& minPos = Vector( -maxDim, -maxDim, -maxDim );
	const auto& maxPos = -minPos;

	std::cout << "Tightening bounding boxes..." << std::endl;

	// Find largest BBox with finite dimensions from the geometry.
	BBox maxBox;
	GRegion* pMaxBoundingRegion = nullptr;
	for ( const auto& iter : geometry.regionsMap ) {
		if ( iter.second->bbox().isValid() &&
				 ( (iter.second->bbox().lowx() >= minPos.x && iter.second->bbox().lowy() >= minPos.y && iter.second->bbox().lowz() >= minPos.z) ||
				 (iter.second->bbox().highx() <= minPos.x && iter.second->bbox().highy() <= minPos.y && iter.second->bbox().highz() <= minPos.z) ) &&
				 iter.second->bbox().size() >= maxBox.size() ) {
			maxBox = iter.second->bbox();
			pMaxBoundingRegion = iter.second;
		}
	}

	std::cout << "Max finite box " << pMaxBoundingRegion->name() << std::endl;
	std::cout << "LowPt [" << maxBox.low() << "] - HighPt [" << maxBox.high() << "]" << std::endl;

	// Tighten all zones' bboxes with that largest BBox.
	_engine.tightenAllZonesBBoxes(minPos, maxPos, maxBox);
} // tightenBBoxes
