/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	11-Oct-2016
 */

#ifndef __GEOMETRY_TETRAMESH_H
#define __GEOMETRY_TETRAMESH_H

#include "tetra.h"

#include "math/bbox.h"
#include "tool/list.h"

class Matrix4;

typedef List<Plane*>	PlaneList;
typedef List<Tetra*>	TetraList;

// ================================= TetraMesh =================================
class TetraMesh {
protected:
	VertexArray	_vertices;		/** array of vertices		*/
	PlaneArray	_planes;		/** array of planes		*/
	TetraArray	_tetrahedra;		/** array of tetrahedra		*/
	BBox		_bbox;			/** bounding box		*/
	double		eps;			/** accuracy of operations	*/

private:
	// temporary arrays used during build up
	std::vector<PlaneList> vertexPlanes;	/* planes for each vertex	*/
	std::vector<TetraList> vertexTetra;	/* tetra for each vertex	*/
	int		processed;

public:
	TetraMesh() : eps(FLOATSMALL), processed(0) {}
	~TetraMesh()	{ free(); };

	/** clean up mesh */
	void	free();
	void	freeVertices();
	void	freePlanes();
	void	freeTetrahedra();
//	void	resize(int nv, int ne, int nf);
	void	allocateVertices(int nv);
//	bool	isEmpty()		const	{ return nvertices() == 0; }

	// --- Vertex operations ---
	int	nvertices()		const	{ return _vertices.size(); }
	Vertex*	vertex(int i)			{ return _vertices[i]; }
const	Vertex*	vertex(int i)		const	{ return _vertices[i]; }
	Vertex*	add(const Vertex& v);

	// --- Planes operations ---
	int	nplanes()		const	{ return _planes.size(); }
	Plane*	plane(int i)			{ return _planes[i]; }
const	Plane*	plane(int i)		const	{ return _planes[i]; }
	Plane*	add(int a, int b, int c);

	// --- Tetra operations ---
	int	ntetrahedra()		const	{ return _tetrahedra.size(); }
	Tetra*	tetra(int i)			{ return _tetrahedra[i]; }
const	Tetra*	tetra(int i)		const	{ return _tetrahedra[i]; }
	Tetra*	add(Tetra* t)			{ _tetrahedra.push_back(t); return t; }
	Tetra*	add(int a, int b, int c, int d);
//	Tetra*	add(Edge* AB, Edge* BC, Edge* CA) { return add(new Tetra(AB, BC, CA)); }
//	Tetra*	add(Vertex* A, Vertex* B, Vertex* C, bool ab=true, bool bc=true, bool ca=true);

//	void	forEachVertex(VertexFunc, void* arg=NULL);
//	void	forEachEdge(EdgeFunc, void* arg=NULL);
//	void	forEachFace(FaceFunc, void* arg=NULL);
//	void	forEachTetra(TetraFunc, void* arg=NULL);

const	BBox&	bbox()			const	{ return _bbox; }
	void	calcBbox();

//	/* operations */
	double	surface() const;
	double	volume() const;

	void	process();
	void	transform(const Matrix4& matrix);

	size_t	memory() const;

	/** print tetramesh */
friend	std::ostream& operator << (std::ostream&, TetraMesh& tetramesh);
}; // TetraMesh

#endif
