/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	01-Apr-2011
 */

#include "vbody.h"

using std::cout;
using std::cerr;
using std::endl;

/* ============================== VBody ================================ */
/* Describe a convex body with up to 6 intersecting quads */

VBody::VBody(GBody* abody)
{
#ifdef THREAD
	pthread_mutex_init(&mutex, NULL);
#endif
	init(abody);
} // VBody

/** ~VBody */
VBody::~VBody()
{
#ifdef THREAD
	pthread_mutex_destroy(&mutex);
#endif
	delVertices();
	_body=NULL;
} /* ~VBody */

/** init */
void VBody::init(GBody* abody)
{
	_body       = abody;

	nC          = 0;
	visible     = false;
	notref      = false;
	_generation = -1;
	location    = GBody::Location::OVERLAP;	// default

	for (int i=0; i<GBody::BODYQUADS; i++) acc[i] = 1.0;
	delVertices();
} // init

/** delVertices */
void VBody::delVertices()
{
	for (int i=0; i<GBody::BODYCONICS; i++)
		V[i].clear();
} // delVertices

/* makeConics
 * Create viewport conics of the body and initializes location information
 * @param view	viewport
 */
void VBody::makeConics(const ViewPort& view)
{
	// clear settings
	delVertices();
	nC     = 0;
	notref = false;
	location = GBody::Location::OVERLAP;	// default

	if (type() == GBody::Type::ERR) {
		location = GBody::Location::OUTSIDE;
		return;
	}

	// Create transformed quadratics
	Quad QN[GBody::BODYQUADS];
	for (int i=0; i<nQ(); i++) {
		QN[i] = Q(i);
		QN[i].transform(view.matrix(), view.invMatrix());
		QN[i].normalize();
#if defined(_DUMP) && _DEBUG>1
		if (show()) {
			cout << "+++ VBody::makeConics:transform " << name() << ":" << i << endl;
			cout << "*-*\t Q[" << i << "] = " << Q(i) << endl;
			cout << "*-*\t QN["<< i << "] = " << QN[i] << endl;
			cout << "*-*\t M=" << endl << view.matrix() << endl;
		}
#endif
	}

	double x = view.Uofs();
	double y = view.Vofs();

	for (int i=0; i<nQ(); i++) acc[i] = 1.0;

	// Create conics
	for (int i=0; i<nQ(); i++) {
		const Quad& QNi = QN[i];
		Conic conic(QNi.Cxx, QNi.Cxy/2.0, QNi.Cyy, QNi.Cx/2.0, QNi.Cy/2.0, QNi.C);
		DUMP(if (show()) cout << "+++ " << QNi << endl);
		DUMP(if (show()) cout << "+++ " << conic << endl);

		if (conic.type()==Conic::Type::LINES) {
			Conic c1, c2;
			conic.splitLines(&c1, &c2);
			c1.parametric();
			c2.parametric();
			DUMP(if (show()) cout << "+1+ " << c1 << endl);
			DUMP(if (show()) cout << "+2+ " << c2 << endl);

			bool in1 = view.inside(c1);
			bool in2 = view.inside(c2);

			// if any of the two is inside then ok
			if (in1 || in2) {
				if (c1.equal(c2,SMALL)) {
					// Conics are the same, so we are tangential
					// to the quad(s) decide with the normal
					// if we are in or out
					location = GBody::Location::OUTSIDE;
					nC = 0;
					return;
				}
				if (in1) {
					c2q[nC] = i;
					C[nC++] = c1;
				}
				if (in2) {
					c2q[nC] = i;
					C[nC++] = c2;
				}
			} else {
				double q = QNi(x,y);
				double gx, gy, gz, g;
				QNi.grad(x, y, &gx, &gy, &gz);
				g = sqrt(gx*gx + gy*gy + gz*gz);
				if (g > SMALL) q /= g;
				DUMP(if (show()) cout << "\tq=" << q << endl);
				DUMP(if (show()) cout << "\tacc=" << QNi.acc(x,y,Conic::CONICPREC) << endl);
				if (q > QNi.acc(x,y,Conic::CONICPREC)) {
					location = GBody::Location::OUTSIDE;
					nC = 0;
					DUMP(if (show()) cout << "\tlocation="
							<< (int)location << " q=" << QNi(x,y) << endl);
					return;
				}
			}
		} else
		if (conic.type()==Conic::Type::LINE      || conic.type()==Conic::Type::ELLIPSE ||
		    conic.type()==Conic::Type::HYPERBOLA || conic.type()==Conic::Type::PARABOLA) {
			conic.parametric();
			if (view.inside(conic)) {
				c2q[nC] = i;
				C[nC++] = conic;
			} else {
				double q = QNi(x,y);
				double gx, gy, gz, g;
				QNi.grad(x, y, &gx, &gy, &gz);
				g = sqrt(gx*gx + gy*gy + gz*gz);
				if (g > SMALL) q /= g;
				DUMP(if (show()) cout << "\tq=" << q << endl);
				DUMP(if (show()) cout << "\tacc=" << QNi.acc(x,y,Conic::CONICPREC) << endl);
				if (q > QNi.acc(x,y,Conic::CONICPREC)) {
					location = GBody::Location::OUTSIDE;
					nC = 0;
					DUMP(if (show()) cout << "\tlocation="
							<< (int)location << " q=" << QNi(x,y) << endl);
					return;
				}
			}
		} else {
			DUMP(if (show()) cout << "\tq=" << QNi(x,y) << endl);
			double q = QNi(x,y);
			double gx, gy, gz, g;
			QNi.grad(x, y, &gx, &gy, &gz);
			g = sqrt(gx*gx + gy*gy + gz*gz);
			if (g > SMALL) q /= g;
			if (q > QNi.acc(x,y,Conic::CONICPREC)) {
				location = GBody::Location::OUTSIDE;
				nC = 0;
				DUMP(if (show()) cout << "\tlocation="
						<< (int)location << " q=" << QNi(x,y) << endl);
				return;
			}
		}
	}
} // makeConics

/** Intersect all the conics of a body
 * @param view   viewport
 */
void VBody::intersectSelf(const ViewPort& view)
{
	for (int i=0; i<nC; i++) {
		const Conic& aconic = C[i];
		if (aconic.type()==Conic::Type::ELLIPSE || aconic.type()==Conic::Type::HYPERBOLA) {
			// For ellipse add starting and ending point
			double x, y;
			aconic.getXY(PI, &x, &y);
			// XXX FIXME possible bug if it is outside the body
			// while at half distance there is another intersecting body!!!
			double rx = view.uv2x(x,y);
			double ry = view.uv2y(x,y);
			double rz = view.uv2z(x,y);
			if (view.inside(x,y) && inside(rx, ry, rz,
					-view.matrix(0,2), -view.matrix(1,2), -view.matrix(2,2),
					c2q[i]))
			{
				V[i].push_back(Segment(-PI, x, y, this));
#if _DEBUG>1
				if (Eq(-PI,-3.1415926535897931,1e-10) and
				   (Eq(x,40.293548333335508, 1e-10))) {
					printf("break2\n");
					aconic.getXY(-PI, &x, &y);
					aconic.getXY( PI, &x, &y);
				}
#endif
				V[i].push_back(Segment( PI, x, y, this));
			}
		}
		for (int j=i+1; j<nC; j++) {
			const Conic& bconic = C[j];
			Vector2D pts[4];
			int n = aconic.intersect(bconic, pts);
			for (int k=0; k<n; k++) {
				double x = pts[k].x;
				double y = pts[k].y;
				double rx = view.uv2x(x,y);
				double ry = view.uv2y(x,y);
				double rz = view.uv2z(x,y);
				if (view.inside(x,y) &&
					inside(rx,ry,rz,
						-view.matrix(0,2), -view.matrix(1,2), -view.matrix(2,2),
						c2q[i], c2q[j]))
				{
					double t = aconic.getT(x,y);
					V[i].push_back(Segment(t,x,y, this));
#if _DEBUG>1
					if (Eq(t,0,1e-10) and
					    Eq(x,7.74, 1e-10))
						printf("break3\n");
#endif
					t = bconic.getT(x,y);
					V[j].push_back(Segment(t,x,y, this));
#if _DEBUG>1
					if (Eq(t,0,1e-10) and
					    Eq(x,7.74, 1e-10))
						printf("break4\n");
#endif
				}
			}
		}
	}
} // intersectSelf

/* bodyIntersections
 * Intersect a body with itself + viewport limits
 * if there is no intersection it updates the location information
 * @param view	viewport
 */
void VBody::intersectViewport(const ViewPort& view)
{
	if (!nC) return;
	// Intersect with the window
	intersectConic(view.ulow,  view);
	intersectConic(view.uhigh, view);
	intersectConic(view.vlow,  view);
	intersectConic(view.vhigh, view);
} // intersectViewport

/* updateLocation
 * if a body has no vertices means it will be either inside or outside the
 * viewport, update this information
 */
void VBody::updateLocation(const ViewPort& view)
{
	// No conics and location already determined from makeConics
	if (!nC && location != GBody::Location::OVERLAP)
		return;

	// Check if body lies outside the plotting window
	for (int i=0; i<nC; i++)
		if (!V[i].empty()) return;

	// No intersection found, remove conics
	nC = 0;

	// find location of body, check several points on screen
	double dx = view.relWidth() / 4.0;
	double dy = view.relHeight() / 4.0;

	// 3 non-linear points should be sufficient
	// 1 or 2 might end-up on a degenerated conic of line or point
	// example a cylinder tangential to the plot will be only one line
	double x,y,z;
	view.origin(&x,&y,&z);

	assert(location == GBody::Location::OVERLAP);
	if (!inside(x, y, z, -view.matrix(0,2), -view.matrix(1,2), -view.matrix(2,2))) {
		location = GBody::Location::OUTSIDE;
		return;
	}

	double rdx = view.uv2x(dx,dy);
	double rdy = view.uv2y(dx,dy);
	double rdz = view.uv2z(dx,dy);

	if (!inside(rdx, rdy, rdz, -view.matrix(0,2), -view.matrix(1,2), -view.matrix(2,2))) {
		location = GBody::Location::OUTSIDE;
		return;
	}

	rdx = view.uv2x(dx,-dy);
	rdy = view.uv2y(dx,-dy);
	rdz = view.uv2z(dx,-dy);

	if (!inside(rdx, rdy, rdz, -view.matrix(0,2), -view.matrix(1,2), -view.matrix(2,2))) {
		location = GBody::Location::OUTSIDE;
		return;
	}

	// set to inside
	location = GBody::Location::INSIDE;

	DUMP(if (show()) cout << "\tlocation=" << (int)location << endl);
} // updateBodyLocation

/** Intersect the conics of two bodies and keep vertices inside the view
 * @param abody  first body
 * @param view   viewport
 */
bool VBody::intersectBody(VBody* abody, const ViewPort& view)
{
	assert(abody);
	assert(this != abody);

	bool intersected = false;

#if defined(_DUMP) && _DEBUG>1
	if (show() && abody->show()) {
		cout << "VBody::intersectBody" << endl;
		cout << "\tname=" << name()
		<< " abody->name=" << abody->name() << endl;
	}
#endif
	// Loop over all conics of this
	for (int i=0; i<nC; i++) {
		// with all conics of abody
		for (int j=0; j<abody->nC; j++) {
			Vector2D pts[4];
#if _DEBUG>1
			if (name()=="249" and abody->name()=="256")
				printf("break intersectBody(%s,%s)",cname(), abody->cname());
#endif
			int n = C[i].intersect(abody->C[j], pts);
			for (int k=0; k<n; k++) {
				double x  = pts[k].x;
				double y  = pts[k].y;
				double rx = view.uv2x(x,y);
				double ry = view.uv2y(x,y);
				double rz = view.uv2z(x,y);
				if (view.inside(x,y) &&
				    inside(rx,ry,rz,
					-view.matrix(0,2), -view.matrix(1,2), -view.matrix(2,2),
						c2q[i]) &&
				    abody->inside(rx,ry,rz,
					-view.matrix(0,2), -view.matrix(1,2), -view.matrix(2,2),
						abody->c2q[j]))
				{
					addVertex(i, x,y, abody);
					abody->addVertex(j, x,y, this);
					intersected = true;
				}
			}
		}
	}
	return intersected;
} // intersectBody

/** intersect body with a conic (normally from the viewport) */
void VBody::intersectConic(const Conic& conic, const ViewPort& view)
{
	Vector2D pts[4];

	for (int i=0; i<nC; i++) {
		int n = C[i].intersect(conic, pts);
		for (int j=0; j<n; j++) {
			double x  = pts[j].x;
			double y  = pts[j].y;
			double rx = view.uv2x(x,y);
			double ry = view.uv2y(x,y);
			double rz = view.uv2z(x,y);
			if (view.inside(x,y) &&
			    inside(rx,ry,rz,
				-view.matrix(0,2), -view.matrix(1,2), -view.matrix(2,2), c2q[i])) {
				double t = C[i].getT(x,y);
				V[i].push_back(Segment(t,x,y, NULL));
#if _DEBUG>1
				if (Eq(t,0,1e-10) and
				    Eq(x,7.74, 1e-10))
					printf("break5\n");
#endif
			}
		}
	}
} // intersectConic

/** removeInvalidVertices
 * remove vertices that belong to invalid bodies and invalidate
 * This is tricky because removing a vertex in a segment change the
 * region checks performed by scanBodySegments
 */
void VBody::removeInvalidVertices()
{
	assert(valid());
	for (int i=0; i<nC; i++) {
		auto& VV = V[i];

		// FIXME More correct but slower!!
		// Is not enough to remove only the vertex and invalidate
		// new segment.
		//       |  |     |  |  |  |  |       |  |
		// -0----1--X2----3--4--N--5--X6------7--N--8---
		//       |   \_______________/         |  |
		//
		// X is to be removed, N to be added later
		// If the new position is several segments away
		// like X(2) -> N, if we remove the X(2)
		// -0----1--x-----2--3--N--4--x-------5--N--6---
		//                       \______________/
		// we will mark the 1-2 as invalid
		// while the 2-3 should be as well (and all intermediates
		//
//#define REMOVE_ALL
#ifdef REMOVE_ALL
		bool invalid = false;
#endif
		for (int j=(int)VV.size(); j--; ) {
			Segment& vertex = VV[j];
			if (vertex.body && vertex.body->invalid()) {
				VV.erase(VV.begin()+j);
#ifdef REMOVE_ALL
				invalid = true;
#endif
				// Invalidate touching segments
				if (j<(int)VV.size())
					VV[j].invalid = true;
			}
		}
#ifdef REMOVE_ALL
		if (invalid)	// mark all vertices as invalid
			for (int j=VV.size(); j--;)
				VV[j].invalid = true;
#endif
	}
} // removeInvalidVertices

/** markInvalidVertices
 * mark segments touching invalid vertices as invalid
 * This is tricky because adding a new vertex in a segment does change the
 * region checks performed by scanBodySegments
 */
void VBody::markInvalidVertices(const VBody* invalidBody)
{
	//assert(valid());
	for (int i=0; i<nC; i++) {
		auto& VV = V[i];
		for (int j=(int)VV.size(); j--;) {
			const Segment& vertex = VV[j];
			if (vertex.body == invalidBody) {
				// Invalidate touching segments
				if (j<(int)VV.size())
					VV[j].invalid = true;
				if (j+1<(int)VV.size())
					VV[j+1].invalid = true;
			}
		}
	}
} // markInvalidVertices

/** remove problematic vertices from the vertex list
 * Segment should be in the order of intersection as
 * vertex-list:   [self|window ] ...(other bodies)... [self|window]
 * For hyperbola accept it twice
 *
 * XXX FIXME Clean up bodies or the reference to bodies when they are not visible
 *   e.g. nC>2 and all vertex are closer than a certain visible distance!!!
 *   or nC=2 like infinite cylinders or parallel bodies with the intersection to the windows
 *           closer than a certain distance
 * XXX Maybe this check can be done when the body is constructed, even to avoid checking any
 *     intersection with it since it wont be visible!
 */
void VBody::removeWrongVertices()
{
	// for all conics
	for (int i=0; i<nC; i++) {
		if (V[i].size()<2) continue;
		if (C[i].type()==Conic::Type::HYPERBOLA) continue;
		auto& VV = V[i];

		// Find first valid point, intersection with self or viewport
		int j, jp=0;
		double tp = -INFINITE;
		for (j=0; j<(int)VV.size(); j++) {
			Segment& v = VV[j];
			if (!Eq0(v.t-tp, SMALL*(Abs(v.t)+1.0))) {
				jp = j;
				tp = v.t;
			}
			if (v.body==NULL || v.body==this) break;
		}
		if (j==(int)VV.size()) {
			// delete everything!!!!
			VV.clear();
			continue;
		}

		// delete everything from 0 to jp
		if (jp>0) VV.erase(VV.begin(), VV.begin()+jp);

		// Do the same in reverse order
		tp = INFINITE;
		for (j=jp=(int)VV.size()-1; j>=0; j--) {
			Segment& v = VV[j];
			if (!Eq0(v.t-tp, SMALL*(Abs(v.t)+1.0))) {
				jp = j;
				tp = v.t;
			}
			if (v.body==NULL || v.body==this) break;
		}

		// delete everything from jp+1 to end
		VV.erase(VV.begin()+jp+1, VV.end());
	}
} // removeWrongVertices

/** sortVertices */
void VBody::sortVertices()
{
	for (int i=0; i<nC; i++)
		std::sort(V[i].begin(), V[i].end(), Segment::compare);
} // sortVertices

/**
 * Calculate the limits of the accuracy to be used for each quad
 * For the moment only report an error or warning if too uncertain
 */
void VBody::calculateAccuracy(const ViewPort& view)
{
	const double MAXACC = 100000.0;
	// Initial values
	for (int i=0; i<nQ(); i++)
		acc[i] = SMALL;

	// For each conic
	for (int i=0; i<nC; i++) {
		int q = c2q[i]; // Quad index
		auto& VV = V[i];
		if (VV.empty()) {
			acc[q] = 10.0;
			continue;
		}

		bool isHyperbola = C[i].type()==Conic::Type::HYPERBOLA;

		// Find limits
		double t0 = VV[0].t;
		double tp = t0;

		// calculate in the middle of the segment
		for (int j=1; j<(int)VV.size(); j++) {
			Segment const &v = VV[j];
			if (Eq0(v.t - tp, SMALL*Abs(v.t+SMALL)))
				continue;
			if (!(isHyperbola &&	 // Skip the segment between the two hyperbolas
			      ((tp<-PIover2 && v.t>-PIover2) ||
			       (tp< PIover2 && v.t> PIover2)))) {
				double x, y;
				C[i].getXY((v.t + tp)/2.0, &x, &y);
				double rx = view.uv2x(x,y) - SMALL*view.matrix(0,2);
				double ry = view.uv2y(x,y) - SMALL*view.matrix(1,2);
				double rz = view.uv2z(x,y) - SMALL*view.matrix(2,2);
				acc[q] = Max(acc[q], accuracy(q, rx, ry, rz));
#if _DEBUG>1
				if (accuracy(q, rx, ry, rz) > MAXACC) {
					cout << endl << "ERROR:" << endl;
					cout << "   body: " << name() << "  c=" << i << "  Q=" << q <<  endl;
					cout << "         " << *_body << endl;
					cout << "   quad[" << q << "]=" << Q(q) << endl;
					cout << "   conic[" << i << "]=" << C[i] << endl;
					cout << "   matrix=" << endl << view.matrix() << endl;
					cout << "   screen= " << x << " " << y << endl;
					cout << "   C()=    " << C[i](x,y) << endl;
					cout << "   pos   = " << rx << " " << ry << " " << rz << endl;
					cout << "   acc   = " << accuracy(q, rx, ry, rz) << endl;
					cout << "   Q()   = " << Q(q)(rx,ry,rz) << endl;
					cout << "   accQ()= " << Q(q).acc(rx,ry,rz, SMALL2) << endl;
				}
#endif
			}
		}
	}
	// Final adjustment
	for (int i=0; i<nQ(); i++) {
		// Allow some margin (x10.0) since all the test will
		// be done slightly outside the viewing plane
		if (acc[i] > MAXACC) {
			if (acc[i] > 10.0*MAXACC)
				cerr << "ERROR: ";
			else
				cerr << "Warning: ";
			cerr << "VBody " << name() << ":" << i << " accuracy=" << acc[i] << endl;
		}
		acc[i] = 64.0 * Range(1.0, acc[i], MAXACC);
#if defined(STRING) && defined(_DUMP)
		if (show()) cout << "Accuracy: " << name() << "." << i << " = " << acc[i] << endl;
#endif
	}
} // calculateAccuracy

#if 0
/**
 * Find if a point (x,y) is on the body surface
 * @param x,y	position
 * @param d	distance from the boundary
 * Returns: true  on the surface
 * XXX FIXME XXX problematic routine on RCC!
 * have to check if the point is inside the body on a non-displayed segment
 */
bool VBody::isOn(const double x, const double y, const double d) const
{
	bool on = false;
	for (int i=0; i<nQ(); i++) {
// XXX if Conic::splitLines is corrected I can use the C[i].adist
		double q = Q(i).adist(x,y);
		if (q > d) {
			return false;	// outside
		} else
		if (q >= -d)
			on = true;
	}
	return on;
} // isOn
#endif

/** 2D bounding box on viewer plane
 * @return bounding box of body
 * @WARNING z coordinates are meaningless
 */
BBox VBody::bbox2D() const
{
	BBox bb;
	for (int i=0; i<nC; i++) {
		if (V[i].size()<2) continue;
		bool isLine = C[i].type()==Conic::Type::LINE;
		const auto& VV = V[i];
		double tp = 0.0;
		for (int k=0; k<(int)VV.size(); k++) {
			const Segment& v = VV[k];
			bb.add(v.x, v.y, 0.0);
			if (!isLine && v.t-tp > 1.0) {
				// split it into several parts at least 6
				double tstep = (v.t - tp) / 6.0;
				for (double t=tp+tstep; t<v.t; t += tstep) {
					double x,y;
					C[i].getXY(t,&x,&y);
					bb.add(x, y, 0.0);
				}
			}
			tp = v.t;
		}
	}
	return bb;
} // bbox2D

/** @return memory used by body */
size_t VBody::memory() const
{
	size_t mem = sizeof(VBody);
	for (int i=0; i<nC; i++)
		mem += sizeof(V[i]) + V[i].capacity() * sizeof(Segment);
	return mem;
} // memory

/** operator << */
std::ostream& operator << (std::ostream& s, const VBody& body)
{
	s << "VBody: " << body.name() << "\t[" << body.typeStr() << "]" << endl;
	for (int i=0; i<body.nQ(); i++)
		s << "   Quad #" << (i+1) << endl << body.Q(i) << endl;
	return s;
} /* operator << */
