/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	13-Jun-2018
 */

#ifndef __GRAPH_ZPAINTER_H
#define __GRAPH_ZPAINTER_H

#include <limits>

#include "painter.h"

// ================================= ZPainter ==================================
/** ZPainter class a Painter including Z-buffer and super-sampling
 * z operations are using single precision "float" type for memory saving
 * x,y pixels position are integers
 */
class ZPainter : public Painter {
public:
constexpr static float	infinite = std::numeric_limits<float>::max();

protected:
	float*	_zdepth;		/** z-depth matrix same dimension as data	*/
	float	_znear;			/** z buffer frustum near limit (def 0)		*/
	float	_zfar;			/** z buffer frustum far limit (def=infinite)	*/
	dword*  _superSamplingBuf;	/** buffer that keeps the info if the pixel
					    has been super sampled */
public:
	ZPainter()
		: Painter(),
		 _zdepth(NULL),
		 _znear(0.0),
		 _zfar(infinite),
		 _superSamplingBuf(NULL)
		{}
//	ZPainter(const int w, const int h) : Painter(w,h), _zdepth(NULL) {}
	~ZPainter();

	bool	init(const int w, const int h);

	float*	depthBuf()		{ return _zdepth; }
	dword*	superSamplingBuf()	{ return _superSamplingBuf; }

	// Z buffer drawing
	void	clear();

	/** define z-frustum from near to far
	 * @param near	near z-range (default 0)
	 * @param far	far z-range (default infinite)
	 */
	void	frustum(float near=0, float far=infinite) {
				_znear = near;
				_zfar  = far;
			}

	// Warning UNSAFE operations
	/** check if x,y,z pixel is valid for drawing
	 * @param x,y	screen position
	 * @param z	z-depth
	 * @return true if near <= z < z-depth
	 */
	bool	zpixel(const int x, const int y, float z) {
				//assert(InRange(0,x,_width));
				//assert(InRange(0,y,_height));
				float* zptr = zpixelPtr(x,y);
				if (z<*zptr and _znear<=z) {
					*zptr = z;
					return true;
				}
				return false;
			}

	/** @return z-depth of x,y pixel */
	float	zpixel(const int x, const int y) const
			{ return _zdepth[y*_width+x]; }

	/** @return pixel pointer of x,y pixel */
	float*	zpixelPtr(const int x, const int y)
			{ return _zdepth + y*_width+x; }

	/** draw 3d z-pixel if depth is inside the range
	 * @param x,y	pixel position
	 * @param z	z depth
	 * @param color	color to draw
	 */
	bool	zpixel(const int x, const int y, float z, dword color) {
				//assert(InRange(0,x,_width));
				//assert(InRange(0,y,_height));
				float* zptr = zpixelPtr(x,y);
				if (z<*zptr and _znear<=z) {
					*zptr = z;
					pixel(x,y,color);
					return true;
				}
				return false;
			}

	/** draw 3d z-pixel if depth is inside the range
	 * else if farther than range, alpha blend it
	 * @param x,y	pixel position
	 * @param z	z depth
	 * @param color	color to draw
	 * @param alpha	alpha blending
	 */
	bool	zpixel(const int x, const int y, float z, dword color, int alpha) {
				//assert(InRange(0,x,_width));
				//assert(InRange(0,y,_height));
				float* zptr = zpixelPtr(x,y);
				if (z > *zptr) {
					if (alpha) {
						Color32 cc(color);
						dword* ptr = pixelPtr(x,y);
						cc.blend(*ptr, alpha);
						*ptr = cc;
						//dword* ptr = pixelPtr(x,y);
						//Color32 cptr(*ptr);
						//Color32 c32(color);
						//cptr.blend(c32, alpha^0xff);
						//*ptr = cptr;
					}
					return false;
				} else
				if (_znear<=z) {
					*zptr = z;
					pixel(x,y,color);
					return true;
				}
				return false;
			}

	/** alpha blend 3d drawing of z-pixel if depth is inside the range
	 * @param x,y	pixel position
	 * @param z	z depth
	 * @param color	color to draw
	 * @param alpha	alpha channel (0..255) to blend if z > z-depth
	 */
	//bool	zpixel(const int x, const int y, float z, dword color, int alpha)

	/** draw a 3d z-depth line
	 * @param x1,y1,z1	(int,int,float) po
	 */
	bool	zline(int x1, int y1, float z1, int x2, int y2, float z2,
				const dword color, int alpha=0) {
			int ox1 = x1;	// remember original points
			int oy1 = y1;
			int ox2 = x2;
			int oy2 = y2;
			if (_clip.line(&x1, &y1, &x2, &y2)) {
				float dx = (float)(ox2-ox1);
				float dy = (float)(oy2-oy1);
				float dz = z2-z1;
				float cz1, cz2;		// correct z1,2
				if (Abs(dx) > Abs(dy)) {
					cz1 = z1 + (float)(x1-ox1) * (dz/dx);
					cz2 = z2 + (float)(x2-ox2) * (dz/dx);
				} else
				if (dy==0.0) {	// integer difference, no need for small number check
					cz1 = z1;
					cz2 = z2;
				} else {
					cz1 = z1 + (float)(y1-oy1) * (dz/dy);
					cz2 = z2 + (float)(y2-oy2) * (dz/dy);
				}
				return unclippedZline(x1,y1,cz1,x2,y2,cz2,color,alpha);
			}
			return false;
		}

	// reset the super sampling buffer
	void	resetSuperSamplingBuf()	{ setSuperSamplingBuf(background()); }

	// is pixel super sampled
	bool	isPixelSuperSampled(const int x, const int y) {
				//assert(InRange(0,x,_width));
				//assert(InRange(0,y,_height));
				return (_superSamplingBuf[y*_width + x] != 0 );
			}

	// set super sampled pixel
	void setPixelSuperSampled(const int x, const int y, dword color) {
				//assert(InRange(0,x,_width));
				//assert(InRange(0,y,_height));
				_superSamplingBuf[y*_width + x] = color;
			}

	// set the complete super sampling buffer to a specific value
	void	setSuperSamplingBuf( dword val ) {
				std::fill_n( _superSamplingBuf, _maxSize, val );
		}

protected:
	bool	unclippedZline(int x1, int y1, float z1, int x2, int y2, float z2,
			const dword color, int alpha);

}; // ZPainter

#endif
