/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#ifndef __MATH_MATRIX_H
#define __MATH_MATRIX_H

#include <ostream>

#include "vector.h"

// ================================== Matrix ===================================
/**
 * Matrix class
 */
class Matrix {
private:
	int	_rows;		/** Matrix rows	*/
	int	_cols;		/** Matrix columns	*/
	double*	_data;		/** allocated data	*/

public:
	Matrix() : _rows(0), _cols(0), _data(NULL) {};
	Matrix(int r, int c=0)    : _rows(0), _cols(0), _data(NULL) { make(r,c); };
	Matrix(const Matrix& mat) : _rows(0), _cols(0), _data(NULL) { copy(mat); }
	~Matrix()	{ if (_data) delete [] _data; };

	Matrix&	operator =(const Matrix& matrix) { copy(matrix); return *this;}
	void	make(int r, int c=0);
	void	copy(const Matrix& src);

	// Get/Set data
	int	rows()		const { return _rows; }
	int	columns()	const { return _cols; }
	int	cols()		const { return _cols; }

	double	operator () (int r, int c) const {
			assert(r<_rows && c<_cols);
			return _data[r*_cols+c];
		}
	double&	operator () (int r, int c) {
			assert(r<_rows && c<_cols);
			return _data[r*_cols+c];
		}
	double	m(int r, int c)	const {return (*this)(r,c);}

	void	identity();
	void	zero();
	void	transpose();
	void	transpose(const Matrix& A);
	void	T() {transpose(); }

	void	multiply(const Matrix& A, const Matrix& B);
	void	multiplyT(const Matrix& A, const Matrix& B);
	void	multiply(const Matrix& B);
	/** operator *= */
	Matrix&	operator *= (const Matrix& B)	{ multiply(B); return *this; }
	void	negate();

	double	det() const;
	void	inverse();
	void	inverse(const Matrix& src) {
			copy(src);
			inverse();
		}

	double	trace() const;

	void	cleanupZero(const double eps);
private:
	void	lubksb(int* indx, double b[]);
	void	ludcmp(int* indx, double* d);
}; // Matrix

bool solveOverDetermined(Matrix& A, Matrix& B, Matrix& X);

std::ostream& operator << (std::ostream&, const Matrix&);
#endif
