/**
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	10-Dec-2018
 */

#ifndef __TOOL_SORTVECTOR_H
#define __TOOL_SORTVECTOR_H

// Populate a sorted vector
#include <vector>
#include <algorithm>

/** insert item so as to keep the vector sorted
 * @param vec	vector to populate
 * @param item	item to insert
 * @return the iterator position of the insertion
 */
template<typename T>
typename std::vector<T>::iterator insert_sorted(std::vector<T>& vec, T const& item)
{
	return vec.insert(
			std::upper_bound(vec.begin(), vec.end(), item),
			item);
} // insert_sorted

/** insert item so as to keep the vector sorted using a predicate function
 * @param vec	vector to populate
 * @param item	item to insert
 * @param pred	binary predicate function for comparison
 * @return the iterator position of the insertion
 */
template<typename T, typename Pred>
typename std::vector<T>::iterator insert_sorted(
			std::vector<T>& vec, T const& item, Pred pred)
{
	return vec.insert(
			std::upper_bound(vec.begin(), vec.end(), item, pred),
			item);
} // insert_sorted

/** find item in a sorted vector */
template<typename T>
bool find_sorted(std::vector<T>& vec, T const& item)
{
	// C++ 20
	//return std::binary_search(vec.begin(), vec.end(), item);
	auto first = std::lower_bound(vec.begin(), vec.end(), item);
	return first!=vec.end() and item==first;
} // find_sorted

/** find item in a sorted vector using a predicate function */
template<typename T, typename Pred>
bool find_sorted(std::vector<T>& vec, T const& item, Pred pred)
{
	// C++ 20
	//return std::binary_search(vec.begin(), vec.end(), item, pred);
	auto first = std::lower_bound(vec.begin(), vec.end(), item, pred);
	return first!=vec.end() and item==first;
} // find_sorted

// =============================== SortedVector ================================
template <typename T>
class SortedVector : public std::vector<T> {
public:
	/** add an item in a sorted vector */
	//std::vector<T>::iterator	add(T& item) {
	//auto	add(T& item) {
	void	add(T item) {
			this->insert(
				std::upper_bound(this->begin(), this->end(), item),
					item);
		}

	/** add an item in a sorted vector, using a comparison function */
	template<typename Cmp>
	void	add(T item, Cmp& cmp) {
			this->insert(
				std::upper_bound(this->begin(), this->end(), item, cmp),
					item);
		}

	/** @return true if item exists in vector */
	bool	has(T const item) const {
			// C++ 20
			//return std::binary_search(begin(), end(), item);
			auto pos = std::lower_bound(this->begin(), this->end(), item);
			return pos!=this->end() and *pos==item;
		}

	/** @return true if item exists in vector, using a comparison function */
	template<typename Cmp>
	bool	has(T const item, Cmp& cmp) const {
			auto pos = std::lower_bound(this->begin(), this->end(), item, cmp);
			return pos!=this->end() and *pos==item;
		}
}; // class SortedVector

#endif
