/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Extract uncertainty information from a USRBIN detector
 *
 * Authors: Christian.Theis@cern.ch
 *          Vasilis.Vlachoudis@cern.ch
 *
 * Date:    14-Apr-2023
 */

#include <iostream>
#include <fstream>
#include <string>
#include <stdexcept>

//#include "os.h"
#include "data/usrbin.h"

bool getCmdOption( const std::vector<std::string>& args, const std::string& optionFlag, std::string& optionVal );

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main( int argc, char** argv )

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	const std::vector<std::string> args( argv, argv + argc );

	cout << "USRBIN uncertainty extractor" << endl;
	cout << "Version 1.0 / 14/04/2023" << endl;
	cout << "Authors: C. Theis, V. Vlachoudis" << endl;
	cout << "--------------------------------" << endl << endl;

	Usrbin   inMesh, outMesh;
	int      detectorIndex = 1;
	string   inFileName, outFileName, detIndexStr;

	if ( argc < 2 ) {
		cout << "Syntax: usrbin_err_extractor -f inputfile -o outputfile -d detector-index" << endl;
		cout << "e.g.: usrbin_err_extractor -f test.bnn -o uncertainty-test.bnn -d 2" << endl;
		cout << "--> this will extract the uncertainties of the second detector in test.bnn and write them to uncertainty-test.bnn" << endl;
		return 0;
	}

	if ( !getCmdOption( args, "-f", inFileName ) ) {
		cout << "\nInput file: ";
		cin >> inFileName;
	}

	if ( !getCmdOption( args, "-o", outFileName ) ) {
		cout << "\nOutput file: ";
		cin >> outFileName;
	}

	if ( !getCmdOption( args, "-d", detIndexStr ) ) {
		auto detNames = Usrbin::getDetectorNames( inFileName.c_str() );
		cout << endl << "The data file includes " << detNames.size() << " data meshes." << endl;
		for ( auto curName : detNames ) {
			cout << detectorIndex++ << ".) " << curName << endl;
		}

		cout << "Please select the index (>=1) of the mesh of which you want to extract the uncertainty: ";
		cin >> detIndexStr;
	}
	else {
		try {
			detectorIndex = stoi( detIndexStr );
		}
		catch ( const std::invalid_argument& ia ) {
			std::cerr << "ERROR: invalid argument for detector index: " << ia.what() << '\n';
			return -1;
		}
	}



	if ( !inMesh.load( inFileName.c_str(), detectorIndex ) ) {
		cerr << "ERROR: Could not read detector # " << detectorIndex << " from " << inFileName << endl;
		return false;
	}

	outMesh.create( 1, inMesh.type(), inMesh.score(), inMesh.xlow, inMesh.xhigh, inMesh.nx,
									inMesh.ylow, inMesh.yhigh, inMesh.ny,
									inMesh.zlow, inMesh.zhigh, inMesh.nz );

	for ( int k = 0; k < inMesh.nz; ++k ) {
		for ( int j = 0; j < inMesh.ny; ++j ) {
			for ( int i = 0; i < inMesh.nx; ++i ) {
				double resRelErr = inMesh.getErr( i, j, k );
				outMesh.set( i, j, k, resRelErr );
			}
		}
	}

	outMesh.save( outFileName.c_str() );

	cout << "...done!" << endl;
	return 1;
}

bool getCmdOption( const std::vector<std::string>& args, const std::string& optionFlag, std::string& optionVal )
{
	for ( auto iter = args.begin(), end = args.end(); iter != end; ++iter ) {
		if ( *iter == optionFlag )
			if ( iter + 1 != args.end() ) {
				optionVal = *(iter + 1);
				return true;
			}
	}

	return false;
}
