/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#ifndef __VIEWER_D3_LAYER_H
#define __VIEWER_D3_LAYER_H

#include "layer.h"
#include "farm/farm.h"
#include "tool/stat.h"
#include "geometry/ray.h"
#include "geometry/light.h"
#include "geometry/filter.h"
#include "decorationlayer.h"

#include <map>

#ifdef THREAD
#	include "tool/threadpool.h"
#endif

#define MORTON

class Color;
class Color3D;
union Color32;
class D3Layer;
class ZPainter;
class GeometryEngine;

#ifdef THREAD
/* ============================= RayShooterWorker  ============================= */
// Base functor to fire rays via threads
class RayShooterWorker : public ThreadPoolWorker {
public:
	D3Layer*	d3;		/** link to D3layer			*/
	GeometryEngine*	engine;		/** engine assigned to worker		*/
	Ray		ray;		/** ray object				*/
	int		ii;		/** pixmap array position		*/
	int		jj;
	VZone*		zone;		/** last zone information		*/
	dword		color;		/** color info of the treated pixel	*/

public:
	RayShooterWorker() : d3(nullptr), engine(nullptr), zone(nullptr) {}
//virtual	void operator()();
}; // RayShooterWorker

/* ============================== Body3DWorker ============================== */
// Derived functor to render objects into the view buffer
class Body3DWorker : public RayShooterWorker {
public:
	ZPainter*	painter;	/** painter object containing all buffers */
	dword*		ptr;		/** pixmap pointer position	*/
	int		W;		/** width of screen		*/
	int		H;		/** height of screen		*/
	int		step;		/** step size			*/

public:
	Body3DWorker(): RayShooterWorker() {}
virtual	void operator()();
}; // Body3DWorker

/* ============================== Body3DFeeder ============================== */
class Body3DFeeder : public ThreadPoolFeeder {
public:
	int		nworkers;	/** number of workers allocated	*/
	Body3DWorker*	workers;	/** array of body workers	*/
	D3Layer*	d3;		/** link to D3layer		*/
	ZPainter*	painter;	/** painter object		*/
	dword*		ptr;		/** current pointer		*/
	int		W;		/** width of screen		*/
	int		H;		/** height of screen		*/
	int		step;		/** step size to scan		*/
	int		ii;		/** current i screen position	*/
	int		jj;		/** current j screen position	*/
private:
#ifndef MORTON
	dword*		pline;		/** current line		*/
#else
	//int		step_shift;	/** step shift			*/
	int		Ws,Hs;		/** W&H in steps		*/
	int		is,js;		/** ii,jj in steps		*/
#endif

public:
	Body3DFeeder(ThreadPool& p, D3Layer* d) :
			ThreadPoolFeeder(p),
			nworkers(0),
			workers(nullptr),
			d3(d),
			painter(nullptr)
			{}

virtual	~Body3DFeeder()		{ if (workers) delete [] workers; }

	// Initialize workers
	void	init(int astep);
	bool	loop();
	void	allocate();
virtual	void	reset(ZPainter* p, Ray& ray);

virtual	ThreadPoolWorker* feed(int threadId);
}; // Body3DFeeder
#endif

/* ================================== D3Layer ================================ */
class D3Layer : public Layer {
	using	VZoneColor = std::pair<VZone*, dword>;

public: /* 3D settings */
	bool	deflights;		/** default lights		*/
	int	ambient;		/** ambient intensity [0..255]	*/
	int	xray;			/** xray transparency		*/
	int	antialias;		/** super-sampling antialias	*/
	bool	drawEdges;		/** detect and draw edges	*/
	bool	shadows;		/** show shadows		*/
	Ray::SkipBlack	skip_black;	/** skip blackhole		*/
	int	light0;			/** light base 0=def, def=custom*/
	VLight	light[MAXLIGHT];	/** rel. lights in the system	*/

protected:
	bool	_usrbinastexture;	/** usrbin mapping as texture	*/
	int	_maxDepth;		/** enable reflections/refractions*/
	int	_step;			/** pixel step of 3D drawing	*/
	bool	_showErrors;		/** temporary copy of showErrors*/
	bool	_edgeDetect;		/** temporary copy of edgeDetect*/
	int	_samples;		/** temporary copy of antialias */
	bool	_beautyPassFinished;	/** flag that the beauty pass has been rendered */

	// CMitchellFilter   _pixelFilter;
	CLanczosSincFilter   _pixelFilter;  /* Lanczos Sinc filter have a bit more overshooting for step func
					       reconstruction but perform better than Mitchell for high frequencies
					       of sinoidal signals*/

	clock_t	_drawTime;		/** how long to draw		*/

	CFARM	m_RenderObj;		/* new renderer FARM (Flair Advanced Render Module) */

#ifdef THREAD
	Body3DFeeder	feeder;		/** threadpool body feeder	*/
#endif
	IFSTAT(Stat stat;)		/** Statistics                  */

public:
	D3Layer(const Geometry& g, GeometryKernel& k, GeometryViewer& v);

	// read configuration from farm.ini
	bool	readFARMConfig(const std::string& path)
			{ return m_RenderObj.ReadConfigFile(path); };

	bool	usrbinAsTexture()	const	{ return _usrbinastexture; }
	void	usrbinAsTexture(bool b)		{ _usrbinastexture = b; }

	bool	maxDepth()		const	{ return _maxDepth; }
	void	maxDepth(int b) {
			_maxDepth = b;
			if (_maxDepth >= 1)
				m_RenderObj.SetRenderMode(CSurfaceIntegrator::cQuality, _maxDepth);
			else
				m_RenderObj.SetRenderMode(CSurfaceIntegrator::cSpeed);
		}

	// ambient occlusion configuration
	int	ambientOcclusionSamples()
				{ return m_RenderObj.GetSurfaceInt()->GetAOCSamples(); }
	void	ambientOcclusionSamples(int val)
				{ m_RenderObj.GetSurfaceInt()->SetAOCSamples(val); }

	clock_t	drawTime()		const	{ return _drawTime?CLOCKS_PER_SEC/_drawTime:0; }
	void	drawTime(int div)		{ _drawTime = div? CLOCKS_PER_SEC/div : 0; }

	/* draw method */
	void	draw(ZPainter& painter, bool checkTime=false);
	void	postDraw(ZPainter& painter);

	bool	nextIntersection(GeometryEngine* eng, Ray* ray);
	Color	shade(GeometryEngine* eng, Ray* ray, const int step);
	Color32	shadeXray(GeometryEngine* eng, Ray* ray,
			const double u, const double v, const int step,
			int alpha=255, VRegion* last_region= nullptr);

	//bool	draw3Dline(ZPainter& painter,
	//		const Point& a, const Point& b,
	//		const dword color);

	void	drawWireframe(ZPainter& painter);
	void	drawWireframe(ZPainter& painter, VBody* body);

	void	drawBodiesBBox(ZPainter& painter);
	void	drawRegionsBBox(ZPainter& painter);

	// run a pass which adaptively performs anti-aliasing based on the existing image data
	void	drawAdaptiveAAPass(ZPainter& DC);

	CFilter* pixelFilter() { return &_pixelFilter; };

protected:
	bool	edgePixel(GeometryEngine* eng, Ray& ray, const double u, const double v, VZone* zone);
	bool	edgePixelNPR( GeometryEngine* eng, Ray& ray, const double u, const double v, VZone* zone, double& lineStrength );
	// determine the color of a pixel by shooting a ray
	dword	shootRay(double u, double v,
				Ray& ray,
				GeometryEngine* eng,
				VZone** zone,
				int step=1);
	void	drawPixelAA(GeometryEngine* eng,
				Ray& ray,
				ZPainter& painter, dword* ptr,
				const int W, const int H,
				const int i, const int j,
				const int step,
				VZone** zone);
	dword	drawPixel(GeometryEngine* eng,
				Ray& ray,
				ZPainter& painter, dword* ptr,
				const int W, const int H,
				const int i, const int j,
				const int step, const int nsample,
				const double u, const double v,
				VZone** zone);

	// check gradient in the vicinity of a specific pixel
	bool isGradientExcessive(ZPainter* pDC, int ii, int jj, const double cMaxDiff);

	// super sample a specific pixel of this layer
	void superSamplePixel(const int i, const int j, ZPainter& DC, int nrOfSamples);

friend class Body3DWorker;
friend class Body3DFeeder;
friend class RayShooterWorker;
}; // D3Layer

#endif
