/**
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	18-Aug-2023
 */

#include <stdio.h>
#include <assert.h>
#include <getopt.h>

#include <iostream>

#include "data/usrbin.h"
#include "data/vtkdata.h"

/** global variables */
char*	prgname;

double	norm = 1.0;

/* --- usage --- */
void usage()
{
	std::cout << "syntax: " << prgname <<" [options]" << std::endl;
	std::cout << "options:" << std::endl;
	std::cout << "\t-?, --help\t\tThis help page" << std::endl;
	std::cout << "\t-d, --detector #\tDetector load (default -1 = all)" << std::endl;
	std::cout << "\t-e, --error #\tFilename with vtk error data mesh" << std::endl;
	std::cout << "\t-n, --norm #\t\tNormalization factor" << std::endl;
	std::cout << "\t-u, --error #\tFilename with usrbin data mesh" << std::endl;
	std::cout << "\t-v, --vtk #\tFilename with vtk data mesh" << std::endl;
	std::cout << "\nauthor: Vasilis.Vlachoudis@cern.ch (2015)" << std::endl;
	exit(0);
} // usage

int main(int ac, char* av[])
{
	prgname = av[0];
	int	det = -1;
	std::string  vtk_file, vtkerror_file, usrbin_file;

	while (true) {
		int option_index = 0;
		static struct option long_options[] = {
			{"detector",required_argument, 0, 'd'},
			{"error",   required_argument, 0, 'e'},
			{"help",    no_argument,       0, '?'},
			{"norm",    required_argument, 0, 'n'},
			{"usrbin",  required_argument, 0, 'u'},
			{"vtk",     required_argument, 0, 'v'},
			{0, 0, 0, 0}
		};

		int c = getopt_long(ac, av, "?v:e:u:n:d:",
				long_options, &option_index);
		if (c == -1)
			break;

		switch (c) {
			case 0:
				usage();
				break;

			case '?':
				usage();
				break;

			case 'v':
				vtk_file = optarg;
				break;

			case 'e':
				vtkerror_file = optarg;
				break;

			case 'u':
				usrbin_file = optarg;
				break;

			case 'd':
				det = atoi(optarg);
				break;

			case 'n':
				norm = atof(optarg);
				if (norm<=0.0) {
					std::cerr << "ERROR invalid normalization value " << norm << std::endl;
					return -1;
				}
				break;

			default:
				std::cerr << "ERROR: ?? getopt returned character code " << c << std::endl;
				usage();
		}
	}
	if (vtk_file.empty()) {
		std::cerr << "ERROR: vtk data file was not provided\n" << std::endl;
		return 1;
	}
	if (usrbin_file.empty()) {
		std::cerr << "ERROR: usrbin file was not provided\n" << std::endl;
		return 1;
	}

	// signle detector
	if (det>=0) {
		VtkData vtkdata;	// Input vtk mesh data
		Usrbin  usrbin;		// Output usrbin mesh data
		// load data
		if (!vtkdata.load(vtk_file.c_str(),det)) {
			std::cerr << "ERROR: reading vtk data file\n" << std::endl;
			return 2;
		}

		// load possible error
		if (!vtkerror_file.empty())
			if (!vtkdata.loadError(vtkerror_file.c_str(),det)) {
				std::cerr << "ERROR: reading vtk error data file\n" << std::endl;
				return 2;
			}

		// copy to usrbin and save
		usrbin.copy(vtkdata);
		usrbin.type(Usrbin::Type::XYZ);
		usrbin.save(usrbin_file.c_str());
	} else {
		std::vector<Usrbin>  usrbins;	// Output usrbin mesh data
		for (int i=0; ; i++) {
			std::cout << "\n*-* Loading detector " << i+1 << std::endl;
			VtkData vtkdata;
			if (!vtkdata.load(vtk_file.c_str(), i)) break;
			if (!vtkerror_file.empty())
				vtkdata.loadError(vtkerror_file.c_str(), i);
			Usrbin usrbin;
			usrbin.copy(vtkdata);
			usrbin.type(Usrbin::Type::XYZ);
			usrbins.push_back(usrbin);
		}
		Usrbin::save(usrbin_file.c_str(), usrbins);
	}

	//std::cout << usrbin << std::endl;
	return 0;
} // main
